@@ -2,9 +2,9 @@ import logging
 
 import numpy as np
 from numpy.linalg import eigh, norm, svd
-from tqdm import tqdm
 
 from aspire.abinitio import CLOrient3D, SyncVotingMixin
+from aspire.operators import PolarFT
 from aspire.utils import (
     J_conjugate,
     Rotation,
@@ -12,7 +12,8 @@ from aspire.utils import (
     all_triplets,
     anorm,
     cyclic_rotations,
-    pairs_to_linear,
+    tqdm,
+    trange,
 )
 from aspire.utils.random import randn
 
@@ -43,13 +44,14 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
         src,
         symmetry=None,
         n_rad=None,
-        n_theta=None,
+        n_theta=360,
         max_shift=0.15,
         shift_step=1,
-        epsilon=1e-3,
+        epsilon=1e-2,
         max_iters=1000,
         degree_res=1,
         seed=None,
+        mask=True,
     ):
         """
         Initialize object for estimating 3D orientations for molecules with C3 and C4 symmetry.
@@ -57,13 +59,15 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
         :param src: The source object of 2D denoised or class-averaged images with metadata
         :param symmetry: A string, 'C3' or 'C4', indicating the symmetry type.
         :param n_rad: The number of points in the radial direction
-        :param n_theta: The number of points in the theta direction
+        :param n_theta: The number of points in the theta direction. Default = 360.
         :param max_shift: Maximum range for shifts as a proportion of resolution. Default = 0.15.
         :param shift_step: Resolution of shift estimation in pixels. Default = 1 pixel.
         :param epsilon: Tolerance for the power method.
         :param max_iter: Maximum iterations for the power method.
         :param degree_res: Degree resolution for estimating in-plane rotations.
         :param seed: Optional seed for RNG.
+        :param mask: Option to mask `src.images` with a fuzzy mask (boolean).
+            Default, `True`, applies a mask.
         """
 
         super().__init__(
@@ -72,8 +76,16 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
             n_theta=n_theta,
             max_shift=max_shift,
             shift_step=shift_step,
+            mask=mask,
         )
 
+        self._check_symmetry(symmetry)
+        self.epsilon = epsilon
+        self.max_iters = max_iters
+        self.degree_res = degree_res
+        self.seed = seed
+
+    def _check_symmetry(self, symmetry):
         if symmetry is None:
             raise NotImplementedError(
                 "Symmetry type not supplied. Please indicate C3 or C4 symmetry."
@@ -85,10 +97,6 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
                     f"Only C3 and C4 symmetry supported. {symmetry} was supplied."
                 )
             self.order = int(symmetry[1])
-        self.epsilon = epsilon
-        self.max_iters = max_iters
-        self.degree_res = degree_res
-        self.seed = seed
 
     def estimate_rotations(self):
         """
@@ -96,8 +104,7 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
 
         :return: Array of rotation matrices, size n_imgx3x3.
         """
-        logger.info(f"Estimating relative viewing directions for {self.n_img} images.")
-        vijs, viis = self._estimate_relative_viewing_directions_c3_c4()
+        vijs, viis = self._estimate_relative_viewing_directions()
 
         logger.info("Performing global handedness synchronization.")
         vijs, viis = self._global_J_sync(vijs, viis)
@@ -110,19 +117,20 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
 
         self.rotations = Ris
 
+        return self.rotations
+
     ###########################################
     # Primary Methods                         #
     ###########################################
 
-    def _estimate_relative_viewing_directions_c3_c4(self):
+    def _estimate_relative_viewing_directions(self):
         """
         Estimate the relative viewing directions vij = vi*vj^T, i<j, and vii = vi*vi^T, where
         vi is the third row of the i'th rotation matrix Ri.
         """
-
+        logger.info(f"Estimating relative viewing directions for {self.n_img} images.")
         # Step 1: Detect a single pair of common-lines between each pair of images
         self.build_clmatrix()
-        clmatrix = self.clmatrix
 
         # Step 2: Detect self-common-lines in each image
         sclmatrix = self._self_clmatrix_c3_c4()
@@ -131,7 +139,7 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
         Riis = self._estimate_all_Riis_c3_c4(sclmatrix)
 
         # Step 4: Calculate relative rotations
-        Rijs = self._estimate_all_Rijs_c3_c4(clmatrix)
+        Rijs = self._estimate_all_Rijs_c3_c4(self.clmatrix)
 
         # Step 5: Inner J-synchronization
         vijs, viis = self._local_J_sync_c3_c4(Rijs, Riis)
@@ -168,20 +176,21 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
         # We use the fact that if v_ii and v_ij are of the same handedness, then v_ii @ v_ij = v_ij.
         # If they are opposite handed then Jv_iiJ @ v_ij = v_ij. We compare each v_ii against all
         # previously synchronized v_ij to get a consensus on the handedness of v_ii.
+        _, pairs_to_linear = all_pairs(n_img, return_map=True)
         for i in range(n_img):
             vii = viis[i]
             vii_J = J_conjugate(vii)
             J_consensus = 0
             for j in range(n_img):
                 if j < i:
-                    idx = pairs_to_linear(n_img, j, i)
+                    idx = pairs_to_linear[j, i]
                     vji = vijs[idx]
 
                     err1 = norm(vji @ vii - vji)
                     err2 = norm(vji @ vii_J - vji)
 
                 elif j > i:
-                    idx = pairs_to_linear(n_img, i, j)
+                    idx = pairs_to_linear[i, j]
                     vij = vijs[idx]
 
                     err1 = norm(vii @ vij - vij)
@@ -269,7 +278,7 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
 
         # Step 1: Construct all rotation matrices Ri_tildes whose third rows are equal to
         # the corresponding third rows vis.
-        Ri_tildes = np.array([self._complete_third_row_to_rot(vi) for vi in vis])
+        Ri_tildes = self._complete_third_row_to_rot(vis)
 
         # Step 2: Construct all in-plane rotation matrices, R_theta_ijs.
         max_angle = (360 // order) * order
@@ -278,23 +287,20 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
 
         # Step 3: Compute the correlation over all shifts.
         # Generate shifts.
-        r_max = pf.shape[0]
+        r_max = pf.shape[-1]
         shifts, shift_phases, _ = self._generate_shift_phase_and_filter(
             r_max, max_shift_1d, shift_step
         )
         n_shifts = len(shifts)
-        all_shift_phases = shift_phases.T
 
         # Q is the n_img x n_img  Hermitian matrix defined by Q = q*q^H,
         # where q = (exp(i*order*theta_0), ..., exp(i*order*theta_{n_img-1}))^H,
         # and theta_i in [0, 2pi/order) is the in-plane rotation angle for the i'th image.
         Q = np.zeros((n_img, n_img), dtype=complex)
 
-        # Transpose pf and reconstruct the full polar Fourier for use in correlation.
-        # self.pf only consists of rays in the range [180, 360) and is in column major order,
-        # ie. self.pf has shape (n_rad-1, n_theta//2, n_img).
-        pf = pf.T
-        pf = np.concatenate((pf, np.conj(pf)), axis=1)
+        # Reconstruct the full polar Fourier for use in correlation. self.pf only consists of
+        # rays in the range [180, 360), with shape (n_img, n_theta//2, n_rad-1).
+        pf = PolarFT.half_to_full(pf)
 
         # Normalize rays.
         pf /= norm(pf, axis=-1)[..., np.newaxis]
@@ -309,7 +315,7 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
 
                 # Generate shifted versions of images.
                 pf_i_shifted = np.array(
-                    [pf_i * shift_phase for shift_phase in all_shift_phases]
+                    [pf_i * shift_phase for shift_phase in shift_phases]
                 )
 
                 Ri_tilde = Ri_tildes[i]
@@ -372,7 +378,9 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
             # Q is a rank-1 Hermitian matrix.
             eig_vals, eig_vecs = eigh(Q)
             leading_eig_vec = eig_vecs[:, -1]
-            logger.info(f"Top 5 eigenvalues of Q are {str(eig_vals[-5:][::-1])}.")
+            logger.info(
+                f"Top 3 eigenvalues of Q (rank-1) are {str(eig_vals[-3:][::-1])}."
+            )
 
             # Calculate R_thetas.
             R_thetas = Rotation.about_axis(
@@ -434,30 +442,27 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
 
         # Compute the correlation over all shifts.
         # Generate Shifts.
-        r_max = pf.shape[0]
+        r_max = pf.shape[-1]
         shifts, shift_phases, _ = self._generate_shift_phase_and_filter(
             r_max, max_shift_1d, shift_step
         )
         n_shifts = len(shifts)
-        all_shift_phases = shift_phases.T
 
-        # Transpose pf and reconstruct the full polar Fourier for use in correlation.
-        # self.pf only consists of rays in the range [180, 360) and is in column major order,
-        # ie. self.pf has shape (n_rad-1, n_theta//2, n_img).
-        pf = pf.T
-        pf_full = np.concatenate((pf, np.conj(pf)), axis=1)
+        # Reconstruct the full polar Fourier for use in correlation. self.pf only consists of
+        # rays in the range [180, 360), with shape (n_img, n_theta//2, n_rad-1).
+        pf_full = PolarFT.half_to_full(pf)
 
         # The self-common-lines matrix holds two indices per image that represent
         # the two self common-lines in the image.
         sclmatrix = np.zeros((n_img, 2))
 
-        for i in tqdm(range(n_img)):
+        for i in trange(n_img):
             pf_i = pf[i]
             pf_full_i = pf_full[i]
 
             # Generate shifted versions of images.
             pf_i_shifted = np.array(
-                [pf_i * shift_phase for shift_phase in all_shift_phases]
+                [pf_i * shift_phase for shift_phase in shift_phases]
             )
             pf_i_shifted = np.reshape(pf_i_shifted, (n_shifts * n_theta // 2, r_max))
 
@@ -494,23 +499,23 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
             # cos_diff should be <= 0.5, but due to discretization that might be violated.
             if np.max(cos_diff) > 0.5:
                 bad_diffs = np.count_nonzero(cos_diff > 0.5)
-                logger.warning(
+                logger.debug(
                     "cos(angular_diff) should be < 0.5."
                     f"Found {bad_diffs} estimates exceeding 0.5, with maximum {np.max(cos_diff)}."
-                    "Setting all bad estimates to 0.5."
+                    " Setting all bad estimates to 0.5."
                 )
-                cos_diff[cos_diff > 0.5] = 0.5
+                cos_diff = np.minimum(cos_diff, 0.5)
             euler_y2 = np.arccos(cos_diff / (1 - cos_diff))
         else:
             # cos_diff should be <= 0, but due to discretization that might be violated.
             if np.max(cos_diff) > 0:
                 bad_diffs = np.count_nonzero(cos_diff > 0)
-                logger.warning(
+                logger.debug(
                     "cos(angular_diff) should be < 0."
-                    f"Found {bad_diffs} estimates exceeding 0, with maximum {np.max(cos_diff)}"
-                    "Setting all bad estimates to 0."
+                    f"Found {bad_diffs} estimates exceeding 0, with maximum {np.max(cos_diff)}."
+                    " Setting all bad estimates to 0."
                 )
-                cos_diff[cos_diff > 0] = 0
+                cos_diff = np.minimum(cos_diff, 0)
             euler_y2 = np.arccos((1 + cos_diff) / (1 - cos_diff))
 
         # Calculate remaining Euler angles in ZYZ convention.
@@ -549,6 +554,7 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
 
         Given the common lines matrix `clmatrix`, a list of images specified in k_list
         and the number of common lines n_theta, find the (i, j) rotation block Rij.
+
         :param clmatrix: The common lines matrix
         :param i: The i image
         :param j: The j image
@@ -556,7 +562,7 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
         :param n_theta: The number of points in the theta direction (common lines)
         :return: The (i,j) rotation block of the synchronization matrix
         """
-        good_k = self._vote_ij(clmatrix, n_theta, i, j, k_list)
+        _, good_k = self._vote_ij(clmatrix, n_theta, i, j, k_list)
 
         rots = self._rotratio_eulerangle_vec(clmatrix, i, j, good_k, n_theta)
 
@@ -681,16 +687,19 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
         itr = 0
 
         # Power method iterations
+        logger.info(
+            "Initiating power method to estimate J-synchronization matrix eigenvector."
+        )
         while itr < max_iters and residual > epsilon:
             itr += 1
-            vec_new = self._signs_times_v(vijs, vec)
+            # Note, this appears to need double precision for accuracy in the following division.
+            vec_new = self._signs_times_v(vijs, vec).astype(np.float64, copy=False)
             vec_new = vec_new / norm(vec_new)
             residual = norm(vec_new - vec)
             vec = vec_new
-
-        logger.info(
-            f"Power method used {itr} iterations. Maximum iterations set to {max_iters}."
-        )
+            logger.info(
+                f"Iteration {itr}, residual {round(residual, 5)} (target {epsilon})"
+            )
 
         # We need only the signs of the eigenvector
         J_sync = np.sign(vec)
@@ -727,6 +736,7 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
         # All pairs (i,j) and triplets (i,j,k) where i<j<k
         n_img = self.n_img
         triplets = all_triplets(n_img)
+        pairs, pairs_to_linear = all_pairs(n_img, return_map=True)
 
         # There are 4 possible configurations of relative handedness for each triplet (vij, vjk, vik).
         # 'conjugate' expresses which node of the triplet must be conjugated (True) to achieve synchronization.
@@ -753,10 +763,10 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
         # condition. Finally, we the multiply the 'edge_signs' by the cooresponding entries of 'vec'.
         v = vijs
         new_vec = np.zeros_like(vec)
-        for (i, j, k) in triplets:
-            ij = pairs_to_linear(n_img, i, j)
-            jk = pairs_to_linear(n_img, j, k)
-            ik = pairs_to_linear(n_img, i, k)
+        for i, j, k in triplets:
+            ij = pairs_to_linear[i, j]
+            jk = pairs_to_linear[j, k]
+            ik = pairs_to_linear[i, k]
             vij, vjk, vik = v[ij], v[jk], v[ik]
             vij_J = J_conjugate(vij)
             vjk_J = J_conjugate(vjk)
@@ -777,7 +787,7 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
             # Update multiplication of signs times vec
             new_vec[ij] += s_ij_jk * vec[jk] + s_ij_ik * vec[ik]
             new_vec[jk] += s_ij_jk * vec[ij] + s_ik_jk * vec[ik]
-            new_vec[ik] += s_ij_jk * vec[ij] + s_ik_jk * vec[jk]
+            new_vec[ik] += s_ij_ik * vec[ij] + s_ik_jk * vec[jk]
 
         return new_vec
 
@@ -787,31 +797,51 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
     @staticmethod
     def _complete_third_row_to_rot(r3):
         """
-        Construct a rotation matrix whose third row is equal to the given row vector.
+        Construct rotation matrices whose third rows are equal to the given row vectors.
         For vector r3 = [a, b, c], where [a, b, c] != [0, 0, 1], we return the matrix
         with rows r1, r2, r3, given by:
 
         r1 = 1/sqrt(a^2 + b^2)[b, -a, 0],
         r2 = 1/sqrt(a^2 + b^2)[ac, bc, -(a^2 + b^2)].
 
-        :param r3: A 1x3 vector of norm 1.
-        :return: A 3x3 rotation matrix whose third row is r3.
+        :param r3: A nx3 array where each row vector has norm 1.
+        :return: An nx3x3 array of rotation matrices whose third rows are r3.
         """
 
+        # Handle singleton vector.
+        singleton = False
+        if r3.shape == (3,):
+            r3 = np.expand_dims(r3, axis=0)
+            singleton = True
+
+        # Initialize output rotation matrices.
+        rots = np.zeros((len(r3), 3, 3), dtype=r3.dtype)
+
+        # Populate 3rd rows.
+        rots[:, 2] = r3
+
+        # Mask for third rows that do not coincide with the z-axis.
+        mask = np.linalg.norm(r3 - [0, 0, 1], axis=1) >= 1e-5
+
         # If the third row coincides with the z-axis we return the identity matrix.
-        if norm(r3 - [0, 0, 1]) < 1e-5:
-            return np.eye(3)
+        rots[~mask] = np.eye(3, dtype=r3.dtype)
 
-        # 'norm_fac' is non-zero since r3 does not coincide with the z-axis.
-        norm_12 = np.sqrt(r3[0] ** 2 + r3[1] ** 2)
+        # 'norm_12' is non-zero since r3 does not coincide with the z-axis.
+        norm_12 = np.sqrt(r3[mask, 0] ** 2 + r3[mask, 1] ** 2)
 
-        # Construct an orthogonal row vector of norm 1.
-        r1 = np.array([r3[1] / norm_12, -r3[0] / norm_12, 0])
+        # Populate 1st rows with vector orthogonal to row 3.
+        rots[mask, 0, 0] = r3[mask, 1] / norm_12
+        rots[mask, 0, 1] = -r3[mask, 0] / norm_12
 
-        # Construct r2 so that r3 = r1xr2
-        r2 = np.array([r3[0] * r3[2] / norm_12, r3[1] * r3[2] / norm_12, -norm_12])
+        # Populate 2nd rows such that r3 = r1 x r2
+        rots[mask, 1, 0] = r3[mask, 0] * r3[mask, 2] / norm_12
+        rots[mask, 1, 1] = r3[mask, 1] * r3[mask, 2] / norm_12
+        rots[mask, 1, 2] = -norm_12
 
-        return np.vstack((r1, r2, r3))
+        if singleton:
+            rots = rots.reshape(3, 3)
+
+        return rots
 
     @staticmethod
     def cl_angles_to_ind(cl_angles, n_theta):
@@ -821,7 +851,13 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
         thetas = np.mod(thetas, 2 * np.pi)
 
         # linear scale from [0,2*pi) to [0,n_theta).
-        return np.mod(np.round(thetas / (2 * np.pi) * n_theta), n_theta).astype(int)
+        ind = np.mod(np.round(thetas / (2 * np.pi) * n_theta), n_theta).astype(int)
+
+        # Return scalar for single value.
+        if ind.size == 1:
+            ind = ind.flat[0]
+
+        return ind
 
     @staticmethod
     def g_sync(rots, order, rots_gt):
@@ -849,7 +885,7 @@ class CLSymmetryC3C4(CLOrient3D, SyncVotingMixin):
 
         pairs = all_pairs(n_img)
 
-        for (i, j) in pairs:
+        for i, j in pairs:
             Ri = rots[i]
             Rj = rots[j]
             Rij = Ri.T @ Rj
