@@ -1,7 +1,8 @@
 # We'll tell isort not to sort these base classes
 # isort: off
 
-from .basis import Basis
+from .basis import Basis, Coef, ComplexCoef
+from .dirac import DiracBasis, DiracBasis2D, DiracBasis3D
 from .steerable import SteerableBasis2D
 from .fb import FBBasisMixin
 
@@ -11,9 +12,9 @@ from .fb_2d import FBBasis2D
 from .fb_3d import FBBasis3D
 from .ffb_2d import FFBBasis2D
 from .ffb_3d import FFBBasis3D
+from .fle_2d import FLEBasis2D
 from .fpswf_2d import FPSWFBasis2D
 from .fpswf_3d import FPSWFBasis3D
 from .fspca import FSPCABasis
-from .polar_2d import PolarBasis2D
 from .pswf_2d import PSWFBasis2D
 from .pswf_3d import PSWFBasis3D
