@@ -5,8 +5,6 @@ from scipy.special import jv
 
 from aspire.basis import FBBasisMixin, SteerableBasis2D
 from aspire.basis.basis_utils import unique_coords_nd
-from aspire.utils import complex_type, real_type, roll_dim, unroll_dim
-from aspire.utils.matlab_compat import m_flatten, m_reshape
 
 logger = logging.getLogger(__name__)
 
@@ -31,9 +29,9 @@ class FBBasis2D(SteerableBasis2D, FBBasisMixin):
             May be a 2-tuple or an integer, in which case a square basis is assumed.
             Currently only square images are supported.
         :ell_max: The maximum order ell of the basis elements. If no input
-            (= None), it will be set to np.Inf and the basis includes all
+            (= None), it will be set to np.inf and the basis includes all
             ell such that the resulting basis vectors are concentrated
-            below the Nyquist frequency (default Inf).
+            below the Nyquist frequency (default inf).
         """
 
         if isinstance(size, int):
@@ -63,7 +61,6 @@ class FBBasis2D(SteerableBasis2D, FBBasisMixin):
 
         # generate 1D indices for basis functions
         self._compute_indices()
-        self._indices = self.indices()
 
         # get normalized factors
         self.radial_norms, self.angular_norms = self.norms()
@@ -109,19 +106,6 @@ class FBBasis2D(SteerableBasis2D, FBBasisMixin):
         self.angular_indices = indices_ells
         self.radial_indices = indices_ks
         self.signs_indices = indices_sgns
-        # Relating to paper: a[i] = a_ell_ks = a_angularindices[i]_radialindices[i]
-        self.complex_angular_indices = indices_ells[self._pos]  # k
-        self.complex_radial_indices = indices_ks[self._pos]  # q
-
-    def indices(self):
-        """
-        Return the precomputed indices for each basis function.
-        """
-        return {
-            "ells": self.angular_indices,
-            "ks": self.radial_indices,
-            "sgns": self.signs_indices,
-        }
 
     def _precomp(self):
         """
@@ -143,7 +127,7 @@ class FBBasis2D(SteerableBasis2D, FBBasisMixin):
             for k in range(1, self.k_max[ell] + 1):
                 # Only normalized by the radial part of basis function
                 radial[:, ind_radial] = (
-                    jv(ell, self.r0[k - 1, ell] * r_unique)
+                    jv(ell, self.r0[ell][k - 1] * r_unique)
                     / self.radial_norms[ind_radial]
                 )
                 ind_radial += 1
@@ -177,7 +161,7 @@ class FBBasis2D(SteerableBasis2D, FBBasisMixin):
         Calculate the normalized factors from radial and angular parts of a specified basis function
         """
         rad_norm = (
-            np.abs(jv(ell + 1, self.r0[k - 1, ell]))
+            np.abs(jv(ell + 1, self.r0[ell][k - 1]))
             * np.sqrt(1 / 2.0)
             * self.nres
             / 2.0
@@ -198,18 +182,19 @@ class FBBasis2D(SteerableBasis2D, FBBasisMixin):
             This is an array whose last dimensions equal `self.sz` and the remaining
             dimensions correspond to first dimensions of `v`.
         """
-        # Transpose here once, instead of several times below  #RCOPT
-        v = v.reshape(-1, self.count).T
+        n_data = v.shape[0]
+        # Transpose here once, instead of several times below
+        v = v.reshape(n_data, self.count).T
 
         r_idx = self.basis_coords["r_idx"]
         ang_idx = self.basis_coords["ang_idx"]
-        mask = m_flatten(self.basis_coords["mask"])
+        mask = self.basis_coords["mask"].flatten()
 
         ind = 0
         ind_radial = 0
         ind_ang = 0
 
-        x = np.zeros(shape=tuple([np.prod(self.sz)] + list(v.shape[1:])), dtype=v.dtype)
+        x = np.zeros((np.prod(self.sz), n_data), dtype=v.dtype)
         for ell in range(0, self.ell_max + 1):
             k_max = self.k_max[ell]
             idx_radial = ind_radial + np.arange(0, k_max, dtype=int)
@@ -230,37 +215,32 @@ class FBBasis2D(SteerableBasis2D, FBBasisMixin):
 
             ind_radial += len(idx_radial)
 
-        x = x.T.reshape(-1, *self.sz)  # RCOPT
+        x = x.T.reshape(n_data, *self.sz)
 
         return x
 
-    def _evaluate_t(self, v):
+    def _evaluate_t(self, x):
         """
         Evaluate coefficient in FB basis from those in standard 2D coordinate basis
 
-        :param v: The coefficient array to be evaluated. The last dimensions
+        :param x: The coefficient array to be evaluated. The last dimensions
             must equal `self.sz`.
         :return: The evaluation of the coefficient array `v` in the dual basis
             of `basis`. This is an array of vectors whose last dimension equals
             `self.count` and whose first dimensions correspond to
             first dimensions of `v`.
         """
-        v = v.asnumpy().T  # RCOPT
-
-        x, sz_roll = unroll_dim(v, self.ndim + 1)
-        x = m_reshape(
-            x, new_shape=tuple([np.prod(self.sz)] + list(x.shape[self.ndim :]))
-        )
+        x = x.reshape(x.shape[0], np.prod(self.sz))
 
         r_idx = self.basis_coords["r_idx"]
         ang_idx = self.basis_coords["ang_idx"]
-        mask = m_flatten(self.basis_coords["mask"])
+        mask = self.basis_coords["mask"].flatten()
 
         ind = 0
         ind_radial = 0
         ind_ang = 0
 
-        v = np.zeros(shape=tuple([self.count] + list(x.shape[1:])), dtype=v.dtype)
+        v = np.zeros((x.shape[0], self.count), dtype=x.dtype)
         for ell in range(0, self.ell_max + 1):
             k_max = self.k_max[ell]
             idx_radial = ind_radial + np.arange(0, k_max)
@@ -274,107 +254,13 @@ class FBBasis2D(SteerableBasis2D, FBBasisMixin):
                 ang = self._precomp["ang"][:, ind_ang]
                 ang_radial = np.expand_dims(ang[ang_idx], axis=1) * radial[r_idx]
                 idx = ind + np.arange(0, k_max)
-                v[idx] = ang_radial.T @ x[mask]
+                v[:, idx] = x[:, mask] @ ang_radial
                 ind += len(idx)
                 ind_ang += 1
 
             ind_radial += len(idx_radial)
 
-        v = roll_dim(v, sz_roll)
-        return v.T  # RCOPT
-
-    def to_complex(self, coef):
-        """
-        Return complex valued representation of coefficients.
-        This can be useful when comparing or implementing methods
-        from literature.
-
-        There is a corresponding method, to_real.
-
-        :param coef: Coefficients from this basis.
-        :return: Complex coefficent representation from this basis.
-        """
-
-        if coef.ndim == 1:
-            coef = coef.reshape(1, -1)
-
-        if coef.dtype not in (np.float64, np.float32):
-            raise TypeError("coef provided to to_complex should be real.")
-
-        # Pass through dtype precions, but check and warn if mismatched.
-        dtype = complex_type(coef.dtype)
-        if coef.dtype != self.dtype:
-            logger.warning(
-                f"coef dtype {coef.dtype} does not match precision of basis.dtype {self.dtype}, returning {dtype}."
-            )
-
-        # Return the same precision as coef
-        imaginary = dtype(1j)
-
-        ccoef = np.zeros((coef.shape[0], self.complex_count), dtype=dtype)
-
-        ind = 0
-        idx = np.arange(self.k_max[0], dtype=int)
-        ind += np.size(idx)
-
-        ccoef[:, idx] = coef[:, idx]
-
-        for ell in range(1, self.ell_max + 1):
-            idx = ind + np.arange(self.k_max[ell], dtype=int)
-            ccoef[:, idx] = (
-                coef[:, self._pos[idx]] - imaginary * coef[:, self._neg[idx]]
-            ) / 2.0
-
-            ind += np.size(idx)
-
-        return ccoef
-
-    def to_real(self, complex_coef):
-        """
-        Return real valued representation of complex coefficients.
-        This can be useful when comparing or implementing methods
-        from literature.
-
-        There is a corresponding method, to_complex.
-
-        :param complex_coef: Complex coefficients from this basis.
-        :return: Real coefficent representation from this basis.
-        """
-        if complex_coef.ndim == 1:
-            complex_coef = complex_coef.reshape(1, -1)
-
-        if complex_coef.dtype not in (np.complex128, np.complex64):
-            raise TypeError("coef provided to to_real should be complex.")
-
-        # Pass through dtype precions, but check and warn if mismatched.
-        dtype = real_type(complex_coef.dtype)
-        if dtype != self.dtype:
-            logger.warning(
-                f"Complex coef dtype {complex_coef.dtype} does not match precision of basis.dtype {self.dtype}, returning {dtype}."
-            )
-
-        coef = np.zeros((complex_coef.shape[0], self.count), dtype=dtype)
-
-        ind = 0
-        idx = np.arange(self.k_max[0], dtype=int)
-        ind += np.size(idx)
-        ind_pos = ind
-
-        coef[:, idx] = complex_coef[:, idx].real
-
-        for ell in range(1, self.ell_max + 1):
-            idx = ind + np.arange(self.k_max[ell], dtype=int)
-            idx_pos = ind_pos + np.arange(self.k_max[ell], dtype=int)
-            idx_neg = idx_pos + self.k_max[ell]
-
-            c = complex_coef[:, idx]
-            coef[:, idx_pos] = 2.0 * np.real(c)
-            coef[:, idx_neg] = -2.0 * np.imag(c)
-
-            ind += np.size(idx)
-            ind_pos += 2 * self.k_max[ell]
-
-        return coef
+        return v
 
     def calculate_bispectrum(
         self, coef, flatten=False, filter_nonzero_freqs=False, freq_cutoff=None
@@ -403,16 +289,8 @@ class FBBasis2D(SteerableBasis2D, FBBasisMixin):
             freq_cutoff=freq_cutoff,
         )
 
-    def rotate(self, coef, radians, refl=None):
+    def filter_to_basis_mat(self, *args, **kwargs):
         """
-        Returns coefs rotated by `radians`.
-
-        :param coef: Basis coefs.
-        :param radians: Rotation in radians.
-        :param refl: Optional reflect image (bool)
-        :return: rotated coefs.
+        See `SteerableBasis2D.filter_to_basis_mat`.
         """
-
-        # Base class rotation expects complex representation of coefficients.
-        #  Convert, rotate and convert back to real representation.
-        return self.to_real(super().rotate(self.to_complex(coef), radians, refl))
+        return super().filter_to_basis_mat(*args, **kwargs)
