@@ -4,9 +4,6 @@ import numpy as np
 
 from aspire.basis import Basis, FBBasisMixin
 from aspire.basis.basis_utils import real_sph_harmonic, sph_bessel, unique_coords_nd
-from aspire.utils import roll_dim, unroll_dim
-from aspire.utils.matlab_compat import m_flatten, m_reshape
-from aspire.volume import Volume
 
 logger = logging.getLogger(__name__)
 
@@ -27,9 +24,9 @@ class FBBasis3D(Basis, FBBasisMixin):
             May be a 3-tuple or an integer, in which case a cubic basis is assumed.
             Currently only cubic images are supported.
         :param ell_max: The maximum order ell of the basis elements. If no input
-            (= None), it will be set to np.Inf and the basis includes all
+            (= None), it will be set to np.inf and the basis includes all
             ell such that the resulting basis vectors are concentrated
-            below the Nyquist frequency (default Inf).
+            below the Nyquist frequency (default inf).
         """
         if isinstance(size, int):
             size = (size, size, size)
@@ -105,7 +102,7 @@ class FBBasis3D(Basis, FBBasisMixin):
 
         for ell in range(0, self.ell_max + 1):
             for k in range(1, self.k_max[ell] + 1):
-                radial[:, ind_radial] = sph_bessel(ell, self.r0[k - 1, ell] * r_unique)
+                radial[:, ind_radial] = sph_bessel(ell, self.r0[ell][k - 1] * r_unique)
                 ind_radial += 1
 
             for m in range(-ell, ell + 1):
@@ -136,7 +133,7 @@ class FBBasis3D(Basis, FBBasisMixin):
         Calculate the normalized factor of a specified basis function.
         """
         return (
-            np.abs(sph_bessel(ell + 1, self.r0[k - 1, ell]))
+            np.abs(sph_bessel(ell + 1, self.r0[ell][k - 1]))
             / np.sqrt(2)
             * np.sqrt((self.nres / 2) ** 3)
         )
@@ -151,21 +148,15 @@ class FBBasis3D(Basis, FBBasisMixin):
             This is an array whose first dimensions equal `self.z` and the
             remaining dimensions correspond to dimensions two and higher of `v`.
         """
-
-        v = v.T
-        v, sz_roll = unroll_dim(v, 2)
-
         r_idx = self.basis_coords["r_idx"]
         ang_idx = self.basis_coords["ang_idx"]
-        mask = m_flatten(self.basis_coords["mask"])
+        mask = self.basis_coords["mask"].flatten()
 
         ind = 0
         ind_radial = 0
         ind_ang = 0
 
-        x = np.zeros(
-            shape=tuple([np.prod(self.sz)] + list(v.shape[1:])), dtype=self.dtype
-        )
+        x = np.zeros((v.shape[0], np.prod(self.sz)), dtype=self.dtype)
         for ell in range(0, self.ell_max + 1):
             k_max = self.k_max[ell]
             idx_radial = ind_radial + np.arange(0, k_max)
@@ -177,47 +168,35 @@ class FBBasis3D(Basis, FBBasisMixin):
                 ang = self._precomp["ang"][:, ind_ang]
                 ang_radial = np.expand_dims(ang[ang_idx], axis=1) * radial[r_idx]
                 idx = ind + np.arange(0, len(idx_radial))
-                x[mask] += ang_radial @ v[idx]
+                x[:, mask] += v[:, idx] @ ang_radial.T
                 ind += len(idx)
                 ind_ang += 1
 
             ind_radial += len(idx_radial)
 
-        x = m_reshape(x, self.sz + x.shape[1:])
-        x = roll_dim(x, sz_roll)
+        return x.reshape(v.shape[0], *self.sz)
 
-        return x.T
-
-    def _evaluate_t(self, v):
+    def _evaluate_t(self, x):
         """
         Evaluate coefficient in FB basis from those in standard 3D coordinate basis
 
-        :param v: The coefficient array to be evaluated. The first dimensions
+        :param x: The coefficient array to be evaluated. The first dimensions
             must equal `self.sz`.
         :return: The evaluation of the coefficient array `v` in the dual
             basis of `basis`. This is an array of vectors whose first dimension
             equals `self.count` and whose remaining dimensions correspond
             to higher dimensions of `v`.
         """
-        # v may be a Volume object or a 7D array passed from Basis.mat_evaluate_t
-        # making this check important
-        if isinstance(v, Volume):
-            v = v.asnumpy()
-        v = v.T
-        x, sz_roll = unroll_dim(v, self.ndim + 1)
-        x = m_reshape(
-            x, new_shape=tuple([np.prod(self.sz)] + list(x.shape[self.ndim :]))
-        )
-
+        x = x.reshape(x.shape[0], np.prod(self.sz))
         r_idx = self.basis_coords["r_idx"]
         ang_idx = self.basis_coords["ang_idx"]
-        mask = m_flatten(self.basis_coords["mask"])
+        mask = self.basis_coords["mask"].flatten()
 
         ind = 0
         ind_radial = 0
         ind_ang = 0
 
-        v = np.zeros(shape=tuple([self.count] + list(x.shape[1:])), dtype=self.dtype)
+        v = np.zeros((x.shape[0], self.count), dtype=self.dtype)
         for ell in range(0, self.ell_max + 1):
             k_max = self.k_max[ell]
             idx_radial = ind_radial + np.arange(0, k_max)
@@ -229,11 +208,10 @@ class FBBasis3D(Basis, FBBasisMixin):
                 ang = self._precomp["ang"][:, ind_ang]
                 ang_radial = np.expand_dims(ang[ang_idx], axis=1) * radial[r_idx]
                 idx = ind + np.arange(0, len(idx_radial))
-                v[idx] = np.real(ang_radial.T @ x[mask])
+                v[:, idx] = x[:, mask] @ ang_radial
                 ind += len(idx)
                 ind_ang += 1
 
             ind_radial += len(idx_radial)
 
-        v = roll_dim(v, sz_roll)
-        return v.T
+        return v
