@@ -8,8 +8,8 @@ from aspire.basis import FBBasis2D
 from aspire.basis.basis_utils import lgwt
 from aspire.nufft import anufft, nufft
 from aspire.numeric import fft, xp
+from aspire.operators import BlkDiagMatrix
 from aspire.utils import complex_type
-from aspire.utils.matlab_compat import m_reshape
 
 logger = logging.getLogger(__name__)
 
@@ -50,7 +50,6 @@ class FFBBasis2D(FBBasis2D):
 
         # generate 1D indices for basis functions
         self._compute_indices()
-        self._indices = self.indices()
 
         # get normalized factors
         self.radial_norms, self.angular_norms = self.norms()
@@ -58,6 +57,16 @@ class FFBBasis2D(FBBasis2D):
         # precompute the basis functions in 2D grids
         self._precomp = self._precomp()
 
+        # include the normalization factor of angular part into radial part
+        self.radial_norm = xp.asarray(self._precomp["radial"]) / xp.asarray(
+            np.expand_dims(self.angular_norms, 1)
+        )
+
+        # precompute weighted nodes
+        self.gl_weighted_nodes = xp.asarray(self._precomp["gl_weights"]) * xp.asarray(
+            self._precomp["gl_nodes"]
+        )
+
     def _precomp(self):
         """
         Precomute the basis functions on a polar Fourier grid
@@ -74,7 +83,7 @@ class FFBBasis2D(FBBasis2D):
         ind_radial = 0
         for ell in range(0, self.ell_max + 1):
             for k in range(1, self.k_max[ell] + 1):
-                radial[ind_radial] = jv(ell, self.r0[k - 1, ell] * r / self.kcut)
+                radial[ind_radial] = jv(ell, self.r0[ell][k - 1] * r / self.kcut)
                 # NOTE: We need to remove the factor due to the discretization here
                 # since it is already included in our quadrature weights
                 # Only normalized by the radial part of basis function
@@ -83,24 +92,13 @@ class FFBBasis2D(FBBasis2D):
                 ind_radial += 1
 
         # Only calculate "positive" frequencies in one half-plane.
-        freqs_x = m_reshape(r, (n_r, 1)) @ m_reshape(
-            np.cos(np.arange(n_theta, dtype=self.dtype) * 2 * pi / (2 * n_theta)),
-            (1, n_theta),
-        )
-        freqs_y = m_reshape(r, (n_r, 1)) @ m_reshape(
-            np.sin(np.arange(n_theta, dtype=self.dtype) * 2 * pi / (2 * n_theta)),
-            (1, n_theta),
-        )
+        theta_grid = np.arange(n_theta, dtype=self.dtype) * 2 * pi / (2 * n_theta)
+        freqs_x = r[:, None] @ np.cos(theta_grid)[None, :]
+        freqs_y = r[:, None] @ np.sin(theta_grid)[None, :]
         freqs = np.vstack((freqs_y[np.newaxis, ...], freqs_x[np.newaxis, ...]))
 
         return {"gl_nodes": r, "gl_weights": w, "radial": radial, "freqs": freqs}
 
-    def get_radial(self):
-        """
-        Return precomputed radial part
-        """
-        return self._precomp["radial"]
-
     def _evaluate(self, v):
         """
         Evaluate coefficients in standard 2D coordinate basis from those in FB basis
@@ -111,33 +109,29 @@ class FFBBasis2D(FBBasis2D):
             coordinate basis. This is Image instance with resolution of `self.sz`
             and the first dimension correspond to remaining dimension of `v`.
         """
-        sz_roll = v.shape[:-1]
-        v = v.reshape(-1, self.count)
+        v = xp.asarray(v)
 
         # number of 2D image samples
         n_data = v.shape[0]
 
         # get information on polar grids from precomputed data
-        n_theta = np.size(self._precomp["freqs"], 2)
-        n_r = np.size(self._precomp["freqs"], 1)
+        n_theta = self._precomp["freqs"].shape[2]
+        n_r = self._precomp["freqs"].shape[1]
 
         # go through  each basis function and find corresponding coefficient
-        pf = np.zeros((n_data, 2 * n_theta, n_r), dtype=complex_type(self.dtype))
-        mask = self._indices["ells"] == 0
+        pf = xp.zeros((2 * n_theta, n_data, n_r), dtype=complex_type(self.dtype))
 
         ind = 0
 
         idx = ind + np.arange(self.k_max[0], dtype=int)
 
-        # include the normalization factor of angular part into radial part
-        radial_norm = self._precomp["radial"] / np.expand_dims(self.angular_norms, 1)
-        pf[:, 0, :] = v[:, mask] @ radial_norm[idx]
-        ind = ind + np.size(idx)
+        pf[0] = v[:, self._zero_angular_inds] @ self.radial_norm[idx]
+        ind = ind + idx.size
 
         ind_pos = ind
 
         for ell in range(1, self.ell_max + 1):
-            idx = ind + np.arange(self.k_max[ell], dtype=int)
+            idx = ind + xp.arange(self.k_max[ell], dtype=int)
             idx_pos = ind_pos + np.arange(self.k_max[ell], dtype=int)
             idx_neg = idx_pos + self.k_max[ell]
 
@@ -146,40 +140,36 @@ class FFBBasis2D(FBBasis2D):
             if np.mod(ell, 2) == 1:
                 v_ell = 1j * v_ell
 
-            pf_ell = v_ell @ radial_norm[idx]
-            pf[:, ell, :] = pf_ell
+            pf_ell = v_ell @ self.radial_norm[idx]
+            pf[ell] = pf_ell
 
             if np.mod(ell, 2) == 0:
-                pf[:, 2 * n_theta - ell, :] = pf_ell.conjugate()
+                pf[2 * n_theta - ell] = pf_ell.conjugate()
             else:
-                pf[:, 2 * n_theta - ell, :] = -pf_ell.conjugate()
+                pf[2 * n_theta - ell] = -pf_ell.conjugate()
 
-            ind = ind + np.size(idx)
+            ind = ind + idx.size
             ind_pos = ind_pos + 2 * self.k_max[ell]
 
         # 1D inverse FFT in the degree of polar angle
-        pf = 2 * pi * xp.asnumpy(fft.ifft(xp.asarray(pf), axis=1))
+        pf = 2 * xp.pi * fft.ifft(pf, axis=0)
 
         # Only need "positive" frequencies.
-        hsize = int(np.size(pf, 1) / 2)
-        pf = pf[:, 0:hsize, :]
-
-        for i_r in range(0, n_r):
-            pf[..., i_r] = pf[..., i_r] * (
-                self._precomp["gl_weights"][i_r] * self._precomp["gl_nodes"][i_r]
-            )
-
-        pf = np.reshape(pf, (n_data, n_r * n_theta))
+        hsize = int(pf.shape[0] / 2)
+        pf = pf[0:hsize]
+        pf *= self.gl_weighted_nodes[None, None, :]
+        pf = pf.transpose(1, 2, 0)
+        pf = pf.reshape(n_data, n_r * n_theta)
 
         # perform inverse non-uniformly FFT transform back to 2D coordinate basis
-        freqs = m_reshape(self._precomp["freqs"], (2, n_r * n_theta))
+        freqs = self._precomp["freqs"].reshape(2, n_r * n_theta)
 
         x = 2 * anufft(pf, 2 * pi * freqs, self.sz, real=True)
 
         # Return X as Image instance with the last two dimensions as *self.sz
-        x = x.reshape((*sz_roll, *self.sz))
+        x = x.reshape((n_data, *self.sz))
 
-        return x
+        return xp.asnumpy(x)
 
     def _evaluate_t(self, x):
         """
@@ -187,9 +177,9 @@ class FFBBasis2D(FBBasis2D):
 
         :param x: The Image instance representing coefficient array in the
             standard 2D coordinate basis to be evaluated.
-        :return v: The evaluation of the coefficient array `v` in the FB basis.
+        :return: The evaluation of the coefficient array `x` in the FB basis.
             This is an array of vectors whose last dimension equals `self.count`
-            and whose first dimension correspond to `x.n_images`.
+            and whose first dimension correspond to `x.shape[0]`.
         """
         # get information on polar grids from precomputed data
         n_theta = np.size(self._precomp["freqs"], 2)
@@ -197,58 +187,112 @@ class FFBBasis2D(FBBasis2D):
         freqs = np.reshape(self._precomp["freqs"], (2, n_r * n_theta))
 
         # number of 2D image samples
-        n_images = x.n_images
-        x_data = x.data
+        n_images = x.shape[0]
 
         # resamping x in a polar Fourier gird using nonuniform discrete Fourier transform
-        pf = nufft(x_data, 2 * pi * freqs)
-        pf = np.reshape(pf, (n_images, n_r, n_theta))
+        pf = nufft(xp.asarray(x), 2 * pi * freqs)
+        pf = pf.reshape(n_images, n_r, n_theta)
 
         # Recover "negative" frequencies from "positive" half plane.
-        pf = np.concatenate((pf, pf.conjugate()), axis=2)
+        pf = xp.concatenate((pf, pf.conjugate()), axis=2)
 
         # evaluate radial integral using the Gauss-Legendre quadrature rule
-        for i_r in range(0, n_r):
-            pf[:, i_r, :] = pf[:, i_r, :] * (
-                self._precomp["gl_weights"][i_r] * self._precomp["gl_nodes"][i_r]
-            )
+        pf = pf * self.gl_weighted_nodes[None, :, None]
 
         #  1D FFT on the angular dimension for each concentric circle
-        pf = 2 * pi / (2 * n_theta) * xp.asnumpy(fft.fft(xp.asarray(pf)))
+        pf = 2 * xp.pi / (2 * n_theta) * fft.fft(pf)
 
         # This only makes it easier to slice the array later.
-        v = np.zeros((n_images, self.count), dtype=x.dtype)
+        v = xp.zeros((n_images, self.count), dtype=x.dtype)
 
         # go through each basis function and find the corresponding coefficient
         ind = 0
-        idx = ind + np.arange(self.k_max[0])
-        mask = self._indices["ells"] == 0
+        idx = ind + xp.arange(self.k_max[0])
 
-        # include the normalization factor of angular part into radial part
-        radial_norm = self._precomp["radial"] / np.expand_dims(self.angular_norms, 1)
-        v[:, mask] = pf[:, :, 0].real @ radial_norm[idx].T
-        ind = ind + np.size(idx)
+        v[:, self._zero_angular_inds] = pf[:, :, 0].real @ self.radial_norm[idx].T
+        ind = ind + idx.size
 
         ind_pos = ind
         for ell in range(1, self.ell_max + 1):
-            idx = ind + np.arange(self.k_max[ell])
-            idx_pos = ind_pos + np.arange(self.k_max[ell])
+            idx = ind + xp.arange(self.k_max[ell])
+            idx_pos = ind_pos + xp.arange(self.k_max[ell])
             idx_neg = idx_pos + self.k_max[ell]
 
-            v_ell = pf[:, :, ell] @ radial_norm[idx].T
+            v_ell = pf[:, :, ell] @ self.radial_norm[idx].T
 
             if np.mod(ell, 2) == 0:
-                v_pos = np.real(v_ell)
-                v_neg = -np.imag(v_ell)
+                v_pos = v_ell.real
+                v_neg = -v_ell.imag
             else:
-                v_pos = np.imag(v_ell)
-                v_neg = np.real(v_ell)
+                v_pos = v_ell.imag
+                v_neg = v_ell.real
 
             v[:, idx_pos] = v_pos
             v[:, idx_neg] = v_neg
 
-            ind = ind + np.size(idx)
+            ind = ind + idx.size
 
             ind_pos = ind_pos + 2 * self.k_max[ell]
 
-        return v
+        return xp.asnumpy(v)
+
+    def filter_to_basis_mat(self, f, **kwargs):
+        """
+        See `SteerableBasis2D.filter_to_basis_mat`.
+        """
+        # Note 'method' and 'truncate' not relevant for this optimized FFB code.
+        if kwargs.get("method", None) is not None:
+            raise NotImplementedError(
+                "`FFBBasis2D.filter_to_basis_mat` method {method} not supported."
+                "  Use `method=None`."
+            )
+
+        pixel_size = kwargs.get("pixel_size", None)
+
+        # These form a circular dependence, import locally until time to clean up.
+        from aspire.basis.basis_utils import lgwt
+
+        # Get the filter's evaluate function.
+        h_fun = f.evaluate
+
+        # Set same dimensions as basis object
+        n_k = self.n_r
+        n_theta = self.n_theta
+        radial = self._precomp["radial"]
+
+        # get 2D grid in polar coordinate
+        k_vals, wts = lgwt(n_k, 0, 0.5, dtype=self.dtype)
+        k, theta = np.meshgrid(
+            k_vals, np.arange(n_theta) * 2 * np.pi / (2 * n_theta), indexing="ij"
+        )
+
+        # Get function values in polar 2D grid and average out angle contribution
+        omegax = k * np.cos(theta)
+        omegay = k * np.sin(theta)
+        omega = 2 * np.pi * np.vstack((omegax.flatten("C"), omegay.flatten("C")))
+
+        h_vals2d = (
+            h_fun(omega, pixel_size=pixel_size).reshape(n_k, n_theta).astype(self.dtype)
+        )
+        h_vals = np.sum(h_vals2d, axis=1) / n_theta
+
+        # Represent 1D function values in basis
+        h_basis = BlkDiagMatrix.empty(2 * self.ell_max + 1, dtype=self.dtype)
+        ind_ell = 0
+        for ell in range(0, self.ell_max + 1):
+            k_max = self.k_max[ell]
+            rmat = 2 * k_vals.reshape(n_k, 1) * self.r0[ell][0:k_max].T
+            basis_vals = np.zeros_like(rmat)
+            ind_radial = np.sum(self.k_max[0:ell])
+            basis_vals[:, 0:k_max] = radial[ind_radial : ind_radial + k_max].T
+            h_basis_vals = basis_vals * h_vals.reshape(n_k, 1)
+            h_basis_ell = basis_vals.T @ (
+                h_basis_vals * k_vals.reshape(n_k, 1) * wts.reshape(n_k, 1)
+            )
+            h_basis[ind_ell] = h_basis_ell
+            ind_ell += 1
+            if ell > 0:
+                h_basis[ind_ell] = h_basis[ind_ell - 1]
+                ind_ell += 1
+
+        return h_basis
