@@ -1,12 +1,11 @@
 import logging
 
 import numpy as np
-from numpy import pi
 
 from aspire.basis import FBBasis3D
 from aspire.basis.basis_utils import lgwt, norm_assoc_legendre, sph_bessel
 from aspire.nufft import anufft, nufft
-from aspire.utils.matlab_compat import m_flatten, m_reshape
+from aspire.numeric import xp
 
 logger = logging.getLogger(__name__)
 
@@ -60,26 +59,29 @@ class FFBBasis3D(FBBasis3D):
 
         r, wt_r = lgwt(n_r, 0.0, self.kcut, dtype=self.dtype)
         z, wt_z = lgwt(n_phi, -1, 1, dtype=self.dtype)
-        r = m_reshape(r, (n_r, 1))
-        wt_r = m_reshape(wt_r, (n_r, 1))
-        z = m_reshape(z, (n_phi, 1))
-        wt_z = m_reshape(wt_z, (n_phi, 1))
-        phi = np.arccos(z)
+        r = xp.asarray(r).reshape(n_r, 1)
+        rh = xp.asnumpy(r)
+        wt_r = xp.asarray(wt_r).reshape(n_r, 1)
+        z = xp.asarray(z).reshape(n_phi, 1)
+        wt_z = xp.asarray(wt_z).reshape(n_phi, 1)
+        phi = xp.arccos(z)
         wt_phi = wt_z
-        theta = 2 * pi * np.arange(n_theta, dtype=self.dtype).T / (2 * n_theta)
-        theta = m_reshape(theta, (n_theta, 1))
+        theta = 2 * xp.pi * xp.arange(n_theta, dtype=self.dtype).T / (2 * n_theta)
+        theta = theta.reshape(n_theta, 1)
 
         # evaluate basis function in the radial dimension
-        radial_wtd = np.zeros(
+        radial_wtd = xp.zeros(
             shape=(n_r, np.max(self.k_max), self.ell_max + 1), dtype=self.dtype
         )
         for ell in range(0, self.ell_max + 1):
             k_max_ell = self.k_max[ell]
-            rmat = r * self.r0[0:k_max_ell, ell].T / self.kcut
-            radial_ell = np.zeros_like(rmat)
+            rmat = rh * self.r0[ell][0:k_max_ell].T / self.kcut  # host
+            radial_ell = xp.zeros_like(rmat)
             for ik in range(0, k_max_ell):
-                radial_ell[:, ik] = sph_bessel(ell, rmat[:, ik])
-            nrm = np.abs(sph_bessel(ell + 1, self.r0[0:k_max_ell, ell].T) / 4)
+                radial_ell[:, ik] = xp.asarray(sph_bessel(ell, rmat[:, ik]))
+            nrm = xp.abs(
+                xp.asarray(sph_bessel(ell + 1, self.r0[ell][0:k_max_ell].T)) / 4
+            )
             radial_ell = radial_ell / nrm
             radial_ell_wtd = r**2 * wt_r * radial_ell
             radial_wtd[:, 0:k_max_ell, ell] = radial_ell_wtd
@@ -94,14 +96,14 @@ class FFBBasis3D(FBBasis3D):
                 - np.mod(self.ell_max, 2) * np.mod(m, 2)
             )
             n_odd_ell = int(self.ell_max - m + 1 - n_even_ell)
-            phi_wtd_m_even = np.zeros((n_phi, n_even_ell), dtype=phi.dtype)
-            phi_wtd_m_odd = np.zeros((n_phi, n_odd_ell), dtype=phi.dtype)
+            phi_wtd_m_even = xp.zeros((n_phi, n_even_ell), dtype=phi.dtype)
+            phi_wtd_m_odd = xp.zeros((n_phi, n_odd_ell), dtype=phi.dtype)
 
             ind_even = 0
             ind_odd = 0
             for ell in range(m, self.ell_max + 1):
-                phi_m_ell = norm_assoc_legendre(ell, m, z)
-                nrm_inv = np.sqrt(0.5 / pi)
+                phi_m_ell = xp.asarray(norm_assoc_legendre(ell, m, z))
+                nrm_inv = np.sqrt(0.5 / np.pi)
                 phi_m_ell = nrm_inv * phi_m_ell
                 phi_wtd_m_ell = wt_phi * phi_m_ell
                 if np.mod(ell, 2) == 0:
@@ -115,32 +117,32 @@ class FFBBasis3D(FBBasis3D):
             ang_phi_wtd_odd.append(phi_wtd_m_odd)
 
         # evaluate basis function in the theta dimension
-        ang_theta = np.zeros((n_theta, 2 * self.ell_max + 1), dtype=theta.dtype)
-
-        ang_theta[:, 0 : self.ell_max] = np.sqrt(2) * np.sin(
-            theta @ m_reshape(np.arange(self.ell_max, 0, -1), (1, self.ell_max))
+        ang_theta = xp.zeros((n_theta, 2 * self.ell_max + 1), dtype=theta.dtype)
+        ang_theta[:, 0 : self.ell_max] = np.sqrt(2) * xp.sin(
+            theta @ xp.arange(self.ell_max, 0, -1).reshape(1, self.ell_max)
         )
-        ang_theta[:, self.ell_max] = np.ones(n_theta, dtype=theta.dtype)
-        ang_theta[:, self.ell_max + 1 : 2 * self.ell_max + 1] = np.sqrt(2) * np.cos(
-            theta @ m_reshape(np.arange(1, self.ell_max + 1), (1, self.ell_max))
+        ang_theta[:, self.ell_max] = xp.ones(n_theta, dtype=theta.dtype)
+        ang_theta[:, self.ell_max + 1 : 2 * self.ell_max + 1] = np.sqrt(2) * xp.cos(
+            theta @ xp.arange(1, self.ell_max + 1).reshape(1, self.ell_max)
         )
 
-        ang_theta_wtd = (2 * pi / n_theta) * ang_theta
+        ang_theta_wtd = (2 * np.pi / n_theta) * ang_theta
 
-        theta_grid, phi_grid, r_grid = np.meshgrid(
-            theta, phi, r, sparse=False, indexing="ij"
+        r_grid, phi_grid, theta_grid = xp.meshgrid(
+            r.flatten(), phi.flatten(), theta.flatten(), sparse=False, indexing="ij"
         )
-        fourier_x = m_flatten(r_grid * np.cos(theta_grid) * np.sin(phi_grid))
-        fourier_y = m_flatten(r_grid * np.sin(theta_grid) * np.sin(phi_grid))
-        fourier_z = m_flatten(r_grid * np.cos(phi_grid))
+
+        fourier_x = (r_grid * xp.cos(theta_grid) * xp.sin(phi_grid)).flatten()
+        fourier_y = (r_grid * xp.sin(theta_grid) * xp.sin(phi_grid)).flatten()
+        fourier_z = (r_grid * xp.cos(phi_grid)).flatten()
         fourier_pts = (
             2
-            * pi
-            * np.vstack(
+            * xp.pi
+            * xp.vstack(
                 (
-                    fourier_z[np.newaxis, ...],
-                    fourier_y[np.newaxis, ...],
-                    fourier_x[np.newaxis, ...],
+                    fourier_z[None, ...],
+                    fourier_y[None, ...],
+                    fourier_x[None, ...],
                 )
             )
         )
@@ -163,9 +165,7 @@ class FFBBasis3D(FBBasis3D):
             coordinate basis. This is an array whose last three dimensions equal
             `self.sz` and the remaining dimensions correspond to `v`.
         """
-        # roll dimensions of v
-        sz_roll = v.shape[:-1]
-        v = v.reshape((-1, self.count))
+        v = xp.asarray(v)
 
         # get information on polar grids from precomputed data
         n_theta = np.size(self._precomp["ang_theta_wtd"], 0)
@@ -175,17 +175,17 @@ class FFBBasis3D(FBBasis3D):
         # number of 3D image samples
         n_data = v.shape[0]
 
-        u_even = np.zeros(
+        u_even = xp.zeros(
             (
-                n_r,
-                int(2 * self.ell_max + 1),
-                n_data,
                 int(np.floor(self.ell_max / 2) + 1),
+                n_data,
+                int(2 * self.ell_max + 1),
+                n_r,
             ),
             dtype=v.dtype,
         )
-        u_odd = np.zeros(
-            (n_r, int(2 * self.ell_max + 1), n_data, int(np.ceil(self.ell_max / 2))),
+        u_odd = xp.zeros(
+            (int(np.ceil(self.ell_max / 2)), n_data, int(2 * self.ell_max + 1), n_r),
             dtype=v.dtype,
         )
 
@@ -196,30 +196,27 @@ class FFBBasis3D(FBBasis3D):
             radial_wtd = self._precomp["radial_wtd"][:, 0:k_max_ell, ell]
 
             ind = self._indices["ells"] == ell
-
-            v_ell = m_reshape(v[:, ind].T, (k_max_ell, (2 * ell + 1) * n_data))
-            v_ell = radial_wtd @ v_ell
-            v_ell = m_reshape(v_ell, (n_r, 2 * ell + 1, n_data))
+            v_ell = v[:, ind].reshape((2 * ell + 1) * n_data, k_max_ell)
+            v_ell = v_ell @ radial_wtd.T
+            v_ell = v_ell.reshape(n_data, 2 * ell + 1, n_r)
 
             if np.mod(ell, 2) == 0:
                 u_even[
+                    int(ell / 2),
                     :,
                     int(self.ell_max - ell) : int(self.ell_max + ell + 1),
                     :,
-                    int(ell / 2),
                 ] = v_ell
             else:
                 u_odd[
+                    int((ell - 1) / 2),
                     :,
                     int(self.ell_max - ell) : int(self.ell_max + ell + 1),
                     :,
-                    int((ell - 1) / 2),
                 ] = v_ell
 
-        u_even = np.transpose(u_even, (3, 0, 1, 2))
-        u_odd = np.transpose(u_odd, (3, 0, 1, 2))
-        w_even = np.zeros((n_phi, n_r, n_data, 2 * self.ell_max + 1), dtype=v.dtype)
-        w_odd = np.zeros((n_phi, n_r, n_data, 2 * self.ell_max + 1), dtype=v.dtype)
+        w_even = xp.zeros((2 * self.ell_max + 1, n_phi, n_data, n_r), dtype=v.dtype)
+        w_odd = xp.zeros((2 * self.ell_max + 1, n_phi, n_data, n_r), dtype=v.dtype)
 
         # evaluate the phi parts
         for m in range(0, self.ell_max + 1):
@@ -235,48 +232,42 @@ class FFBBasis3D(FBBasis3D):
                 sgns = (1, -1)
 
             for sgn in sgns:
-
                 end = np.size(u_even, 0)
                 u_m_even = u_even[end - n_even_ell : end, :, self.ell_max + sgn * m, :]
                 end = np.size(u_odd, 0)
                 u_m_odd = u_odd[end - n_odd_ell : end, :, self.ell_max + sgn * m, :]
 
-                u_m_even = m_reshape(u_m_even, (n_even_ell, n_r * n_data))
-                u_m_odd = m_reshape(u_m_odd, (n_odd_ell, n_r * n_data))
+                u_m_even = u_m_even.reshape(n_even_ell, n_data * n_r)
+                u_m_odd = u_m_odd.reshape(n_odd_ell, n_data * n_r)
 
                 w_m_even = ang_phi_wtd_m_even @ u_m_even
                 w_m_odd = ang_phi_wtd_m_odd @ u_m_odd
 
-                w_m_even = m_reshape(w_m_even, (n_phi, n_r, n_data))
-                w_m_odd = m_reshape(w_m_odd, (n_phi, n_r, n_data))
+                w_m_even = w_m_even.reshape(n_phi, n_data, n_r)
+                w_m_odd = w_m_odd.reshape(n_phi, n_data, n_r)
 
-                w_even[:, :, :, self.ell_max + sgn * m] = w_m_even
-                w_odd[:, :, :, self.ell_max + sgn * m] = w_m_odd
+                w_even[self.ell_max + sgn * m] = w_m_even
+                w_odd[self.ell_max + sgn * m] = w_m_odd
 
-        w_even = np.transpose(w_even, (3, 0, 1, 2))
-        w_odd = np.transpose(w_odd, (3, 0, 1, 2))
+        w_even = w_even.transpose((2, 3, 1, 0))
+        w_odd = w_odd.transpose((2, 3, 1, 0))
         u_even = w_even
         u_odd = w_odd
 
-        u_even = m_reshape(u_even, (2 * self.ell_max + 1, n_phi * n_r * n_data))
-        u_odd = m_reshape(u_odd, (2 * self.ell_max + 1, n_phi * n_r * n_data))
+        u_even = u_even.reshape(n_data * n_r * n_phi, 2 * self.ell_max + 1)
+        u_odd = u_odd.reshape(n_data * n_r * n_phi, 2 * self.ell_max + 1)
 
         # evaluate the theta parts
-        w_even = self._precomp["ang_theta_wtd"] @ u_even
-        w_odd = self._precomp["ang_theta_wtd"] @ u_odd
+        w_even = u_even @ self._precomp["ang_theta_wtd"].T
+        w_odd = u_odd @ self._precomp["ang_theta_wtd"].T
 
         pf = w_even + 1j * w_odd
-        pf = m_reshape(pf, (n_theta * n_phi * n_r, n_data))
-        pf = np.moveaxis(pf, 0, -1)
+        pf = pf.reshape(n_data, n_r * n_phi * n_theta)
 
         # perform inverse non-uniformly FFT transformation back to 3D rectangular coordinates
-        freqs = m_reshape(self._precomp["fourier_pts"], (3, n_r * n_theta * n_phi))
-        x = anufft(pf, freqs, self.sz, real=True)
+        x = anufft(pf, self._precomp["fourier_pts"], self.sz, real=True)
 
-        # Roll, return the x with the last three dimensions as self.sz
-        # Higher dimensions should be like v.
-        x = x.reshape((*sz_roll, *self.sz))
-        return x
+        return xp.asnumpy(x)
 
     def _evaluate_t(self, x):
         """
@@ -284,47 +275,42 @@ class FFBBasis3D(FBBasis3D):
 
         :param x: The coefficient array in the standard 3D coordinate basis
             to be evaluated. The last three dimensions must equal `self.sz`.
-        :return v: The evaluation of the coefficient array `v` in the FB basis.
+        :return: The evaluation of the coefficient array `x` in the FB basis.
             This is an array of vectors whose last dimension equals
             `self.count` and whose remaining dimensions correspond to higher
             dimensions of `x`.
         """
-        x = x.asnumpy()
-
-        # roll dimensions
-        sz_roll = x.shape[:-3]
-        x = x.reshape((-1, *self.sz))
+        x = xp.asarray(x)
 
         n_data = x.shape[0]
         n_r = np.size(self._precomp["radial_wtd"], 0)
         n_phi = np.size(self._precomp["ang_phi_wtd_even"][0], 0)
         n_theta = np.size(self._precomp["ang_theta_wtd"], 0)
 
-        # resamping x in a polar Fourier gird using nonuniform discrete Fourier transform
+        # resamping x in a polar Fourier grid using nonuniform discrete Fourier transform
         pf = nufft(x, self._precomp["fourier_pts"])
-
-        pf = m_reshape(pf.T, (n_theta, n_phi * n_r * n_data))
+        pf = pf.reshape(n_data * n_r * n_phi, n_theta)
 
         # evaluate the theta parts
-        u_even = self._precomp["ang_theta_wtd"].T @ np.real(pf)
-        u_odd = self._precomp["ang_theta_wtd"].T @ np.imag(pf)
+        u_even = pf.real @ self._precomp["ang_theta_wtd"]
+        u_odd = pf.imag @ self._precomp["ang_theta_wtd"]
 
-        u_even = m_reshape(u_even, (2 * self.ell_max + 1, n_phi, n_r, n_data))
-        u_odd = m_reshape(u_odd, (2 * self.ell_max + 1, n_phi, n_r, n_data))
+        u_even = u_even.reshape(n_data, n_r, n_phi, 2 * self.ell_max + 1)
+        u_odd = u_odd.reshape(n_data, n_r, n_phi, 2 * self.ell_max + 1)
 
-        u_even = np.transpose(u_even, (1, 2, 3, 0))
-        u_odd = np.transpose(u_odd, (1, 2, 3, 0))
-
-        w_even = np.zeros(
-            (int(np.floor(self.ell_max / 2) + 1), n_r, 2 * self.ell_max + 1, n_data),
+        w_even = xp.zeros(
+            (n_data, 2 * self.ell_max + 1, n_r, int(np.floor(self.ell_max / 2) + 1)),
             dtype=x.dtype,
         )
-        w_odd = np.zeros(
-            (int(np.ceil(self.ell_max / 2)), n_r, 2 * self.ell_max + 1, n_data),
+        w_odd = xp.zeros(
+            (n_data, 2 * self.ell_max + 1, n_r, int(np.ceil(self.ell_max / 2))),
             dtype=x.dtype,
         )
 
-        # evaluate the phi parts
+        # Transpose and copy as contiguous for faster slicing and matrix multiplication.
+        u_even = xp.ascontiguousarray(u_even.transpose(3, 0, 1, 2))
+        u_odd = xp.ascontiguousarray(u_odd.transpose(3, 0, 1, 2))
+
         for m in range(0, self.ell_max + 1):
             ang_phi_wtd_m_even = self._precomp["ang_phi_wtd_even"][m]
             ang_phi_wtd_m_odd = self._precomp["ang_phi_wtd_odd"][m]
@@ -338,27 +324,25 @@ class FFBBasis3D(FBBasis3D):
                 sgns = (1, -1)
 
             for sgn in sgns:
-                u_m_even = u_even[:, :, :, self.ell_max + sgn * m]
-                u_m_odd = u_odd[:, :, :, self.ell_max + sgn * m]
+                u_m_even = u_even[self.ell_max + sgn * m]
+                u_m_odd = u_odd[self.ell_max + sgn * m]
 
-                u_m_even = m_reshape(u_m_even, (n_phi, n_r * n_data))
-                u_m_odd = m_reshape(u_m_odd, (n_phi, n_r * n_data))
+                u_m_even = u_m_even.reshape(n_data * n_r, n_phi)
+                u_m_odd = u_m_odd.reshape(n_data * n_r, n_phi)
 
-                w_m_even = ang_phi_wtd_m_even.T @ u_m_even
-                w_m_odd = ang_phi_wtd_m_odd.T @ u_m_odd
+                w_m_even = u_m_even @ ang_phi_wtd_m_even
+                w_m_odd = u_m_odd @ ang_phi_wtd_m_odd
 
-                w_m_even = m_reshape(w_m_even, (n_even_ell, n_r, n_data))
-                w_m_odd = m_reshape(w_m_odd, (n_odd_ell, n_r, n_data))
-                end = np.size(w_even, 0)
-                w_even[end - n_even_ell : end, :, self.ell_max + sgn * m, :] = w_m_even
-                end = np.size(w_odd, 0)
-                w_odd[end - n_odd_ell : end, :, self.ell_max + sgn * m, :] = w_m_odd
+                w_m_even = w_m_even.reshape(n_data, n_r, n_even_ell)
+                w_m_odd = w_m_odd.reshape(n_data, n_r, n_odd_ell)
 
-        w_even = np.transpose(w_even, (1, 2, 3, 0))
-        w_odd = np.transpose(w_odd, (1, 2, 3, 0))
+                end = np.size(w_even, -1)
+                w_even[:, self.ell_max + sgn * m, :, end - n_even_ell : end] = w_m_even
+                end = np.size(w_odd, -1)
+                w_odd[:, self.ell_max + sgn * m, :, end - n_odd_ell : end] = w_m_odd
 
         # evaluate the radial parts
-        v = np.zeros((n_data, self.count), dtype=x.dtype)
+        v = xp.zeros((n_data, self.count), dtype=x.dtype)
         for ell in range(0, self.ell_max + 1):
             k_max_ell = self.k_max[ell]
             radial_wtd = self._precomp["radial_wtd"][:, 0:k_max_ell, ell]
@@ -378,17 +362,12 @@ class FFBBasis3D(FBBasis3D):
                     int((ell - 1) / 2),
                 ]
 
-            v_ell = m_reshape(v_ell, (n_r, (2 * ell + 1) * n_data))
-
-            v_ell = radial_wtd.T @ v_ell
-
-            v_ell = m_reshape(v_ell, (k_max_ell * (2 * ell + 1), n_data))
+            v_ell = v_ell.reshape(n_data * (2 * ell + 1), n_r)
+            v_ell = v_ell @ radial_wtd
+            v_ell = v_ell.reshape(n_data, (2 * ell + 1) * k_max_ell)
 
             # TODO: Fix this to avoid lookup each time.
             ind = self._indices["ells"] == ell
-            v[:, ind] = v_ell.T
+            v[:, ind] = v_ell
 
-        # Roll dimensions, last dimension should be self.count,
-        # Higher dimensions like x.
-        v = v.reshape((*sz_roll, self.count))
-        return v
+        return xp.asnumpy(v)
