@@ -3,8 +3,7 @@ from collections import OrderedDict
 
 import numpy as np
 
-from aspire.basis import FFBBasis2D, SteerableBasis2D
-from aspire.covariance import BatchedRotCov2D
+from aspire.basis import Coef, ComplexCoef, FFBBasis2D, SteerableBasis2D
 from aspire.operators import BlkDiagMatrix
 from aspire.utils import complex_type, fix_signs, real_type
 
@@ -42,7 +41,7 @@ class FSPCABasis(SteerableBasis2D):
             Default value of `None` will use `self.basis.count`.
         :param noise_var: Optionally assign noise variance.
             Default value of `None` will estimate noise with WhiteNoiseEstimator.
-            Use 0 when using clean images so cov2d skips applying noisy covar coeffs..
+            Use 0 when using clean images so cov2d skips applying noisy covar coefs..
         :param batch_size: Batch size for computing basis coefficients.
             `batch_size` is also passed to BatchedRotCov2D.
         """
@@ -71,7 +70,7 @@ class FSPCABasis(SteerableBasis2D):
         self.complex_count = self.basis.complex_count
         self.angular_indices = self.basis.angular_indices
         self.radial_indices = self.basis.radial_indices
-        self.signs_indices = self.basis._indices["sgns"]
+        self.signs_indices = self.basis.signs_indices
         self.complex_angular_indices = self.basis.complex_angular_indices
         self.complex_radial_indices = self.basis.complex_radial_indices
 
@@ -80,6 +79,11 @@ class FSPCABasis(SteerableBasis2D):
             len(self.complex_indices_map) == self.complex_count
         ), f"{len(self.complex_indices_map)} != {self.complex_count}"
 
+        # Map Cached SteerableBasis2D index maps
+        self._zero_angular_inds = self.basis._zero_angular_inds
+        self._pos_angular_inds = self.basis._pos_angular_inds
+        self._neg_angular_inds = self.basis._neg_angular_inds
+
         self.noise_var = noise_var  # noise_var is handled during `build` call.
 
         self.build()
@@ -142,6 +146,9 @@ class FSPCABasis(SteerableBasis2D):
             self.noise_var = WhiteNoiseEstimator(self.src).estimate()
         logger.info(f"Setting noise_var={self.noise_var}")
 
+        # Import BatchedRotCov2D here to prevent circular imports.
+        from aspire.covariance import BatchedRotCov2D
+
         cov2d = BatchedRotCov2D(
             src=self.src, basis=self.basis, batch_size=self.batch_size
         )
@@ -157,13 +164,13 @@ class FSPCABasis(SteerableBasis2D):
         }
         self.mean_coef_est = cov2d.get_mean()
         self.covar_coef_est = cov2d.get_covar(
-            mean_coeff=self.mean_coef_est,
+            mean_coef=self.mean_coef_est,
             noise_var=self.noise_var,
             covar_est_opt=covar_opt,
         )
 
         # Create the arrays to be packed by _compute_spca
-        self.eigvals = np.zeros(self.basis.count, dtype=self.dtype)
+        self._eigvals = np.zeros(self.basis.count, dtype=self.dtype)
 
         self.eigvecs = BlkDiagMatrix.empty(2 * self.basis.ell_max + 1, dtype=self.dtype)
 
@@ -188,7 +195,6 @@ class FSPCABasis(SteerableBasis2D):
         eigval_index = 0
         basis_inds = []
         for angular_index, C_k in enumerate(self.covar_coef_est):
-
             # # Eigen/SVD, covariance block C_k should be symmetric.
             eigvals_k, eigvecs_k = np.linalg.eigh(C_k)
 
@@ -207,7 +213,7 @@ class FSPCABasis(SteerableBasis2D):
             basis_inds.append(_basis_inds)
 
             # Store the eigvals for this block, note this is a flat array.
-            self.eigvals[_basis_inds] = eigvals_k
+            self._eigvals[_basis_inds] = eigvals_k
 
             # Store the eigvecs, note this is a BlkDiagMatrix and is assigned incrementally.
             self.eigvecs[angular_index] = eigvecs_k
@@ -227,7 +233,7 @@ class FSPCABasis(SteerableBasis2D):
         #
         # We can pass a full or truncated slice of sorted_indices to any array indexed by
         # the coefs.  This is used later for compression and index re-generation.
-        self.sorted_indices = np.argsort(-np.abs(self.eigvals))
+        self.sorted_indices = np.argsort(-np.abs(self._eigvals))
 
         compressed_indices = self._get_compressed_indices()
 
@@ -236,19 +242,19 @@ class FSPCABasis(SteerableBasis2D):
         )
 
         # Compute coefficient vector of mean image at zeroth component
-        self.mean_coef_zero = self.mean_coef_est[self.angular_indices == 0]
+        self.mean_coef_zero = self.mean_coef_est.asnumpy()[0][self.angular_indices == 0]
 
         # Define mask for zero angular mode, used in loop below
-        zero_ell_mask = self.basis._indices["ells"] == 0
+        zero_ell_mask = self.basis.angular_indices == 0
 
         # Apply Data matrix batchwise
         num_batches = (self.src.n + self.batch_size - 1) // self.batch_size
         for i in range(num_batches):
-
             # Compute the coefficients for this batch
             start = i * self.batch_size
             finish = min((i + 1) * self.batch_size, self.src.n)
             batch_coef = self.basis.evaluate_t(self.src.images[start:finish])
+            batch_coef = batch_coef.asnumpy()
 
             # Make the Data matrix (A_k)
             # # Construct A_k, matrix of expansion coefficients a^i_k_q
@@ -271,9 +277,9 @@ class FSPCABasis(SteerableBasis2D):
             for ell in range(
                 1, self.basis.ell_max + 1
             ):  # `ell` in this code is `k` from paper
-                mask_ell = self.basis._indices["ells"] == ell
-                mask_pos = mask_ell & (self.basis._indices["sgns"] == +1)
-                mask_neg = mask_ell & (self.basis._indices["sgns"] == -1)
+                mask_ell = self.basis.angular_indices == ell
+                mask_pos = mask_ell & (self.basis.signs_indices == +1)
+                mask_neg = mask_ell & (self.basis.signs_indices == -1)
 
                 A.append(batch_coef[:, mask_pos])
                 A.append(batch_coef[:, mask_neg])
@@ -291,7 +297,6 @@ class FSPCABasis(SteerableBasis2D):
             #     on the correspong block.
             blk_spca_coef = np.empty_like(batch_coef)
             for angular_index, a_blk in enumerate(A):
-
                 # To compute new expansion coefficients using spca basis
                 #   we combine the basis coefs using the eigen decomposition.
                 # Note image stack slow moving axis, otherwise this is just a
@@ -328,13 +333,15 @@ class FSPCABasis(SteerableBasis2D):
             Fourier Bessel basis.
         :return: Stack of coefs in the FSPCABasis.
         """
+        if not isinstance(x, Coef):
+            raise TypeError(f"'x' should be `Coef` instance, received {type(x)}.")
 
         # Apply linear combination defined by FSPCA (eigvecs)
-        c_fspca = x @ self.eigvecs
+        c_fspca = x.asnumpy() @ self.eigvecs
 
         assert c_fspca.shape == (x.shape[0], self.count)
 
-        return c_fspca
+        return Coef(self, c_fspca, pixel_size=x.pixel_size)
 
     def evaluate_to_image_basis(self, c):
         """
@@ -344,6 +351,9 @@ class FSPCABasis(SteerableBasis2D):
         :return: The Image instance representing a stack of images in the
             standard 2D coordinate basis..
         """
+        if not isinstance(c, Coef):
+            raise TypeError(f"'c' should be `Coef` instance, received {type(c)}.")
+
         c_fb = self.evaluate(c)
 
         return self.basis.evaluate(c_fb)
@@ -356,6 +366,9 @@ class FSPCABasis(SteerableBasis2D):
         :return: The (real) coefs representing a stack of images in self.basis
         """
 
+        if not isinstance(c, Coef):
+            raise TypeError(f"'c' should be `Coef` instance, received {type(c)}.")
+
         # apply FSPCA eigenvector to coefs c, yields coefs in self.basis
         eigvecs = self.eigvecs
         if isinstance(eigvecs, BlkDiagMatrix):
@@ -366,7 +379,7 @@ class FSPCABasis(SteerableBasis2D):
         # corrected_c[:, self.angular_indices!=0] *= 2
         # return corrected_c @ eigvecs.T
 
-        return c @ eigvecs.T
+        return Coef(self.basis, c @ eigvecs.T, pixel_size=c.pixel_size)
 
     # TODO: Python>=3.8 @cached_property
     def _get_compressed_indices(self):
@@ -387,17 +400,17 @@ class FSPCABasis(SteerableBasis2D):
         #    instead of the larger component.
         # Note, that we only use OrderedDict for its key's (ie as an OrderedSet)
         ordered_components = OrderedDict()
-        for (k, q) in unsigned_components:
+        for k, q in unsigned_components:
             ordered_components.setdefault((k, q))  # inserts when not exists yet
 
         # Select the top n (k,q) pairs
         top_components = list(ordered_components)[: self.components]
 
         # Now we need to find the locations of both the + and - sgns.
-        pos_mask = self.basis._indices["sgns"] == 1
-        neg_mask = self.basis._indices["sgns"] == -1
+        pos_mask = self.basis.signs_indices == 1
+        neg_mask = self.basis.signs_indices == -1
         compressed_indices = []
-        for (k, q) in top_components:
+        for k, q in top_components:
             # Compute the locations of coefs we're interested in.
             k_maps = self.angular_indices == k
             q_maps = self.radial_indices == q
@@ -433,7 +446,7 @@ class FSPCABasis(SteerableBasis2D):
         compressed_indices = self._get_compressed_indices()
         self.count = len(compressed_indices)
 
-        self.eigvals = self.eigvals[compressed_indices]
+        self._eigvals = self._eigvals[compressed_indices]
         if isinstance(self.eigvecs, BlkDiagMatrix):
             self.eigvecs = self.eigvecs.dense()
         self.eigvecs = self.eigvecs[:, compressed_indices]
@@ -460,11 +473,12 @@ class FSPCABasis(SteerableBasis2D):
         There is a corresponding method, to_real.
 
         :param coef: Coefficients from this basis.
-        :return: Complex coefficent representation from this basis.
+        :return: Complex coeficent representation from this basis.
         """
-
-        if coef.ndim == 1:
-            coef = coef.reshape(1, -1)
+        if not isinstance(coef, Coef):
+            raise TypeError(f"'coef' should be `Coef` instance, received {type(coef)}.")
+        px_sz = coef.pixel_size
+        coef = coef.asnumpy()
 
         if coef.dtype not in (np.float64, np.float32):
             raise TypeError("coef provided to to_complex should be real.")
@@ -501,7 +515,7 @@ class FSPCABasis(SteerableBasis2D):
         for i, k in enumerate(ccoef_d.keys()):
             ccoef[:, i] = ccoef_d[k]
 
-        return ccoef
+        return ComplexCoef(self, ccoef, pixel_size=px_sz)
 
     def to_real(self, complex_coef):
         """
@@ -512,7 +526,7 @@ class FSPCABasis(SteerableBasis2D):
         There is a corresponding method, to_complex.
 
         :param complex_coef: Complex coefficients from this basis.
-        :return: Real coefficent representation from this basis.
+        :return: Real coefficient representation from this basis.
         """
 
         if complex_coef.ndim == 1:
@@ -541,21 +555,7 @@ class FSPCABasis(SteerableBasis2D):
                 coef[:, pos_i] = 2.0 * complex_coef[:, i].real
                 coef[:, neg_i] = -2.0 * complex_coef[:, i].imag
 
-        return coef
-
-    def rotate(self, coef, radians, refl=None):
-        """
-        Returns coefs rotated by `radians`.
-
-        :param coef: Basis coefs.
-        :param radians: Rotation in radians.
-        :param refl: Optional reflect image (bool)
-        :return: rotated coefs.
-        """
-
-        # Sterrable class rotation expects complex representation of coefficients.
-        #  Convert, rotate and convert back to real representation.
-        return self.to_real(super().rotate(self.to_complex(coef), radians, refl))
+        return Coef(self, coef, pixel_size=complex_coef.pixel_size)
 
     def calculate_bispectrum(
         self, coef, flatten=False, filter_nonzero_freqs=False, freq_cutoff=None
@@ -569,6 +569,13 @@ class FSPCABasis(SteerableBasis2D):
             freq_cutoff=freq_cutoff,
         )
 
+    @property
+    def eigvals(self):
+        """
+        Return the eigenvals of FSPCABasis as Numpy array.
+        """
+        return self._eigvals
+
     def eigen_images(self):
         """
         Return the eigen images of the FSPCA basis, evaluated to image space.
@@ -582,7 +589,7 @@ class FSPCABasis(SteerableBasis2D):
         if isinstance(eigvecs, BlkDiagMatrix):
             eigvecs = eigvecs.dense()
 
-        return self.basis.evaluate(eigvecs.T)
+        return Coef(self.basis, eigvecs.T).evaluate()
 
     def shift(self, coef, shifts):
         """
@@ -609,3 +616,15 @@ class FSPCABasis(SteerableBasis2D):
         return self.expand_from_image_basis(
             self.evaluate_to_image_basis(coef).shift(shifts)
         )
+
+    def filter_to_basis_mat(self, f, **kwargs):
+        """
+        Convert a filter into a basis representation.
+
+        :param f: `Filter` object, usually a `CTFFilter`.
+
+        :return: Representation of filter in `basis`.
+            Return type will be based on the class's `matrix_type`.
+        """
+        # This is possible to implement, but there are no current use cases.
+        raise NotImplementedError("Not currently implemented for compressed basis.")
