@@ -3,18 +3,12 @@ import logging
 import matplotlib.pyplot as plt
 import numpy as np
 from sklearn.neighbors import NearestNeighbors
-from tqdm import tqdm
 
-from aspire.basis import FSPCABasis
-from aspire.classification import (
-    BFSReddyChatterjiAverager2D,
-    Class2D,
-    ClassSelector,
-    RandomClassSelector,
-)
+from aspire.basis import Coef, ComplexCoef, FSPCABasis
+from aspire.classification import Class2D
 from aspire.classification.legacy_implementations import bispec_2drot_large, pca_y
 from aspire.numeric import ComplexPCA
-from aspire.utils.random import rand
+from aspire.utils import random, trange
 
 logger = logging.getLogger(__name__)
 
@@ -26,18 +20,14 @@ class RIRClass2D(Class2D):
         pca_basis=None,
         fspca_components=None,
         alpha=1 / 3,
-        sample_n=4000,
+        sample_n=4000,  # Paper had 4000, but MATLAB code suggested 50000
         bispectrum_components=300,
         n_nbor=100,
-        n_classes=50,
         bispectrum_freq_cutoff=None,
         large_pca_implementation="legacy",
         nn_implementation="legacy",
         bispectrum_implementation="legacy",
-        selector=None,
-        num_procs=None,
         batch_size=512,
-        averager=None,
         dtype=None,
         seed=None,
     ):
@@ -54,8 +44,7 @@ class RIRClass2D(Class2D):
         Z. Zhao, Y. Shkolnisky, A. Singer, Rotationally Invariant Image Representation
         for Viewing Direction Classification in Cryo-EM. (2014)
 
-        :param src: Source instance.  Note it is possible to use one `source` for classification (ie CWF),
-            and a different `source` for stacking in the `averager`.
+        :param src: Source instance, for classification.
         :param pca_basis: Optional FSPCA Basis instance
         :param fspca_components: Optinally set number of components (top eigvals) to keep from full FSCPA.
             Default value of None will infer from `pca_basis` when provided, otherwise defaults to 400.
@@ -63,16 +52,11 @@ class RIRClass2D(Class2D):
         :param sample_n: Threshold for random sampling of bispectrum coefs. Default 4000,
             high values such as 50000 reduce random sampling.
         :param n_nbor: Number of nearest neighbors to compute.
-        :param n_classes: Number of class averages to return.
         :param bispectrum_freq_cutoff: Truncate (zero) high k frequecies above (int) value, defaults off (None).
         :param large_pca_implementation: See `pca`.
         :param nn_implementation: See `nn_classification`.
         :param bispectrum_implementation: See `bispectrum`.
-        :param selector: A ClassSelector subclass. Defaults to RandomClassSelector.
-        :param num_procs: Number of processes to use.
-            `None` will attempt computing a suggestion based on machine resources.
         :param batch_size: Chunk size (typically number of images) for batched methods.
-        :param averager: An Averager2D subclass. Defaults to BFSReddyChatterjiAverager2D.
         :param dtype: Optional dtype, otherwise taken from src.
         :param seed: Optional RNG seed to be passed to random methods, (example Random NN).
         :return: RIRClass2D instance to be used to compute bispectrum-like rotationally invariant 2D classification.
@@ -81,11 +65,9 @@ class RIRClass2D(Class2D):
         super().__init__(
             src=src,
             n_nbor=n_nbor,
-            n_classes=n_classes,
             seed=seed,
             dtype=dtype,
         )
-        self.num_procs = num_procs
         self.batch_size = int(batch_size)
 
         # Implementation Checks
@@ -98,6 +80,7 @@ class RIRClass2D(Class2D):
             raise ValueError(
                 f"Provided nn_implementation={nn_implementation} not in {nn_implementations.keys()}"
             )
+        self._nn_implementation = nn_implementation  # Save str for logger
         self._nn_classification = nn_implementations[nn_implementation]
 
         # # Do we have a sane Large Dataset PCA
@@ -128,13 +111,6 @@ class RIRClass2D(Class2D):
             )
         self._bispectrum = bispectrum_implementations[bispectrum_implementation]
 
-        # Setup class selection
-        if selector is None:
-            selector = RandomClassSelector(seed=self.seed)
-        elif not isinstance(selector, ClassSelector):
-            raise RuntimeError("`selector` must be subclass of `ClassSelector`")
-        self.selector = selector
-
         # For now, only run with FSPCA basis
         if pca_basis and not isinstance(pca_basis, FSPCABasis):
             raise NotImplementedError(
@@ -174,7 +150,6 @@ class RIRClass2D(Class2D):
         self.sample_n = sample_n
         self.alpha = alpha
         self.bispectrum_freq_cutoff = bispectrum_freq_cutoff
-        self.averager = averager
 
         if self.src.n < self.bispectrum_components:
             raise RuntimeError(
@@ -198,55 +173,31 @@ class RIRClass2D(Class2D):
                 self.src, components=self.fspca_components, batch_size=self.batch_size
             )
 
-        # For convenience, assign the fb_basis used in the pca_basis.
-        self.fb_basis = self.pca_basis.basis
-
-        # When not provided by a user, the averager is instantiated after
-        #  we are certain our pca_basis has been constructed.
-        if self.averager is None:
-            self.averager = BFSReddyChatterjiAverager2D(
-                self.fb_basis, self.src, num_procs=self.num_procs, dtype=self.dtype
-            )
-
         # Get the expanded coefs in the compressed FSPCA space.
         self.fspca_coef = self.pca_basis.spca_coef
 
         # Compute Bispectrum
-        coef_b, coef_b_r = self.bispectrum(self.fspca_coef)
+        coef_b, coef_b_r = self.bispectrum(Coef(self.pca_basis, self.fspca_coef))
 
         # # Stage 2: Compute Nearest Neighbors
-        logger.info("Calculate Nearest Neighbors")
-        classes, reflections, distances = self.nn_classification(coef_b, coef_b_r)
+        logger.info(f"Calculate Nearest Neighbors using {self._nn_implementation}.")
+        self.classes, self.reflections, self.distances = self.nn_classification(
+            coef_b, coef_b_r
+        )
 
         if diagnostics:
             # Lets peek at the distribution of distances
             # zero index is self, distance 0, ignored
-            plt.hist(distances[:, 1:].flatten(), bins="auto")
+            plt.hist(self.distances[:, 1:].flatten(), bins="auto")
             plt.show()
 
             # Report some information about reflections
             logger.info(
-                f"Count reflected: {np.sum(reflections)}"
-                f" {100 * np.mean(reflections) } %"
+                f"Count reflected: {np.sum(self.reflections)}"
+                f" {100 * np.mean(self.reflections)} %"
             )
 
-        return classes, reflections, distances
-
-    def averages(self, classes, reflections, distances):
-        # # Stage 3: Class Selection
-        # This is an area open to active research.
-        logger.info(f"Select {self.n_classes} Classes from Nearest Neighbors")
-        self.selection = selection = self.selector.select(
-            self.n_classes, classes, reflections, distances
-        )
-        classes, reflections = classes[selection], reflections[selection]
-
-        # # Stage 4: Averager
-        logger.info(
-            f"Begin Averaging of {classes.shape[0]} Classes using {self.averager}."
-        )
-
-        return self.averager.average(classes, reflections)
+        return self.classes, self.reflections, self.distances
 
     def pca(self, M):
         """
@@ -272,7 +223,7 @@ class RIRClass2D(Class2D):
         each having shape (n_img, features)
         where features = min(self.bispectrum_components, n_img).
 
-        Result is array (n_img, n_nbor) with entry `i` reprsenting
+        Result is array (n_img, n_nbor) with entry `i` representing
         index `i` into class input img array (src).
 
         To extend with an additonal Nearest Neighbor algo,
@@ -296,11 +247,21 @@ class RIRClass2D(Class2D):
         :param coef: complex steerable coefficients (eg. from FSPCABasis).
         :returns: tuple of arrays (coef_b, coef_b_r)
         """
+
+        if not isinstance(coef, Coef):
+            raise TypeError(
+                f"`coef` should be a `Coef` instance, received {type(coef)}"
+            )
+
         # _bispectrum is assigned during initialization.
         return self._bispectrum(coef)
 
-    def _sk_nn_classification(self, coeff_b, coeff_b_r):
-        # Before we get clever lets just use a generally accepted implementation.
+    def _sk_nn_classification(self, coef_b, coef_b_r):
+        """
+        Perform nearest neighbor classification using scikit learn.
+
+        Note "distances" are as computed by scikit, defaults to Euclidean.
+        """
 
         n_img = self.src.n
 
@@ -310,10 +271,10 @@ class RIRClass2D(Class2D):
         #   so we'll pretend we have 2*n_features of real values.
         # Don't worry about the copy because NearestNeighbors wants
         #   C-contiguous anyway... (it would copy internally otherwise)
-        X = np.column_stack((coeff_b.real, coeff_b.imag))
+        X = np.column_stack((coef_b.real, coef_b.imag))
         # We'll also want to consider the neighbors under reflection.
-        #   These coefficients should be provided by coeff_b_r
-        X_r = np.column_stack((coeff_b_r.real, coeff_b_r.imag))
+        #   These coefficients should be provided by coef_b_r
+        X_r = np.column_stack((coef_b_r.real, coef_b_r.imag))
 
         # We can compare both non-reflected and reflected representations as one large set by
         #   taking care later that we store refl=True where indices>=n_img
@@ -332,14 +293,16 @@ class RIRClass2D(Class2D):
 
         return classes, refl, distances
 
-    def _legacy_nn_classification(self, coeff_b, coeff_b_r):
+    def _legacy_nn_classification(self, coef_b, coef_b_r):
         """
-        Perform nearest neighbor classification.
+        Perform nearest neighbor classification using port of ASPIRE legacy MATLAB code.
+
+        Note `distances` returned from this method are dot products, ie "corr".
         """
 
         # Note kept ordering from legacy code (n_features, n_img)
-        coeff_b = coeff_b.T
-        coeff_b_r = coeff_b_r.T
+        coef_b = coef_b.T
+        coef_b_r = coef_b_r.T
 
         n_im = self.src.n
         # Shouldn't have more neighbors than images
@@ -350,28 +313,39 @@ class RIRClass2D(Class2D):
             )
             n_nbor = n_im - 1
 
-        concat_coeff = np.concatenate((coeff_b, coeff_b_r), axis=1)
+        concat_coef = np.concatenate((coef_b, coef_b_r), axis=1)
 
         num_batches = (n_im + self.batch_size - 1) // self.batch_size
 
         classes = np.zeros((n_im, n_nbor), dtype=int)
         distances = np.zeros((n_im, n_nbor), dtype=self.dtype)
-        for i in range(num_batches):
+
+        for i in trange(num_batches):
             start = i * self.batch_size
             finish = min((i + 1) * self.batch_size, n_im)
-            corr = np.real(
-                np.dot(np.conjugate(coeff_b[:, start:finish]).T, concat_coeff)
-            )
+            batch = np.conjugate(coef_b[:, start:finish])
+            corr = np.real(np.dot(batch.T, concat_coef))
+
+            assert np.all(
+                np.abs(corr) <= 1.01  # Allow some numerical wiggle
+            ), f"Corr out of [-1,1] bounds {np.min(corr)} {np.max(corr)}."
+            # Clamp
+            corr = np.maximum(corr, -1)
+            corr = np.minimum(corr, 1)
+
             # Note legacy did not include the original image?
             # classes[start:finish] = np.argsort(-corr, axis=1)[:, 1 : n_nbor + 1]
             # This now does include the original image
             # (Matches sklean implementation.)
-            # Check with Joakim about preference.
-            # I (GBW) think class[i] should have class[i][0] be the original image index.
-            classes[start:finish] = np.argsort(-corr, axis=1)[:, :n_nbor]
+            #
+            # Also we've converted from correlation to distance=1-correlation
+            # https://github.com/ComputationalCryoEM/ASPIRE-Python/discussions/867
+            dist = np.sqrt(2.0 - 2.0 * corr)
+
+            classes[start:finish] = np.argsort(dist, axis=1)[:, :n_nbor]
             # Store the corr values for the n_nbors in this batch
             distances[start:finish] = np.take_along_axis(
-                corr, classes[start:finish], axis=1
+                dist, classes[start:finish], axis=1
             )
 
         # There were two sets of vectors each n_img long.
@@ -392,7 +366,7 @@ class RIRClass2D(Class2D):
 
         # ### The following was from legacy code. Be careful wrt order.
         M = M.T
-        u, s, v = pca_y(M, self.bispectrum_components)
+        u, s, v = pca_y(M, self.bispectrum_components, seed=self.seed)
 
         # Contruct coefficients
         coef_b = np.einsum("i, ij -> ij", s, np.conjugate(v))
@@ -453,14 +427,14 @@ class RIRClass2D(Class2D):
             self.pca_basis.complex_angular_indices != 0
         ]  # filter non_zero_freqs eq 18,19
         pm = m / np.sum(m)
-        x = rand(len(m))
+        x = random(len(m))
         m_mask = x < self.sample_n * pm
 
         M = None
 
-        for i in tqdm(range(self.src.n)):
+        for i in trange(self.src.n):
             B = self.pca_basis.calculate_bispectrum(
-                coef_normed[i, np.newaxis],
+                ComplexCoef(self.pca_basis, coef_normed[i]),
                 filter_nonzero_freqs=True,
                 freq_cutoff=self.bispectrum_freq_cutoff,
             )
@@ -496,27 +470,59 @@ class RIRClass2D(Class2D):
 
         return coef_b, coef_b_r
 
-    def _legacy_bispectrum(self, coef):
+    def _legacy_bispectrum(self, coef, retry_attempts=3):
         """
         This code was ported to Python by an unkown author,
         and is the closest viable reference material.
 
-        It is copied here to compare while
-        fresh code is developed for this class.
+        :param coef: Real valued basis coefficients.
+        :param retry_attempts: Optional, max attempts to retry randomized
+            bispec_2drot_large.  Defaults to 3.
+
+        :return: Compressed feature and reflected feature vectors.
         """
 
+        if not isinstance(coef, Coef):
+            raise TypeError(
+                f"`coef` should be a `Coef` instance, received {type(coef)}"
+            )
+
         # The legacy code expects the complex representation
-        coef = self.pca_basis.to_complex(coef)
-        complex_eigvals = self.pca_basis.to_complex(self.pca_basis.eigvals).reshape(
-            self.pca_basis.complex_count
+        coef = self.pca_basis.to_complex(coef).asnumpy()
+        complex_eigvals = (
+            Coef(self.pca_basis, self.pca_basis.eigvals)
+            .to_complex()
+            .asnumpy()
+            .reshape(self.pca_basis.complex_count)
         )  # flatten
 
-        coef_b, coef_b_r = bispec_2drot_large(
-            coeff=coef.T,  # Note F style tranpose here and in return
-            freqs=self.pca_basis.complex_angular_indices,
-            eigval=complex_eigvals,
-            alpha=self.alpha,
-            sample_n=self.sample_n,
-        )
+        # bispec_2drot_large has a random selection component.
+        # Sometimes this can fail to return a complete feature vector.
+        # In this case we can retry, but if not successful raise an error.
+        # This seems to occur more frequently at very low resolutions (<=32),
+        # and likely requires tuning other RIR parameters for small problems.
+        attempt = 0
+        # create a local seed, convert None to an integer for this method.
+        _seed = self.seed or 0
+        while attempt < retry_attempts:
+            coef_b, coef_b_r = bispec_2drot_large(
+                coef=coef.T,  # Note F style transpose here and in return
+                freqs=self.pca_basis.complex_angular_indices,
+                eigval=complex_eigvals,
+                alpha=self.alpha,
+                sample_n=self.sample_n,
+                seed=_seed + attempt,
+            )
+            attempt += 1
+            # If we have produced a feature vector
+            if coef_b.size != 0:
+                break  # Return feature vector.
+
+        # while-else: we've exceeded retry attempts.
+        else:
+            raise RuntimeError(
+                "bispec_2drot_large failed to return valid feature vector."
+                f" Returned {coef_b.shape} after {attempt} attempts."
+            )
 
         return coef_b.T, coef_b_r.T
