@@ -3,9 +3,11 @@ import logging
 import click
 
 from aspire.basis import FFBBasis2D
-from aspire.denoising.denoiser_cov2d import DenoiserCov2D
+from aspire.commands import log_level_option
+from aspire.denoising import DenoisedSource, DenoiserCov2D
 from aspire.noise import AnisotropicNoiseEstimator, WhiteNoiseEstimator
 from aspire.source.relion import RelionSource
+from aspire.utils.logging import setConsoleLoggingLevel
 
 logger = logging.getLogger(__name__)
 
@@ -46,6 +48,7 @@ logger = logging.getLogger(__name__)
     type=str,
     help="Specified method for denoising 2D images",
 )
+@log_level_option
 def denoise(
     data_folder,
     starfile_in,
@@ -55,10 +58,14 @@ def denoise(
     max_resolution,
     noise_type,
     denoise_method,
+    loglevel,
 ):
     """
     Denoise the images and output the clean images using the default CWF method.
     """
+    # Set desired logging option for the command line
+    setConsoleLoggingLevel(loglevel)
+
     # Create a source object for 2D images
     logger.info(f"Read in images from {starfile_in} and preprocess the images.")
     source = RelionSource(
@@ -68,10 +75,10 @@ def denoise(
     logger.info(f"Set the resolution to {max_resolution} X {max_resolution}")
     if max_resolution < source.L:
         # Downsample the images
-        source.downsample(max_resolution)
+        source = source.downsample(max_resolution)
     else:
         logger.warn(f"Unable to downsample to {max_resolution}, using {source.L}")
-    source.cache()
+    source = source.cache()
 
     # Specify the fast FB basis method for expending the 2D images
     basis = FFBBasis2D((source.L, source.L))
@@ -89,12 +96,12 @@ def denoise(
 
     # Whiten the noise of images
     logger.info("Whiten the noise of images from the noise estimator")
-    source.whiten(noise_estimator.filter)
+    source = source.whiten(noise_estimator)
 
     if denoise_method == "CWF":
         logger.info("Denoise the images using CWF cov2D method.")
         denoiser = DenoiserCov2D(source, basis)
-        denoised_src = denoiser.denoise(batch_size=512)
+        denoised_src = DenoisedSource(source, denoiser)
         denoised_src.save(
             starfile_out, batch_size=512, save_mode="single", overwrite=False
         )
