@@ -4,10 +4,10 @@ import numpy as np
 from numpy.linalg import eig, inv
 from scipy.linalg import solve, sqrtm
 
-from aspire.operators import BlkDiagMatrix, RadialCTFFilter
+from aspire.basis import Coef, FFBBasis2D
+from aspire.operators import BlkDiagMatrix, DiagMatrix
 from aspire.optimization import conj_grad, fill_struct
 from aspire.utils import make_symmat
-from aspire.utils.matlab_compat import m_reshape
 
 logger = logging.getLogger(__name__)
 
@@ -100,126 +100,141 @@ class RotCov2D:
         self.dtype = self.basis.dtype
         assert basis.ndim == 2, "Only two-dimensional basis functions are needed."
 
-    def _get_mean(self, coeffs):
+    def _ctf_identity_mat(self):
+        """
+        Returns CTF identity corresponding to the `matrix_type` of `self.basis`.
+
+        :return: Identity BlkDiagMatrix or DiagMatrix
+        """
+        if self.basis.matrix_type == DiagMatrix:
+            return DiagMatrix.eye(self.basis.count, dtype=self.dtype)
+        else:
+            return BlkDiagMatrix.eye(self.basis.blk_diag_cov_shape, dtype=self.dtype)
+
+    def _get_mean(self, coefs):
         """
         Calculate the mean vector from the expansion coefficients of 2D images without CTF information.
 
-        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be averaged.
+        :param coefs: A coefficient vector (or an array of coefficient vectors) to be averaged.
         :return: The mean value vector for all images.
         """
-        if coeffs.size == 0:
+
+        if coefs.size == 0:
             raise RuntimeError("The coefficients need to be calculated first!")
-        mask = self.basis._indices["ells"] == 0
-        mean_coeff = np.zeros(self.basis.count, dtype=coeffs.dtype)
-        mean_coeff[mask] = np.mean(coeffs[..., mask], axis=0)
 
-        return mean_coeff
+        mean_coef = np.zeros(self.basis.count, dtype=coefs.dtype)
+        mean_coef[self.basis._zero_angular_inds] = np.mean(
+            coefs[..., self.basis._zero_angular_inds], axis=0
+        )
 
-    def _get_covar(self, coeffs, mean_coeff=None, do_refl=True):
+        return mean_coef
+
+    def _get_covar(self, coefs, mean_coef=None, do_refl=True):
         """
         Calculate the covariance matrix from the expansion coefficients without CTF information.
 
-        :param coeffs: A coefficient vector (or an array of coefficient vectors) calculated from 2D images.
-        :param mean_coeff: The mean vector calculated from the `coeffs`.
+        :param coefs: A coefficient vector (an array of coefficient vectors) calculated from 2D images.
+        :param mean_coef: The mean vector calculated from the `coefs`.
         :param do_refl: If true, enforce invariance to reflection (default false).
         :return: The covariance matrix of coefficients for all images.
         """
-        if coeffs.size == 0:
+        if coefs.size == 0:
             raise RuntimeError("The coefficients need to be calculated first!")
-        if mean_coeff is None:
-            mean_coeff = self._get_mean(coeffs)
+        if mean_coef is None:
+            mean_coef = self._get_mean(coefs)
 
         # Initialize a totally empty BlkDiagMatrix, build incrementally.
-        covar_coeff = BlkDiagMatrix.empty(0, dtype=coeffs.dtype)
+        covar_coef = BlkDiagMatrix.empty(0, dtype=coefs.dtype)
         ell = 0
-        mask = self.basis._indices["ells"] == ell
-        coeff_ell = coeffs[..., mask] - mean_coeff[mask]
-        covar_ell = np.array(coeff_ell.T @ coeff_ell / coeffs.shape[0])
-        covar_coeff.append(covar_ell)
+
+        mask = self.basis.angular_indices == ell
+
+        coef_ell = coefs[..., mask] - mean_coef[mask]
+        covar_ell = np.array(coef_ell.T @ coef_ell / coefs.shape[0])
+        covar_coef.append(covar_ell)
 
         for ell in range(1, self.basis.ell_max + 1):
-            mask = self.basis._indices["ells"] == ell
-            mask_pos = [
-                mask[i] and (self.basis._indices["sgns"][i] == +1)
-                for i in range(len(mask))
-            ]
-            mask_neg = [
-                mask[i] and (self.basis._indices["sgns"][i] == -1)
-                for i in range(len(mask))
-            ]
+            mask_ell = self.basis.angular_indices == ell
+            mask_pos = mask_ell & (self.basis.signs_indices == +1)
+            mask_neg = mask_ell & (self.basis.signs_indices == -1)
+
             covar_ell_diag = np.array(
-                coeffs[:, mask_pos].T @ coeffs[:, mask_pos]
-                + coeffs[:, mask_neg].T @ coeffs[:, mask_neg]
-            ) / (2 * coeffs.shape[0])
+                coefs[:, mask_pos].T @ coefs[:, mask_pos]
+                + coefs[:, mask_neg].T @ coefs[:, mask_neg]
+            ) / (2 * coefs.shape[0])
 
             if do_refl:
-                covar_coeff.append(covar_ell_diag)
-                covar_coeff.append(covar_ell_diag)
+                covar_coef.append(covar_ell_diag)
+                covar_coef.append(covar_ell_diag)
             else:
                 covar_ell_off = np.array(
                     (
-                        coeffs[:, mask_pos] @ coeffs[:, mask_neg].T / coeffs.shape[0]
-                        - coeffs[:, mask_pos].T @ coeffs[:, mask_neg]
+                        coefs[:, mask_pos] @ coefs[:, mask_neg].T / coefs.shape[0]
+                        - coefs[:, mask_pos].T @ coefs[:, mask_neg]
                     )
-                    / (2 * coeffs.shape[0])
+                    / (2 * coefs.shape[0])
                 )
 
                 hsize = covar_ell_diag.shape[0]
-                covar_coeff_blk = np.zeros((2, hsize, 2, hsize))
+                covar_coef_blk = np.zeros((2, hsize, 2, hsize))
 
-                covar_coeff_blk[0:2, :, 0:2, :] = covar_ell_diag[:hsize, :hsize]
-                covar_coeff_blk[0, :, 1, :] = covar_ell_off[:hsize, :hsize]
-                covar_coeff_blk[1, :, 0, :] = covar_ell_off.T[:hsize, :hsize]
+                covar_coef_blk[0:2, :, 0:2, :] = covar_ell_diag[:hsize, :hsize]
+                covar_coef_blk[0, :, 1, :] = covar_ell_off[:hsize, :hsize]
+                covar_coef_blk[1, :, 0, :] = covar_ell_off.T[:hsize, :hsize]
 
-                covar_coeff.append(covar_coeff_blk.reshape(2 * hsize, 2 * hsize))
+                covar_coef.append(covar_coef_blk.reshape(2 * hsize, 2 * hsize))
 
-        return covar_coeff
+        return covar_coef
 
-    def get_mean(self, coeffs, ctf_fb=None, ctf_idx=None):
+    def get_mean(self, coefs, ctf_basis=None, ctf_idx=None):
         """
         Calculate the mean vector from the expansion coefficients with CTF information.
 
-        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be averaged.
-        :param ctf_fb: The CFT functions in the FB expansion.
-        :param ctf_idx: An array of the CFT function indices for all 2D images.
-            If ctf_fb or ctf_idx is None, the identity filter will be applied.
+        :param coefs: A coefficient vector (or an array of coefficient vectors) to be averaged.
+        :param ctf_basis: The CTF functions in the Basis expansion.
+        :param ctf_idx: An array of the CTF function indices for all 2D images.
+            If ctf_basis or ctf_idx is None, the identity filter will be applied.
         :return: The mean value vector for all images.
         """
 
-        if coeffs.size == 0:
+        if not isinstance(coefs, Coef):
+            raise TypeError(
+                f"`coefs` should be instance of `Coef`, received {type(Coef)}."
+            )
+
+        coefs = coefs.asnumpy()
+
+        # TODO: Redundant, remove?
+        if coefs.size == 0:
             raise RuntimeError("The coefficients need to be calculated!")
 
         # should assert we require none or both...
-        if (ctf_fb is None) or (ctf_idx is None):
-            ctf_idx = np.zeros(coeffs.shape[0], dtype=int)
-            ctf_fb = [
-                BlkDiagMatrix.eye_like(
-                    RadialCTFFilter().fb_mat(self.basis), dtype=self.dtype
-                )
-            ]
+        if (ctf_basis is None) or (ctf_idx is None):
+            ctf_idx = np.zeros(coefs.shape[0], dtype=int)
+            ctf_basis = [self._ctf_identity_mat()]
 
-        b = np.zeros(self.basis.count, dtype=coeffs.dtype)
+        b = np.zeros(self.basis.count, dtype=coefs.dtype)
 
-        A = BlkDiagMatrix.zeros_like(ctf_fb[0])
+        A = BlkDiagMatrix.zeros(self.basis.blk_diag_cov_shape)
         for k in np.unique(ctf_idx[:]).T:
-            coeff_k = coeffs[ctf_idx == k]
-            weight = coeff_k.shape[0] / coeffs.shape[0]
-            mean_coeff_k = self._get_mean(coeff_k)
+            coef_k = coefs[ctf_idx == k]
+            weight = coef_k.shape[0] / coefs.shape[0]
+            mean_coef_k = self._get_mean(coef_k)
 
-            ctf_fb_k = ctf_fb[k]
-            ctf_fb_k_t = ctf_fb_k.T
-            b += weight * ctf_fb_k_t.apply(mean_coeff_k)
-            A += weight * (ctf_fb_k_t @ ctf_fb_k)
+            ctf_basis_k = ctf_basis[k]
+            ctf_basis_k_t = ctf_basis_k.T
+            b += weight * ctf_basis_k_t.apply(mean_coef_k)
+            A += weight * (ctf_basis_k_t @ ctf_basis_k)
 
-        mean_coeff = A.solve(b)
-        return mean_coeff
+        mean_coef = A.solve(b)
+        return Coef(self.basis, mean_coef)
 
     def get_covar(
         self,
-        coeffs,
-        ctf_fb=None,
+        coefs,
+        ctf_basis=None,
         ctf_idx=None,
-        mean_coeff=None,
+        mean_coef=None,
         do_refl=True,
         noise_var=0,
         covar_est_opt=None,
@@ -228,34 +243,36 @@ class RotCov2D:
         """
         Calculate the covariance matrix from the expansion coefficients and CTF information.
 
-        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be calculated.
-        :param ctf_fb: The CFT functions in the FB expansion.
-        :param ctf_idx: An array of the CFT function indices for all 2D images.
-            If ctf_fb or ctf_idx is None, the identity filter will be applied.
-        :param mean_coeff: The mean value vector from all images.
-        :param noise_var: The estimated variance of noise. The value should be zero for `coeffs`
+        :param coefs: A coefficient vector (or an array of coefficient vectors) to be calculated.
+        :param ctf_basis: The CTF functions in the Basis expansion.
+        :param ctf_idx: An array of the CTF function indices for all 2D images.
+            If ctf_basis or ctf_idx is None, the identity filter will be applied.
+        :param mean_coef: The mean value vector from all images.
+        :param noise_var: The estimated variance of noise. The value should be zero for `coefs`
             from clean images of simulation data.
         :param covar_est_opt: The optimization parameter list for obtaining the Cov2D matrix.
         :param make_psd: If True, make the covariance matrix positive semidefinite
         :return: The basis coefficients of the covariance matrix in
             the form of cell array representing a block diagonal matrix. These
             block diagonal matrices are implemented as BlkDiagMatrix instances.
-            The covariance is calculated from the images represented by the coeffs array,
+            The covariance is calculated from the images represented by the coefs array,
             along with all possible rotations and reflections. As a result, the computed covariance
-            matrix is invariant to both reflection and rotation. The effect of the filters in ctf_fb
+            matrix is invariant to both reflection and rotation. The effect of the filters in ctf_basis
             are accounted for and inverted to yield a covariance estimate of the unfiltered images.
         """
 
-        if coeffs.size == 0:
+        if not isinstance(coefs, Coef):
+            raise TypeError(
+                f"`coefs` should be instance of `Coef`, received {type(Coef)}."
+            )
+        coefs = coefs.asnumpy()
+
+        if coefs.size == 0:
             raise RuntimeError("The coefficients need to be calculated!")
 
-        if (ctf_fb is None) or (ctf_idx is None):
-            ctf_idx = np.zeros(coeffs.shape[0], dtype=int)
-            ctf_fb = [
-                BlkDiagMatrix.eye_like(
-                    RadialCTFFilter().fb_mat(self.basis), dtype=self.dtype
-                )
-            ]
+        if (ctf_basis is None) or (ctf_idx is None):
+            ctf_idx = np.zeros(coefs.shape[0], dtype=int)
+            ctf_basis = [self._ctf_identity_mat()]
 
         def identity(x):
             return x
@@ -273,45 +290,48 @@ class RotCov2D:
 
         covar_est_opt = fill_struct(covar_est_opt, default_est_opt)
 
-        if mean_coeff is None:
-            mean_coeff = self.get_mean(coeffs, ctf_fb, ctf_idx)
+        if mean_coef is None:
+            mean_coef = self.get_mean(Coef(self.basis, coefs), ctf_basis, ctf_idx)
 
-        b_coeff = BlkDiagMatrix.zeros_like(ctf_fb[0])
-        b_noise = BlkDiagMatrix.zeros_like(ctf_fb[0])
+        b_coef = BlkDiagMatrix.zeros(self.basis.blk_diag_cov_shape)
+        b_noise = BlkDiagMatrix.zeros(self.basis.blk_diag_cov_shape)
         A = []
-        for _ in range(len(ctf_fb)):
-            A.append(BlkDiagMatrix.zeros_like(ctf_fb[0]))
+        for _ in range(len(ctf_basis)):
+            A.append(BlkDiagMatrix.zeros(self.basis.blk_diag_cov_shape))
 
-        M = BlkDiagMatrix.zeros_like(ctf_fb[0])
+        M = BlkDiagMatrix.zeros(self.basis.blk_diag_cov_shape)
 
         for k in np.unique(ctf_idx[:]):
+            coef_k = coefs[ctf_idx == k].astype(self.dtype)
+            weight = coef_k.shape[0] / coefs.shape[0]
 
-            coeff_k = coeffs[ctf_idx == k].astype(self.dtype)
-            weight = coeff_k.shape[0] / coeffs.shape[0]
+            ctf_basis_k = ctf_basis[k]
+            ctf_basis_k_t = ctf_basis_k.T
+            mean_coef_k = ctf_basis_k.apply(mean_coef.asnumpy()[0])
+            covar_coef_k = self._get_covar(coef_k, mean_coef_k)
 
-            ctf_fb_k = ctf_fb[k]
-            ctf_fb_k_t = ctf_fb_k.T
-            mean_coeff_k = ctf_fb_k.apply(mean_coeff)
-            covar_coeff_k = self._get_covar(coeff_k, mean_coeff_k)
+            b_coef += weight * (ctf_basis_k_t @ covar_coef_k @ ctf_basis_k)
 
-            b_coeff += weight * (ctf_fb_k_t @ covar_coeff_k @ ctf_fb_k)
+            ctf_basis_k_sq = ctf_basis_k_t @ ctf_basis_k
+            b_noise += weight * ctf_basis_k_sq
 
-            ctf_fb_k_sq = ctf_fb_k_t @ ctf_fb_k
-            b_noise += weight * ctf_fb_k_sq
+            A_k = np.sqrt(weight) * ctf_basis_k_sq
+            if not isinstance(A_k, BlkDiagMatrix):
+                A_k = DiagMatrix(A_k).as_blk_diag(self.basis.blk_diag_cov_shape)
 
-            A[k] = np.sqrt(weight) * ctf_fb_k_sq
+            A[k] = A_k
             M += A[k]
 
-        if not b_coeff.check_psd():
+        if not b_coef.check_psd():
             logger.warning("Left side b in Cov2D is not positive semidefinite.")
 
         if covar_est_opt["shrinker"] is None:
-            b = b_coeff - noise_var * b_noise
+            b = b_coef - noise_var * b_noise
         else:
             b = self.shrink_covar_backward(
-                b_coeff,
+                b_coef,
                 b_noise,
-                np.size(coeffs, 0),
+                np.size(coefs, 0),
                 noise_var,
                 covar_est_opt["shrinker"],
             )
@@ -325,45 +345,43 @@ class RotCov2D:
 
         cg_opt = covar_est_opt
 
-        covar_coeff = BlkDiagMatrix.zeros_like(ctf_fb[0])
+        covar_coef = BlkDiagMatrix.zeros(self.basis.blk_diag_cov_shape)
 
         def precond_fun(S, x):
             p = np.size(S, 0)
             assert np.size(x) == p * p, "The sizes of S and x are not consistent."
-            x = m_reshape(x, (p, p))
+            x = x.reshape(p, p)
             y = S @ x @ S
-            y = m_reshape(y, (p**2,))
-            return y
+            return y.flatten()
 
         def apply(A, x):
             p = np.size(A[0], 0)
-            x = m_reshape(x, (p, p))
+            x = x.reshape(p, p)
             y = np.zeros_like(x)
             for k in range(0, len(A)):
                 y = y + A[k] @ x @ A[k].T
-            y = m_reshape(y, (p**2,))
-            return y
+            return y.flatten()
 
         for ell in range(0, len(b)):
             A_ell = []
             for k in range(0, len(A)):
                 A_ell.append(A[k][ell])
             p = np.size(A_ell[0], 0)
-            b_ell = m_reshape(b[ell], (p**2,))
+            b_ell = b[ell].flatten()
             S = inv(M[ell])
             cg_opt["preconditioner"] = lambda x, S=S: precond_fun(S, x)
-            covar_coeff_ell, _, _ = conj_grad(
+            covar_coef_ell, _, _ = conj_grad(
                 lambda x, A_ell=A_ell: apply(A_ell, x), b_ell, cg_opt
             )
-            covar_coeff[ell] = m_reshape(covar_coeff_ell, (p, p))
+            covar_coef[ell] = covar_coef_ell.reshape(p, p)
 
-        if not covar_coeff.check_psd():
+        if not covar_coef.check_psd():
             logger.warning("Covariance matrix in Cov2D is not positive semidefinite.")
             if make_psd:
                 logger.info("Convert matrices to positive semidefinite.")
-                covar_coeff = covar_coeff.make_psd()
+                covar_coef = covar_coef.make_psd()
 
-        return covar_coeff
+        return covar_coef
 
     def shrink_covar_backward(self, b, b_noise, n, noise_var, shrinker):
         """
@@ -389,76 +407,83 @@ class RotCov2D:
             b_out[ell] = b_ell
         return b_out
 
-    def get_cwf_coeffs(
+    def get_cwf_coefs(
         self,
-        coeffs,
-        ctf_fb=None,
+        coefs,
+        ctf_basis=None,
         ctf_idx=None,
-        mean_coeff=None,
-        covar_coeff=None,
+        mean_coef=None,
+        covar_coef=None,
         noise_var=0,
     ):
         """
         Estimate the expansion coefficients using the Covariance Wiener Filtering (CWF) method.
 
-        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be calculated.
-        :param ctf_fb: The CFT functions in the FB expansion.
-        :param ctf_idx: An array of the CFT function indices for all 2D images.
-            If ctf_fb or ctf_idx is None, the identity filter will be applied.
-        :param mean_coeff: The mean value vector from all images.
-        :param covar_coeff: The block diagonal covariance matrix of the clean coefficients represented by a cell array.
-        :param noise_var: The estimated variance of noise. The value should be zero for `coeffs`
+        :param coefs: A coefficient vector (or an array of coefficient vectors) to be calculated.
+        :param ctf_basis: The CTF functions in the Basis expansion.
+        :param ctf_idx: An array of the CTF function indices for all 2D images.
+            If ctf_basis or ctf_idx is None, the identity filter will be applied.
+        :param mean_coef: The mean value vector from all images.
+        :param covar_coef: The block diagonal covariance matrix of the clean coefficients represented by a cell array.
+        :param noise_var: The estimated variance of noise. The value should be zero for `coefs`
             from clean images of simulation data.
         :return: The estimated coefficients of the unfiltered images in certain math basis.
             These are obtained using a Wiener filter with the specified covariance for the clean images
             and white noise of variance `noise_var` for the noise.
         """
 
-        if mean_coeff is None:
-            mean_coeff = self.get_mean(coeffs, ctf_fb, ctf_idx)
-
-        if covar_coeff is None:
-            covar_coeff = self.get_covar(
-                coeffs, ctf_fb, ctf_idx, mean_coeff, noise_var=noise_var
+        if not isinstance(coefs, Coef):
+            raise TypeError(
+                f"`coefs` should be instance of `Coef`, received {type(Coef)}."
             )
 
+        if mean_coef is None:
+            mean_coef = self.get_mean(coefs, ctf_basis, ctf_idx)
+
+        if covar_coef is None:
+            covar_coef = self.get_covar(
+                coefs, ctf_basis, ctf_idx, mean_coef, noise_var=noise_var
+            )
+
+        coefs = coefs.asnumpy()
+
         # Handle CTF arguments.
-        if (ctf_fb is None) ^ (ctf_idx is None):
+        if (ctf_basis is None) ^ (ctf_idx is None):
             raise RuntimeError(
-                "Both `ctf_fb` and `ctf_idx` should be provided,"
+                "Both `ctf_basis` and `ctf_idx` should be provided,"
                 " or both should be `None`."
-                f' Given {"ctf_fb" if ctf_idx is None else "ctf_idx"}'
+                f' Given {"ctf_basis" if ctf_idx is None else "ctf_idx"}'
             )
-        elif ctf_fb is None:
+        elif ctf_basis is None:
             # Setup defaults for CTF
-            ctf_idx = np.zeros(coeffs.shape[0], dtype=int)
-            ctf_fb = [BlkDiagMatrix.eye_like(covar_coeff)]
+            ctf_idx = np.zeros(coefs.shape[0], dtype=int)
+            ctf_basis = [BlkDiagMatrix.eye_like(covar_coef)]
 
-        noise_covar_coeff = noise_var * BlkDiagMatrix.eye_like(covar_coeff)
+        noise_covar_coef = noise_var * BlkDiagMatrix.eye_like(covar_coef)
 
-        coeffs_est = np.zeros_like(coeffs)
+        coefs_est = np.zeros_like(coefs)
 
         for k in np.unique(ctf_idx[:]):
-            coeff_k = coeffs[ctf_idx == k]
-            ctf_fb_k = ctf_fb[k]
-            ctf_fb_k_t = ctf_fb_k.T
+            coef_k = coefs[ctf_idx == k]
+            ctf_basis_k = ctf_basis[k]
+            ctf_basis_k_t = ctf_basis_k.T
 
-            mean_coeff_k = ctf_fb_k.apply(mean_coeff)
-            coeff_est_k = coeff_k - mean_coeff_k
+            mean_coef_k = ctf_basis_k.apply(mean_coef.asnumpy()[0])
+            coef_est_k = coef_k - mean_coef_k
 
             if noise_var == 0:
-                coeff_est_k = ctf_fb_k.solve(coeff_est_k.T).T
+                coef_est_k = ctf_basis_k.solve(coef_est_k.T).T
             else:
-                sig_covar_coeff = ctf_fb_k @ covar_coeff @ ctf_fb_k_t
-                sig_noise_covar_coeff = sig_covar_coeff + noise_covar_coeff
+                sig_covar_coef = ctf_basis_k @ covar_coef @ ctf_basis_k_t
+                sig_noise_covar_coef = sig_covar_coef + noise_covar_coef
 
-                coeff_est_k = sig_noise_covar_coeff.solve(coeff_est_k.T).T
-                coeff_est_k = (covar_coeff @ ctf_fb_k_t).apply(coeff_est_k.T).T
+                coef_est_k = sig_noise_covar_coef.solve(coef_est_k.T).T
+                coef_est_k = (covar_coef @ ctf_basis_k_t).apply(coef_est_k.T).T
 
-            coeff_est_k = coeff_est_k + mean_coeff
-            coeffs_est[ctf_idx == k] = coeff_est_k
+            coef_est_k = coef_est_k + mean_coef
+            coefs_est[ctf_idx == k] = coef_est_k
 
-        return coeffs_est
+        return Coef(self.basis, coefs_est)
 
 
 class BatchedRotCov2D(RotCov2D):
@@ -498,57 +523,64 @@ class BatchedRotCov2D(RotCov2D):
         src = self.src
 
         if self.basis is None:
-            from aspire.basis import FFBBasis2D
-
             self.basis = FFBBasis2D((src.L, src.L), dtype=self.dtype)
 
         if not src.unique_filters:
             logger.info("CTF filters are not included in Cov2D denoising")
             # set all CTF filters to an identity filter
             self.ctf_idx = np.zeros(src.n, dtype=int)
-            self.ctf_fb = [BlkDiagMatrix.eye_like(RadialCTFFilter().fb_mat(self.basis))]
+            self.ctf_basis = [self._ctf_identity_mat()]
+
         else:
-            logger.info("Represent CTF filters in FB basis")
+            logger.info("Represent CTF filters in basis")
             unique_filters = src.unique_filters
             self.ctf_idx = src.filter_indices
-            self.ctf_fb = [f.fb_mat(self.basis) for f in unique_filters]
+            self.ctf_basis = [
+                self.basis.filter_to_basis_mat(f, pixel_size=self.src.pixel_size)
+                for f in unique_filters
+            ]
 
     def _calc_rhs(self):
         src = self.src
         basis = self.basis
 
-        ctf_fb = self.ctf_fb
+        ctf_basis = self.ctf_basis
         ctf_idx = self.ctf_idx
 
-        zero_coeff = np.zeros((basis.count,), dtype=self.dtype)
+        zero_coef = np.zeros((basis.count,), dtype=self.dtype)
 
-        b_mean = [np.zeros(basis.count, dtype=self.dtype) for _ in ctf_fb]
+        b_mean = [np.zeros(basis.count, dtype=self.dtype) for _ in ctf_basis]
 
-        b_covar = BlkDiagMatrix.zeros_like(ctf_fb[0])
+        b_covar = BlkDiagMatrix.zeros(self.basis.blk_diag_cov_shape, dtype=self.dtype)
 
         for start in range(0, src.n, self.batch_size):
             batch = np.arange(start, min(start + self.batch_size, src.n))
 
             im = src.images[batch[0] : batch[0] + len(batch)]
-            coeff = basis.evaluate_t(im)
+            coef = basis.evaluate_t(im).asnumpy()
 
             for k in np.unique(ctf_idx[batch]):
-                coeff_k = coeff[ctf_idx[batch] == k]
-                weight = np.size(coeff_k, 0) / src.n
+                coef_k = coef[ctf_idx[batch] == k]
+                weight = np.size(coef_k, 0) / src.n
 
-                mean_coeff_k = self._get_mean(coeff_k)
+                mean_coef_k = self._get_mean(coef_k)
 
-                ctf_fb_k = ctf_fb[k]
-                ctf_fb_k_t = ctf_fb_k.T
+                ctf_basis_k = ctf_basis[k]
+                ctf_basis_k_t = ctf_basis_k.T
 
-                b_mean_k = weight * ctf_fb_k_t.apply(mean_coeff_k)
+                b_mean_k = weight * ctf_basis_k_t.apply(mean_coef_k)
+
+                if isinstance(b_mean_k, DiagMatrix):
+                    # Convert to a column vector
+                    b_mean_k = b_mean_k.asnumpy().T
 
                 b_mean[k] += b_mean_k
 
-                covar_coeff_k = self._get_covar(coeff_k, zero_coeff)
+                covar_coef_k = self._get_covar(coef_k, zero_coef)
 
-                b_covar_k = ctf_fb_k_t @ covar_coeff_k
-                b_covar_k = b_covar_k @ ctf_fb_k
+                b_covar_k = ctf_basis_k_t @ covar_coef_k
+
+                b_covar_k = b_covar_k @ ctf_basis_k
                 b_covar_k *= weight
 
                 b_covar += b_covar_k
@@ -559,24 +591,23 @@ class BatchedRotCov2D(RotCov2D):
     def _calc_op(self):
         src = self.src
 
-        ctf_fb = self.ctf_fb
+        ctf_basis = self.ctf_basis
         ctf_idx = self.ctf_idx
 
-        A_mean = BlkDiagMatrix.zeros_like(ctf_fb[0])
-        A_covar = [None for _ in ctf_fb]
-        M_covar = BlkDiagMatrix.zeros_like(ctf_fb[0])
+        A_mean = BlkDiagMatrix.zeros(self.basis.blk_diag_cov_shape, self.dtype)
+        A_covar = [None for _ in ctf_basis]
+        M_covar = BlkDiagMatrix.zeros_like(A_mean)
 
         for k in np.unique(ctf_idx):
-            weight = np.count_nonzero(ctf_idx == k) / src.n
+            weight = float(np.count_nonzero(ctf_idx == k) / src.n)
 
-            ctf_fb_k = ctf_fb[k]
-            ctf_fb_k_t = ctf_fb_k.T
+            ctf_basis_k = ctf_basis[k]
+            ctf_basis_k_t = ctf_basis_k.T
 
-            ctf_fb_k_sq = ctf_fb_k_t @ ctf_fb_k
-            A_mean_k = weight * ctf_fb_k_sq
+            ctf_basis_k_sq = ctf_basis_k_t @ ctf_basis_k
+            A_mean_k = weight * ctf_basis_k_sq
             A_mean += A_mean_k
-
-            A_covar_k = np.sqrt(weight) * ctf_fb_k_sq
+            A_covar_k = np.sqrt(weight).astype(self.dtype) * ctf_basis_k_sq
             A_covar[k] = A_covar_k
 
             M_covar += A_covar_k
@@ -585,34 +616,34 @@ class BatchedRotCov2D(RotCov2D):
         self.A_covar = A_covar
         self.M_covar = M_covar
 
-    def _mean_correct_covar_rhs(self, b_covar, b_mean, mean_coeff):
+    def _mean_correct_covar_rhs(self, b_covar, b_mean, mean_coef):
         src = self.src
 
-        ctf_fb = self.ctf_fb
+        ctf_basis = self.ctf_basis
         ctf_idx = self.ctf_idx
 
-        partition = ctf_fb[0].partition
+        partition = self.basis.blk_diag_cov_shape
 
         # Note: If we don't do this, we'll be modifying the stored `b_covar`
         # since the operations below are in-place.
         b_covar = b_covar.copy()
 
         for k in np.unique(ctf_idx):
-            weight = np.count_nonzero(ctf_idx == k) / src.n
+            weight = float(np.count_nonzero(ctf_idx == k) / src.n)
 
-            ctf_fb_k = ctf_fb[k]
-            ctf_fb_k_t = ctf_fb_k.T
+            ctf_basis_k = ctf_basis[k]
+            ctf_basis_k_t = ctf_basis_k.T
 
-            mean_coeff_k = ctf_fb_k.apply(mean_coeff)
-            mean_coeff_k = ctf_fb_k_t.apply(mean_coeff_k)
+            mean_coef_k = ctf_basis_k.apply(mean_coef.asnumpy()[0])
+            mean_coef_k = ctf_basis_k_t.apply(mean_coef_k)
 
-            mean_coeff_k = mean_coeff_k[: partition[0][0]]
+            mean_coef_k = mean_coef_k[: partition[0][0]]
             b_mean_k = b_mean[k][: partition[0][0]]
 
             correction = (
-                np.outer(mean_coeff_k, b_mean_k)
-                + np.outer(b_mean_k, mean_coeff_k)
-                - weight * np.outer(mean_coeff_k, mean_coeff_k)
+                np.outer(mean_coef_k, b_mean_k)
+                + np.outer(b_mean_k, mean_coef_k)
+                - weight * np.outer(mean_coef_k, mean_coef_k)
             )
 
             b_covar[0] -= correction
@@ -631,42 +662,60 @@ class BatchedRotCov2D(RotCov2D):
         return b_covar
 
     def _solve_covar(self, A_covar, b_covar, M, covar_est_opt):
-        ctf_fb = self.ctf_fb
+        method = self._solve_covar_cg
+        if self.basis.matrix_type == DiagMatrix:
+            method = self._solve_covar_direct
 
+        return method(A_covar, b_covar, M, covar_est_opt)
+
+    def _solve_covar_direct(self, A_covar, b_covar, M, covar_est_opt):
+        # A_covar is a list of DiagMatrix, representing each ctf in self.basis.
+        # b_covar is a BlkDiagMatrix
+        # M is sum of weighted A squared, only used for cg, ignore here.
+        A_covar = DiagMatrix(np.concatenate([x.asnumpy() for x in A_covar]))
+        A2i = A_covar * A_covar
+
+        res = BlkDiagMatrix.empty(b_covar.nblocks, self.dtype)
+        for b in range(b_covar.nblocks):
+            res.data[b] = b_covar[b] / A2i[b]
+
+        return res
+
+    def _solve_covar_cg(self, A_covar, b_covar, M, covar_est_opt):
         def precond_fun(S, x):
             p = np.size(S, 0)
             assert np.size(x) == p * p, "The sizes of S and x are not consistent."
-            x = m_reshape(x, (p, p))
+            x = x.reshape(p, p)
             y = S @ x @ S
-            y = m_reshape(y, (p**2,))
-            return y
+            return y.flatten()
 
         def apply(A, x):
             p = np.size(A[0], 0)
-            x = m_reshape(x, (p, p))
+            x = x.reshape(p, p)
             y = np.zeros_like(x)
             for k in range(0, len(A)):
                 y = y + A[k] @ x @ A[k].T
-            y = m_reshape(y, (p**2,))
-            return y
+            return y.flatten()
 
         cg_opt = covar_est_opt
-        covar_coeff = BlkDiagMatrix.zeros_like(ctf_fb[0])
+        covar_coef = BlkDiagMatrix.zeros(
+            self.basis.blk_diag_cov_shape, dtype=self.dtype
+        )
 
         for ell in range(0, len(b_covar)):
             A_ell = []
             for k in range(0, len(A_covar)):
                 A_ell.append(A_covar[k][ell])
             p = np.size(A_ell[0], 0)
-            b_ell = m_reshape(b_covar[ell], (p**2,))
+            b_ell = b_covar[ell].flatten()
             S = inv(M[ell])
             cg_opt["preconditioner"] = lambda x, S=S: precond_fun(S, x)
-            covar_coeff_ell, _, _ = conj_grad(
+            covar_coef_ell, _, _ = conj_grad(
                 lambda x, A_ell=A_ell: apply(A_ell, x), b_ell, cg_opt
             )
-            covar_coeff[ell] = m_reshape(covar_coeff_ell, (p, p))
+            covar_coef[ell] = covar_coef_ell.reshape(p, p)
 
-        return covar_coeff
+        return covar_coef
 
     def get_mean(self):
         """
@@ -683,19 +732,17 @@ class BatchedRotCov2D(RotCov2D):
             self._calc_op()
 
         b_mean_all = np.stack(self.b_mean).sum(axis=0)
-        mean_coeff = self.A_mean.solve(b_mean_all)
+        mean_coef = self.A_mean.solve(b_mean_all)
 
-        return mean_coeff
+        return Coef(self.basis, mean_coef)
 
-    def get_covar(
-        self, noise_var=0, mean_coeff=None, covar_est_opt=None, make_psd=True
-    ):
+    def get_covar(self, noise_var=0, mean_coef=None, covar_est_opt=None, make_psd=True):
         """
         Calculate the block diagonal covariance matrix in the basis
         coefficients.
 
         :param noise_var: The variance of the noise in the images (default 1)
-        :param mean_coeff: If specified, overrides the mean coefficient vector
+        :param mean_coef: If specified, overrides the mean coefficient vector
             used to calculate the covariance (default `self.get_mean()`).
         :param :covar_est_opt: The estimation parameters for obtaining the covariance
             matrix in the form of a dictionary. Keys include:
@@ -744,12 +791,12 @@ class BatchedRotCov2D(RotCov2D):
         if not self.A_covar or self.M_covar:
             self._calc_op()
 
-        if mean_coeff is None:
-            mean_coeff = self.get_mean()
+        if mean_coef is None:
+            mean_coef = self.get_mean()
 
         b_covar = self.b_covar
 
-        b_covar = self._mean_correct_covar_rhs(b_covar, self.b_mean, mean_coeff)
+        b_covar = self._mean_correct_covar_rhs(b_covar, self.b_mean, mean_coef)
         if not b_covar.check_psd():
             logger.warning("Left side b in Batched Cov2D is not positive semidefinite.")
 
@@ -762,78 +809,84 @@ class BatchedRotCov2D(RotCov2D):
                 "in Batched Cov2D is not positive semidefinite."
             )
 
-        covar_coeff = self._solve_covar(
+        covar_coef = self._solve_covar(
             self.A_covar, b_covar, self.M_covar, covar_est_opt
         )
-        if not covar_coeff.check_psd():
+        if not covar_coef.check_psd():
             logger.warning(
                 "Covariance matrix in Batched Cov2D is not positive semidefinite."
             )
             if make_psd:
                 logger.info("Convert matrices to positive semidefinite.")
-                covar_coeff = covar_coeff.make_psd()
+                covar_coef = covar_coef.make_psd()
 
-        return covar_coeff
+        return covar_coef
 
-    def get_cwf_coeffs(
-        self, coeffs, ctf_fb, ctf_idx, mean_coeff, covar_coeff, noise_var=0
+    def get_cwf_coefs(
+        self, coefs, ctf_basis, ctf_idx, mean_coef, covar_coef, noise_var=0
     ):
         """
         Estimate the expansion coefficients using the Covariance Wiener Filtering (CWF) method.
 
-        :param coeffs: A coefficient vector (or an array of coefficient vectors) to be calculated.
-        :param ctf_fb: The CFT functions in the FB expansion.
-        :param ctf_idx: An array of the CFT function indices for all 2D images.
-            If ctf_fb or ctf_idx is None, the identity filter will be applied.
-        :param mean_coeff: The mean value vector from all images.
-        :param covar_coeff: The block diagonal covariance matrix of the clean coefficients represented by a cell array.
-        :param noise_var: The estimated variance of noise. The value should be zero for `coeffs`
+        :param coefs: A coefficient vector (or an array of coefficient vectors) to be calculated.
+        :param ctf_basis: The CTF functions in the Basis expansion.
+        :param ctf_idx: An array of the CTF function indices for all 2D images.
+            If ctf_basis or ctf_idx is None, the identity filter will be applied.
+        :param mean_coef: The mean value vector from all images.
+        :param covar_coef: The block diagonal covariance matrix of the clean coefficients represented by a cell array.
+        :param noise_var: The estimated variance of noise. The value should be zero for `coefs`
             from clean images of simulation data.
         :return: The estimated coefficients of the unfiltered images in certain math basis.
             These are obtained using a Wiener filter with the specified covariance for the clean images
             and white noise of variance `noise_var` for the noise.
         """
 
-        if mean_coeff is None:
-            mean_coeff = self.get_mean()
+        if not isinstance(coefs, Coef):
+            raise TypeError(
+                f"`coefs` should be instance of `Coef`, received {type(Coef)}."
+            )
+        coefs = coefs.asnumpy()
 
-        if covar_coeff is None:
-            covar_coeff = self.get_covar(noise_var=noise_var, mean_coeff=mean_coeff)
+        if mean_coef is None:
+            mean_coef = self.get_mean()
+
+        if covar_coef is None:
+            covar_coef = self.get_covar(noise_var=noise_var, mean_coef=mean_coef)
 
         # Handle CTF arguments.
-        if (ctf_fb is None) ^ (ctf_idx is None):
+        if (ctf_basis is None) ^ (ctf_idx is None):
             raise RuntimeError(
-                "Both `ctf_fb` and `ctf_idx` should be provided,"
+                "Both `ctf_basis` and `ctf_idx` should be provided,"
                 " or both should be `None`."
-                f' Given {"ctf_fb" if ctf_idx is None else "ctf_idx"}'
+                f' Given {"ctf_basis" if ctf_idx is None else "ctf_idx"}'
             )
-        elif ctf_fb is None:
+        elif ctf_basis is None:
             # Setup defaults for CTF
-            ctf_idx = np.zeros(coeffs.shape[0], dtype=int)
-            ctf_fb = [BlkDiagMatrix.eye_like(covar_coeff)]
+            ctf_idx = np.zeros(coefs.shape[0], dtype=int)
+            ctf_basis = [BlkDiagMatrix.eye_like(covar_coef)]
 
-        noise_covar_coeff = noise_var * BlkDiagMatrix.eye_like(covar_coeff)
+        noise_covar_coef = noise_var * BlkDiagMatrix.eye_like(covar_coef)
 
-        coeffs_est = np.zeros_like(coeffs)
+        coefs_est = np.zeros_like(coefs)
 
         for k in np.unique(ctf_idx[:]):
-            coeff_k = coeffs[ctf_idx == k]
-            ctf_fb_k = ctf_fb[k]
-            ctf_fb_k_t = ctf_fb_k.T
+            coef_k = coefs[ctf_idx == k]
+            ctf_basis_k = ctf_basis[k]
+            ctf_basis_k_t = ctf_basis_k.T
 
-            mean_coeff_k = ctf_fb_k.apply(mean_coeff)
-            coeff_est_k = coeff_k - mean_coeff_k
+            mean_coef_k = ctf_basis_k.apply(mean_coef.asnumpy()[0])
+            coef_est_k = coef_k - mean_coef_k
 
             if noise_var == 0:
-                coeff_est_k = ctf_fb_k.solve(coeff_est_k.T).T
+                coef_est_k = ctf_basis_k.solve(coef_est_k.T).T
             else:
-                sig_covar_coeff = ctf_fb_k @ covar_coeff @ ctf_fb_k_t
-                sig_noise_covar_coeff = sig_covar_coeff + noise_covar_coeff
+                sig_covar_coef = ctf_basis_k @ covar_coef @ ctf_basis_k_t
+                sig_noise_covar_coef = sig_covar_coef + noise_covar_coef
 
-                coeff_est_k = sig_noise_covar_coeff.solve(coeff_est_k.T).T
-                coeff_est_k = (covar_coeff @ ctf_fb_k_t).apply(coeff_est_k.T).T
+                coef_est_k = sig_noise_covar_coef.solve(coef_est_k.T).T
+                coef_est_k = (covar_coef @ ctf_basis_k_t).apply(coef_est_k.T).T
 
-            coeff_est_k = coeff_est_k + mean_coeff
-            coeffs_est[ctf_idx == k] = coeff_est_k
+            coef_est_k = coef_est_k + mean_coef
+            coefs_est[ctf_idx == k] = coef_est_k
 
-        return coeffs_est
+        return Coef(self.basis, coefs_est)
