@@ -1,8 +1,26 @@
 """
+Contains code supporting CTF parameter estimation.
+Generally, this is a port of ASPIRE-CTF from MATLAB.
+
+See paper:
+
+  |    "Reducing bias and variance for CTF estimation in single particle cryo-EM"
+  |    Ayelet Heimowitz, Joakim Andén, Amit Singer
+  |    Ultramicroscopy, Volume 212, 2020
+  |    https://doi.org/10.1016/j.ultramic.2020.112950.
+
+Note:
+``CtfEstimator`` computes the background as a monotonically decreasing
+function of spatial frequency. This practice may lead to an inaccurate
+background estimation for experimental images produced using a K2
+camera in counting mode, as the background in this case is not
+monotonically decreasing. Despite this, CTF parameters are captured
+successfully in such situations.
+
 Created on Sep 10, 2019
 @author: Ayelet Heimowitz, Amit Moscovich
 
-Integrated into ASPIRE by Garrett Wright Feb 2021.
+Integrated into ASPIRE-Python by Garrett Wright Feb 2021.
 """
 
 import logging
@@ -12,16 +30,14 @@ from collections import OrderedDict
 import mrcfile
 import numpy as np
 from numpy import linalg as npla
-from pandas import DataFrame
 from scipy.optimize import linprog
 from scipy.signal.windows import dpss
 
-from aspire.basis.ffb_2d import FFBBasis2D
+from aspire.basis import Coef, FFBBasis2D
 from aspire.image import Image
 from aspire.numeric import fft
-from aspire.operators import voltage_to_wavelength
 from aspire.storage import StarFile
-from aspire.utils import abs2, complex_type, grid_1d, grid_2d
+from aspire.utils import abs2, complex_type, grid_1d, grid_2d, voltage_to_wavelength
 
 logger = logging.getLogger(__name__)
 
@@ -44,7 +60,7 @@ class CtfEstimator:
         """
         Instantiate a CtfEstimator instance.
 
-        :param pixel_size: Size of the pixel in \u212b (Angstrom).
+        :param pixel_size: Size of the pixel in \u212b (angstrom).
         :param cs: Spherical aberration in mm.
         :param amplitude_contrast: Amplitude contrast.
         :param voltage: Voltage of electron microscope.
@@ -59,7 +75,7 @@ class CtfEstimator:
         self.voltage = voltage
         self.psd_size = psd_size
         self.num_tapers = num_tapers
-        self.lmbd = voltage_to_wavelength(voltage) / 10.0  # (Angstrom)
+        self.lmbd = voltage_to_wavelength(voltage) / 10.0  # Convert angstrom to nm
         self.dtype = np.dtype(dtype)
 
         grid = grid_2d(psd_size, normalized=True, indexing="yx", dtype=self.dtype)
@@ -144,8 +160,14 @@ class CtfEstimator:
         # verify block_size is even
         assert block_size % 2 == 0
 
-        size_x = micrograph.shape[1]
-        size_y = micrograph.shape[0]
+        if micrograph.ndim == 3:
+            assert (
+                micrograph.shape[0] == 1
+            ), f"micrograph should be 2D or stack of 1 2D image: {micrograph.shape}"
+            micrograph = micrograph[0]
+
+        size_x = micrograph.shape[-1]
+        size_y = micrograph.shape[-2]
 
         step_size = block_size // 2
         range_y = size_y // step_size - 1
@@ -262,34 +284,33 @@ class CtfEstimator:
         """
 
         # RCOPT, come back and change the indices for this method
-        coeffs_s = ffbbasis.evaluate_t(amplitude_spectrum).T
-        coeffs_n = coeffs_s.copy()
+        coefs_s = ffbbasis.evaluate_t(amplitude_spectrum).asnumpy().copy().T
+        coefs_n = coefs_s.copy()
 
-        coeffs_s[np.argwhere(ffbbasis._indices["ells"] == 1)] = 0
+        coefs_s[np.argwhere(ffbbasis.angular_indices == 1)] = 0
         if circular:
-            coeffs_s[np.argwhere(ffbbasis._indices["ells"] == 2)] = 0
+            coefs_s[np.argwhere(ffbbasis.angular_indices == 2)] = 0
             noise = amplitude_spectrum
         else:
-            coeffs_n[np.argwhere(ffbbasis._indices["ells"] == 0)] = 0
-            coeffs_n[np.argwhere(ffbbasis._indices["ells"] == 2)] = 0
-            noise = ffbbasis.evaluate(coeffs_n.T)
+            coefs_n[np.argwhere(ffbbasis.angular_indices == 0)] = 0
+            coefs_n[np.argwhere(ffbbasis.angular_indices == 2)] = 0
+            noise = Coef(ffbbasis, coefs_n.T).evaluate()
 
-        psd = ffbbasis.evaluate(coeffs_s.T)
+        psd = Coef(ffbbasis, coefs_s.T).evaluate()
 
         return psd, noise
 
     def background_subtract_1d(
-        self, amplitude_spectrum, linprog_method="interior-point", n_low_freq_cutoffs=14
+        self, amplitude_spectrum, linprog_method="highs", n_low_freq_cutoffs=14
     ):
         """
         Estimate and subtract the background from the power spectrum
 
         :param amplitude_spectrum: Estimated power spectrum
-        :param linprog_method: Method passed to linear progam solver (scipy.optimize.linprog).  Defaults to 'interior-point'.
+        :param linprog_method: Method passed to linear program solver (scipy.optimize.linprog).
         :param n_low_freq_cutoffs: Low frequency cutoffs (loop iterations).
         :return: 2-tuple of NumPy arrays (PSD after noise subtraction and estimated noise)
         """
-
         # compute radial average
         center = amplitude_spectrum.shape[-1] // 2
 
@@ -304,7 +325,7 @@ class CtfEstimator:
                 f"Invalid ndimension for amplitude_spectrum {amplitude_spectrum.shape}"
             )
 
-        amplitude_spectrum = amplitude_spectrum[center, center:]
+        amplitude_spectrum = amplitude_spectrum.asnumpy()[0, center, center:]
         amplitude_spectrum = amplitude_spectrum[
             0 : 3 * amplitude_spectrum.shape[-1] // 4
         ]
@@ -352,20 +373,22 @@ class CtfEstimator:
                 axis=0,
             )
 
-            x_bound_lst = [
-                (signal[i], signal[i], -1 * np.inf, np.inf)
-                for i in range(signal.shape[0])
-            ]
-            x_bound = np.asarray(x_bound_lst, A.dtype)
-            x_bound = np.concatenate((x_bound[:, :2], x_bound[:, 2:]), axis=0)
+            # The original code used `bounds`,
+            #   but for many problems, linprog reports infeasable constraints.
+            # In practice for a micrograph from the paper, and our tutorial,
+            #   the code seems to work better without it...
+            # ASPIRE #417
 
             x = linprog(
                 f,
                 A_ub=A,
                 b_ub=np.zeros(A.shape[0]),
-                bounds=x_bound,
                 method=linprog_method,
             )
+
+            if not x.success:
+                raise RuntimeError("Linear program did not succeed. Halting")
+
             background = x.x[N:]
 
             bs_psd = signal - background
@@ -391,9 +414,9 @@ class CtfEstimator:
         Find optimal defocus for the radially symmetric case (where no astigmatism is present)
 
         :param amplitude_spectrum: Estimated power specrtum.
-        :param pixel_size: Pixel size in \u212b (Angstrom).
+        :param pixel_size: Pixel size in \u212b (angstrom).
         :param cs: Spherical aberration in mm.
-        :param lmbd: Electron wavelength \u212b (Angstrom).
+        :param lmbd: Electron wavelength \u212b (angstrom).
         :param w: Amplitude contrast.
         :param N: Number of rows (or columns) in the estimate power spectrum.
         :param min_defocus: Start of defocus loop scan.
@@ -404,7 +427,7 @@ class CtfEstimator:
         center = N // 2
 
         grid = grid_1d(N, normalized=True, dtype=self.dtype)
-        rb = grid["x"][center:] / 2
+        rb = grid["r"][center:] / 2
 
         r_ctf = rb * (10 / pixel_size)  # units: inverse nm
 
@@ -476,7 +499,7 @@ class CtfEstimator:
         """
 
         :param signal: Estimated power spectrum.
-        :param pixel_size: Pixel size in \u212b (Angstrom).
+        :param pixel_size: Pixel size in \u212b (angstrom).
         :param g_min: Inverse of minimun resolution for PSD.
         :param g_max: Inverse of maximum resolution for PSD.
         :return: ratio.
@@ -496,7 +519,7 @@ class CtfEstimator:
         X = grid["x"]
         Y = grid["y"]
 
-        signal -= np.min(signal)
+        signal = signal - np.min(signal)
 
         rad_sq_min = N * pixel_size / g_min
         rad_sq_max = N * pixel_size / g_max
@@ -551,11 +574,11 @@ class CtfEstimator:
         :param angle_ast: Angle between df1 and the x-axis, Radians.
         :param r: Magnitude of spatial frequencies.
         :param theta: Phase of spatial frequencies.
-        :param pixel_size: Pixel size in \u212b (Angstrom).
+        :param pixel_size: Pixel size in \u212b (angstrom).
         :param g_min: Inverse of minimun resolution for PSD.
         :param g_max: Inverse of maximum resolution for PSD.
         :param amplitude_contrast: Amplitude contrast.
-        :param lmbd: Electron wavelength \u212b (Angstrom).
+        :param lmbd: Electron wavelength \u212b (angstrom).
         :param cs: Spherical aberration in mm.
         :return: Optimal defocus parameters
         """
@@ -678,16 +701,17 @@ class CtfEstimator:
         """
         data_block = {}
         data_block["_rlnMicrographName"] = name
-        data_block["_rlnDefocusU"] = params_dict["defocus_u"]
-        data_block["_rlnDefocusV"] = params_dict["defocus_v"]
-        data_block["_rlnDefocusAngle"] = params_dict["defocus_ang"]
+        data_block["_rlnDefocusU"] = params_dict["defocus_u"]  # Should already be in A
+        data_block["_rlnDefocusV"] = params_dict["defocus_v"]  # Should already be in A
+        data_block["_rlnDefocusAngle"] = (
+            params_dict["defocus_ang"] * 180 / np.pi
+        )  # Convert from radian to degree
         data_block["_rlnSphericalAberration"] = params_dict["cs"]
         data_block["_rlnAmplitudeContrast"] = params_dict["amplitude_contrast"]
         data_block["_rlnVoltage"] = params_dict["voltage"]
-        data_block["_rlnDetectorPixelSize"] = params_dict["pixel_size"]
-        df = DataFrame([data_block])
+        data_block["_rlnMicrographPixelSize"] = params_dict["pixel_size"]
         blocks = OrderedDict()
-        blocks["root"] = df
+        blocks["root"] = data_block
         star = StarFile(blocks=blocks)
         star.write(os.path.join(output_dir, os.path.splitext(name)[0]) + ".star")
 
@@ -726,7 +750,7 @@ def estimate_ctf(
         amplitude_contrast / np.sqrt(1 - amplitude_contrast**2)
     )
 
-    lmbd = voltage_to_wavelength(voltage) / 10  # (Angstrom)
+    lmbd = voltage_to_wavelength(voltage) / 10  # Convert from angstrom to nm
 
     ctf_object = CtfEstimator(
         pixel_size, cs, amplitude_contrast, voltage, psd_size, num_tapers, dtype=dtype
@@ -758,15 +782,17 @@ def estimate_ctf(
 
         # Optionally changing to: linprog_method='simplex',
         # will more deterministically repro results in exchange for speed.
+        # linprog_method was changed from 'interior-point' to 'highs' due to
+        # "interior-point' being deprecated.
         signal_1d, background_1d = ctf_object.background_subtract_1d(
-            amplitude_spectrum, linprog_method="interior-point"
+            amplitude_spectrum, linprog_method="highs"
         )
 
         avg_defocus, low_freq_skip = ctf_object.opt1d(
             signal_1d,
             pixel_size,
             cs,
-            lmbd,  # (Angstrom)
+            lmbd,  # (angstrom)
             amplitude_contrast,
             signal_observed.shape[-1],
         )
@@ -806,7 +832,7 @@ def estimate_ctf(
                 g_min,
                 g_max,
                 amplitude_contrast,
-                lmbd,  # (Angstrom)
+                lmbd,  # (angstrom)
                 cs,
             )
 
@@ -817,8 +843,8 @@ def estimate_ctf(
         ml = np.argmax(cc_array[:, 3], -1)
 
         result = {
-            "defocus_u": cc_array[ml, 0],
-            "defocus_v": cc_array[ml, 1],
+            "defocus_u": cc_array[ml, 0] * 10,  # Convert from nm to A
+            "defocus_v": cc_array[ml, 1] * 10,  # Convert from nm to A
             "defocus_ang": cc_array[ml, 2],  # Radians
             "cs": cs,
             "voltage": voltage,
@@ -837,12 +863,12 @@ def estimate_ctf(
                 os.path.join(output_dir, os.path.splitext(name)[0] + "_noise.mrc"),
                 overwrite=True,
             ) as mrc:
-                mrc.set_data(background_2d[0].astype(np.float32))
+                mrc.set_data(background_2d.asnumpy()[0].astype(np.float32))
                 mrc.voxel_size = pixel_size
 
         if save_ctf_images:
-            ctf_object.set_df1(cc_array[ml, 0])
-            ctf_object.set_df2(cc_array[ml, 1])
+            ctf_object.set_df1(cc_array[ml, 0])  # angstrom
+            ctf_object.set_df2(cc_array[ml, 1])  # angstrom
             ctf_object.set_angle(cc_array[ml, 2])  # Radians
             ctf_object.generate_ctf()
             df = (cc_array[ml, 0] + cc_array[ml, 1]) * np.ones(
@@ -856,7 +882,7 @@ def estimate_ctf(
             )
             ctf_signal = np.zeros(ctf_im.shape, ctf_im.dtype)
             ctf_signal[: ctf_im.shape[0] // 2, :] = ctf_im[: ctf_im.shape[0] // 2, :]
-            ctf_signal[ctf_im.shape[0] // 2 + 1 :, :] = signal[
+            ctf_signal[ctf_im.shape[0] // 2 + 1 :, :] = signal.asnumpy()[
                 :, :, ctf_im.shape[0] // 2 + 1
             ]
 
