@@ -1,16 +1,16 @@
+import abc
 import logging
 
 import numpy as np
 from joblib import Memory
 
 from aspire.image import Image
-from aspire.operators import PowerFilter, ZeroFilter
-from aspire.utils.random import randn
+from aspire.utils import crop_pad_2d
 
 logger = logging.getLogger(__name__)
 
 
-class Xform:
+class Xform(abc.ABC):
     """
     An Xform is anything that implements a `forward` method (and an `adjoint` method, in the case of a LinearXform),
     that takes in a square Image object and spits out a square Image object corresponding to forward/adjoint operations.
@@ -68,14 +68,16 @@ class Xform:
             indices = np.arange(im.n_images)
         return self._forward(im, indices=indices)
 
+    @abc.abstractmethod
     def _forward(self, im, indices):
-        raise NotImplementedError(
-            "Subclasses must implement the _forward method applicable to im/indices."
-        )
+        """
+        Subclasses must implement the _forward method applicable to im/indices.
+        """
 
     def enabled(self):
         """
         Enable this Xform in a context manager, regardless of its `active` attribute value.
+
         :return: A context manager in which this Xform is enabled.
         """
         return Xform.XformActiveContextManager(self, active=True)
@@ -83,6 +85,7 @@ class Xform:
     def disabled(self):
         """
         Disable this Xform in a context manager, regardless of its `active` attribute value.
+
         :return: A context manager in which this Xform is disabled.
         """
         return Xform.XformActiveContextManager(self, active=False)
@@ -110,10 +113,11 @@ class LinearXform(Xform):
             indices = np.arange(im.n_images)
         return self._adjoint(im, indices=indices)
 
+    @abc.abstractmethod
     def _adjoint(self, im, indices):
-        raise NotImplementedError(
-            "Subclasses must implement the _adjoint method applicable to im/indices."
-        )
+        """
+        Subclasses must implement the _adjoint method applicable to im/indices.
+        """
 
 
 class SymmetricXform(LinearXform):
@@ -142,7 +146,7 @@ class Multiply(SymmetricXform):
 
     def _forward(self, im, indices):
         if self.multipliers.size == 1:  # if we have a scalar multiplier
-            im_new = im * self.multipliers
+            im_new = im * self.multipliers.astype(im.dtype)
         else:
             im_new = im * self.multipliers[indices]
 
@@ -196,19 +200,98 @@ class Downsample(LinearXform):
     A Xform that downsamples an Image object to a resolution specified by this Xform's resolution.
     """
 
-    def __init__(self, resolution):
+    def __init__(self, resolution, zero_nyquist=True, centered_fft=True):
+        """
+        Initialize Xform to downsample Image to a specific resolution.
+
+        :param resolution: int - new resolution, should be <= the current resolution
+            of this Image
+        :param zero_nyquist: Option to keep or remove Nyquist frequency for even
+            resolution (boolean). Defaults to zero_nyquist=True, removing the Nyquist frequency.
+        :param centered_fft: Default of True uses `centered_fft` to
+            maintain ASPIRE-Python centering conventions.
+        """
         self.resolution = resolution
+        self.zero_nyquist = zero_nyquist
+        self.centered_fft = centered_fft
         super().__init__()
 
     def _forward(self, im, indices):
-        return im.downsample(self.resolution)
+        original_stack_shape = im.stack_shape
+        data = im.stack_reshape(-1)._data
+        im_ds = Image._downsample(
+            data,
+            self.resolution,
+            zero_nyquist=self.zero_nyquist,
+            centered_fft=self.centered_fft,
+        )
+
+        # pixel_size has already been adjusted in the ImageSource and passed
+        # to `im`, so we instantiate the new Image with im.pixel_size.
+        return Image(im_ds, pixel_size=im.pixel_size).stack_reshape(
+            original_stack_shape
+        )
 
     def _adjoint(self, im, indices):
         # TODO: Implement up-sampling with zero-padding
         raise NotImplementedError("Adjoint of downsampling not implemented yet.")
 
     def __str__(self):
-        return f"Downsample (Resolution {self.resolution})"
+        return f"Downsample (resolution={self.resolution}, zero_nyquist={self.zero_nyquist}, centered_fft={self.centered_fft}) Xform"
+
+
+class CropPad(Xform):
+    """
+    A Xform that crops or pads an Image object to a specified size.
+    """
+
+    def __init__(self, L, fill_value=0):
+        """
+        Initialize Xform to crop Image to a specific size.
+
+        :param L: int - new size
+        :param fill_value: Optional value for padding, default 0.
+        """
+        self.L = L
+        self.fill_value = fill_value
+        super().__init__()
+
+    def _forward(self, im, indices):
+        return crop_pad_2d(im, self.L, self.fill_value)
+
+    def __str__(self):
+        return f"CropPad({self.L}, {self.fill_value}) Xform"
+
+
+class LegacyWhiten(Xform):
+    """
+    A Xform that implements MATLAB legacy whitening.
+    """
+
+    def __init__(self, psd, delta):
+        """
+        Initialize LegacyWhiten Xform.
+
+        :param psd: PSD (as computed by LegacyNoiseEstimator).
+        :param delta: Threshold used to determine which frequencies to whiten
+            and which to set to zero. By default all `sqrt(psd)` values
+            less than `delta` are zeroed out in the whitening filter.
+        """
+
+        self.psd = psd
+        self.delta = delta
+        super().__init__()
+
+    def _forward(self, im, indices):
+        """
+        Apply the legacy MATLAB whitening transformation to `im`.
+
+        The tranform is applied to all of `im`, `indices` are unused.
+        """
+        return im.legacy_whiten(self.psd, self.delta)
+
+    def __str__(self):
+        return "LegacyWhiten() Xform"
 
 
 class FilterXform(SymmetricXform):
@@ -289,37 +372,6 @@ class LambdaXform(Xform):
         return f"LambdaXform ({self.lambda_fun.__name__})"
 
 
-class NoiseAdder(Xform):
-    """
-    A Xform that adds white noise, optionally passed through a Filter object, to all incoming images.
-    """
-
-    def __init__(self, seed=0, noise_filter=None):
-        """
-        Initialize the random state of this NoiseAdder using specified values.
-
-        :param seed: The random seed used to generate white noise
-        :param noise_filter: An optional aspire.operators.Filter object to use to filter the generated white noise.
-            By default, a ZeroFilter is used, generating no noise.
-        """
-        super().__init__()
-        self.seed = seed
-        noise_filter = noise_filter or ZeroFilter()
-        self.noise_filter = PowerFilter(noise_filter, power=0.5)
-
-    def _forward(self, im, indices):
-        im = im.copy()
-
-        for i, idx in enumerate(indices):
-            # Note: The following random seed behavior is directly taken from MATLAB Cov3D code.
-            random_seed = self.seed + 191 * (idx + 1)
-            im_s = randn(2 * im.res, 2 * im.res, seed=random_seed)
-            im_s = Image(im_s).filter(self.noise_filter)[0]
-            im[i] += im_s[: im.res, : im.res]
-
-        return im
-
-
 class IndexedXform(Xform):
     """
     An IndexedXform is a Xform where individual Xform objects are used at specific indices of the incoming Image object.
@@ -378,11 +430,9 @@ class IndexedXform(Xform):
             # Apply the transformation to the selected indices in the Image object
             if len(im_data_indices) > 0:
                 fn_handle = getattr(xform, which)
-                im_data[im_data_indices] = fn_handle(
-                    Image(im[im_data_indices])
-                ).asnumpy()
+                im_data[im_data_indices] = fn_handle(im[im_data_indices]).asnumpy()
 
-        return Image(im_data)
+        return Image(im_data, pixel_size=im.pixel_size)
 
     def _forward(self, im, indices):
         return self._indexed_operation(im, indices, "forward")
@@ -399,10 +449,10 @@ def _apply_xform(xform, im, indices, adjoint=False):
     method.
     """
     if not adjoint:
-        logger.info("  Applying " + str(xform))
+        logger.debug("  Applying " + str(xform))
         return xform.forward(im, indices=indices)
     else:
-        logger.info("  Applying Adjoint " + str(xform))
+        logger.debug("  Applying Adjoint " + str(xform))
         return xform.adjoint(im, indices=indices)
 
 
@@ -463,10 +513,10 @@ class Pipeline(Xform):
         memory = Memory(location=self.memory, verbose=0)
         _apply_transform_cached = memory.cache(_apply_xform)
 
-        logger.info("Applying forward transformations in pipeline")
+        logger.debug("Applying forward transformations in pipeline")
         for xform in self.xforms:
             im = _apply_transform_cached(xform, im, indices, False)
-        logger.info("All forward transformations applied")
+        logger.debug("All forward transformations applied")
 
         return im
 
@@ -476,9 +526,9 @@ class LinearPipeline(Pipeline, LinearXform):
         memory = Memory(location=self.memory, verbose=0)
         _apply_transform_cached = memory.cache(_apply_xform)
 
-        logger.info("Applying adjoint transformations in pipeline")
+        logger.debug("Applying adjoint transformations in pipeline")
         for xform in self.xforms[::-1]:
             im = _apply_transform_cached(xform, im, indices, True)
-        logger.info("All adjoint transformations applied")
+        logger.debug("All adjoint transformations applied")
 
         return im
