@@ -3,6 +3,9 @@ Define a BlkDiagMatrix module which implements operations for
 block diagonal matrices as used by ASPIRE.
 """
 
+import logging
+import warnings
+
 import numpy as np
 from numpy.linalg import norm, solve
 from scipy.linalg import block_diag
@@ -10,6 +13,28 @@ from scipy.linalg import block_diag
 from aspire.utils import make_psd
 from aspire.utils.cell import Cell2D
 
+logger = logging.getLogger(__name__)
+
+
+def is_scalar_type(x):
+    """
+    Helper function checking scalar-ness for elementwise ops.
+
+    Essentially we are checking for a single numeric object, as opposed to
+    something like an `ndarray` or `BlkDiagMatrix`. We do this by
+    checking `numpy.isscalar(x)`.
+
+    In the future this check may require extension to include ASPIRE or
+    other third party types beyond what is provided by numpy, so we
+    implement it now as a method.
+
+    :param x: Value to check
+
+    :return: bool.
+    """
+
+    return np.isscalar(x)
+
 
 class BlkDiagMatrix:
     """
@@ -66,7 +91,7 @@ class BlkDiagMatrix:
         :param blk: Block to append (ndarray).
         """
 
-        self.data.append(blk)
+        self.data.append(blk.astype(self.dtype, copy=False))
         self.nblocks += 1
         self.reset_cache()
 
@@ -105,7 +130,7 @@ class BlkDiagMatrix:
         Convenience wrapper, setter on self.data.
         """
 
-        self.data[key] = value
+        self.data[key] = value.astype(self.dtype, copy=False)
         self.reset_cache()
 
     def __len__(self):
@@ -115,25 +140,6 @@ class BlkDiagMatrix:
 
         return self.nblocks
 
-    def _is_scalar_type(self, x):
-        """
-        Internal helper function checking scalar-ness for elementwise ops.
-
-        Essentially we are checking for a single numeric object, as opposed to
-        something like an `ndarray` or `BlkDiagMatrix`. We do this by
-        checking `numpy.isscalar(x)`.
-
-        In the future this check may require extension to include ASPIRE or
-        other third party types beyond what is provided by numpy, so we
-        implement it now as a class method.
-
-        :param x: Value to check
-
-        :return: bool.
-        """
-
-        return np.isscalar(x)
-
     def __check_size_compatible_add(self, other):
         """
         Sanity check two BlkDiagMatrix instances are compatible in size
@@ -246,8 +252,10 @@ class BlkDiagMatrix:
             to self + other.
         """
 
-        if self._is_scalar_type(other):
+        if is_scalar_type(other):
             return self.__scalar_add(other, inplace=inplace)
+        elif not isinstance(other, BlkDiagMatrix):
+            return NotImplemented
 
         self.__check_compatible(other)
 
@@ -296,7 +304,7 @@ class BlkDiagMatrix:
             to self + other.
         """
 
-        assert self._is_scalar_type(scalar)
+        assert is_scalar_type(scalar)
 
         if inplace:
             C = self
@@ -319,8 +327,10 @@ class BlkDiagMatrix:
             self - other.
         """
 
-        if self._is_scalar_type(other):
+        if is_scalar_type(other):
             return self.__scalar_sub(other, inplace=inplace)
+        elif not isinstance(other, BlkDiagMatrix):
+            return NotImplemented
 
         self.__check_compatible(other)
 
@@ -349,7 +359,7 @@ class BlkDiagMatrix:
         Operator overloading for in-place subtraction.
         """
 
-        if self._is_scalar_type(other):
+        if is_scalar_type(other):
             return self.__scalar_sub(other, inplace=True)
 
         return self.sub(other, inplace=True)
@@ -375,7 +385,7 @@ class BlkDiagMatrix:
             self + other.
         """
 
-        assert self._is_scalar_type(scalar)
+        assert is_scalar_type(scalar)
 
         if inplace:
             C = self
@@ -397,12 +407,16 @@ class BlkDiagMatrix:
         :return: A BlkDiagMatrix of self @ other.
         """
 
-        if not isinstance(other, BlkDiagMatrix):
-            if inplace:
-                raise RuntimeError(
-                    "`inplace` method not supported when "
-                    "mixing `BlkDiagMatrix` and `Numpy`."
-                )
+        if not isinstance(other, BlkDiagMatrix) and inplace:
+            raise RuntimeError(
+                "`inplace` method not supported when " "mixing `BlkDiagMatrix`."
+            )
+
+        # Convert operator to BlkDiagMatrix is possible.
+        # Its possible to optimize this in the future if its too slow.
+        if hasattr(other, "as_blk_diag"):
+            other = other.as_blk_diag(self.partition)
+        elif not isinstance(other, BlkDiagMatrix):
             return self.apply(other)
 
         self.__check_compatible(other, size_compat="mul")
@@ -464,28 +478,22 @@ class BlkDiagMatrix:
         :return: A BlkDiagMatrix of self * other.
         """
 
-        if isinstance(val, BlkDiagMatrix):
-            raise RuntimeError(
-                "Attempt numeric multiplication (*,mul) of two "
-                "BlkDiagMatrix instances, try (matmul,@)."
-            )
-
-        elif not self._is_scalar_type(val):
-            raise RuntimeError(
-                "Attempt numeric multiplication (*,mul) of a "
-                "BlkDiagMatrix and {}.".format(type(val))
-            )
+        # Convert scalar to reduce code branching.
+        if is_scalar_type(val):
+            val = np.full((self.nblocks), fill_value=val, dtype=self.dtype)
+        elif not isinstance(val, BlkDiagMatrix):
+            raise NotImplementedError(f"mul not implemented for {type(val)}.")
 
         if inplace:
             for i in range(self.nblocks):
-                self[i] *= val
+                self[i] *= val[i]
 
             C = self
         else:
             C = BlkDiagMatrix(self.partition, dtype=self.dtype)
 
             for i in range(self.nblocks):
-                C[i] = self[i] * val
+                C[i] = self[i] * val[i]
 
         return C
 
@@ -703,14 +711,14 @@ class BlkDiagMatrix:
 
         cols = self.partition[:, 1]
 
-        if np.sum(cols) != np.size(X, 0):
-            raise RuntimeError("Sizes of matrix `self` and `X` are not compatible.")
-
         vector = False
         if np.ndim(X) == 1:
             X = X[:, np.newaxis]
             vector = True
 
+        if np.sum(cols) != np.size(X, 0):
+            raise RuntimeError("Sizes of matrix `self` and `X` are not compatible.")
+
         rows = np.array(
             [
                 np.size(X, 1),
@@ -751,6 +759,7 @@ class BlkDiagMatrix:
     def eigvals(self):
         """
         Compute the eigenvalues of a BlkDiagMatrix.
+
         :return: Array of eigvals, with length equal to the fully expanded matrix diagonal.
 
         """
@@ -762,7 +771,7 @@ class BlkDiagMatrix:
 
         :return: True if all blocks have non-negative eigenvalues.
         """
-        return np.alltrue(self.eigvals() > 0.0)
+        return np.all(self.eigvals() > 0.0)
 
     def make_psd(self):
         """
@@ -923,3 +932,59 @@ class BlkDiagMatrix:
             A.data[i] = np.array(blk_diag[i], dtype=dtype)
 
         return A
+
+    def diag(self):
+        """
+        Return the diagonal elements of this `BlkDiagMatrix`.
+        """
+
+        # Avoid circular import
+        from .diag_matrix import DiagMatrix
+
+        diag = []
+        for blk in self:
+            diag.extend(list(np.diag(blk)))
+
+        return DiagMatrix(np.array(diag, dtype=self.dtype))
+
+    @staticmethod
+    def from_dense(A, blk_partition, warn_eps=1e-3):
+        """
+        Create BlkDiagMatrix with `blk_partition` from dense matrix `A`.
+
+        :param A: Dense `Numpy` array.
+        :param blk_partition: List of block partition shapes.
+        :param warn_eps: Optionally warn if off block values from `A`
+            exceed `warn_eps`.  `None` disables warnings.
+        :return: `BlkDiagMatrix` with values from `A`.
+        """
+
+        # Instantiate an empty BlkDiagMatrix with `blk_partition`
+        B = BlkDiagMatrix.zeros(blk_partition, dtype=A.dtype)
+
+        # Set the data
+        inds = np.array([0, 0])
+        for i in range(B.nblocks):
+            ends = inds + B.partition[i]
+            B[i][:, :] = A[inds[0] : ends[0], inds[1] : ends[1]]
+            inds = ends
+
+        # We should reach exactly the end of A when partition was correct
+        if not np.all(inds == A.shape):
+            raise RuntimeError(
+                "Block partition appears to mismatch shape of dense matrix A."
+            )
+
+        if warn_eps is not None:
+            max_diff = np.max(np.abs((A - B.dense())))
+            if max_diff > warn_eps:
+                # Warn (once)
+                warnings.warn(
+                    f"BlkDiagMatrix.from_dense truncating values exceeding {warn_eps}",
+                    UserWarning,
+                    stacklevel=2,
+                )
+                # Log the specifics for debugging
+                logger.debug(f"BlkDiagMatrix.from_dense truncated max value {max_diff}")
+
+        return B
