@@ -1,58 +1,44 @@
 import inspect
 import logging
-import math
+from functools import lru_cache
 
 import numpy as np
 from scipy.interpolate import RegularGridInterpolator
 
-from aspire.utils import grid_2d
-from aspire.utils.filter_to_fb_mat import filter_to_fb_mat
+from aspire import config
+from aspire.utils import cart2pol, grid_2d, voltage_to_wavelength
 
 logger = logging.getLogger(__name__)
 
 
-def voltage_to_wavelength(voltage):
-    """
-    Convert from electron voltage to wavelength.
-
-    :param voltage: float, The electron voltage in kV.
-    :return: float, The electron wavelength in nm.
-    """
-    return 12.2643247 / math.sqrt(voltage * 1e3 + 0.978466 * voltage**2)
-
-
-def wavelength_to_voltage(wavelength):
-    """
-    Convert from electron voltage to wavelength.
-
-    :param wavelength: float, The electron wavelength in nm.
-    :return: float, The electron voltage in kV.
-    """
-    return (
-        -1e3 + math.sqrt(1e6 + 4 * 12.2643247**2 * 0.978466 / wavelength**2)
-    ) / (2 * 0.978466)
-
-
-def evaluate_src_filters_on_grid(src):
+def evaluate_src_filters_on_grid(src, indices=None):
     """
     Given an ImageSource object, compute the source's unique filters
     at the filter_indices specified in its metadata.
 
+    :param src: Source instance
+    :param indices: Optional, subset of src indices to compute.
+        Defaults to the entire src.
+
     :return: an `src.L x src.L x len(src.filter_indices)`
         array containing the evaluated filters at each gridpoint
     """
 
+    if indices is None:
+        indices = np.arange(src.n, dtype=int)
+
     grid2d = grid_2d(src.L, indexing="yx", dtype=src.dtype)
     omega = np.pi * np.vstack((grid2d["x"].flatten(), grid2d["y"].flatten()))
 
-    h = np.empty((omega.shape[-1], len(src.filter_indices)), dtype=src.dtype)
+    # Initialize h as ones to mimic an IdentityFilter when src.unique_filters is None.
+    h = np.ones((omega.shape[-1], len(indices)), dtype=src.dtype)
     for i, filt in enumerate(src.unique_filters):
-        idx_k = np.where(src.filter_indices == i)[0]
+        idx_k = np.where(src.filter_indices[indices] == i)[0]
         if len(idx_k) > 0:
-            filter_values = filt.evaluate(omega)
+            filter_values = filt.evaluate(omega, pixel_size=src.pixel_size)
             h[:, idx_k] = np.column_stack((filter_values,) * len(idx_k))
 
-    h = np.reshape(h, grid2d["x"].shape + (len(src.filter_indices),))
+    h = np.reshape(h, grid2d["x"].shape + (len(indices),))
 
     return h
 
@@ -69,11 +55,12 @@ class Filter:
     def __str__(self):
         """
         Show class name of Filter
+
         :return: A string of class name
         """
         return self.__class__.__name__
 
-    def evaluate(self, omega):
+    def evaluate(self, omega, **kwargs):
         """
         Evaluate the filter at specified frequencies.
 
@@ -85,7 +72,9 @@ class Filter:
         if omega.ndim == 1:
             assert self.radial, "Cannot evaluate a non-radial filter on 1D input array."
         elif omega.ndim == 2 and self.dim:
-            assert omega.shape[0] == self.dim, f"Omega must be of size {self.dim} x n"
+            assert (
+                omega.shape[0] == self.dim
+            ), f"Omega must be of size {self.dim} x n; Passed omega.shape {omega.shape}"
 
         if self.radial:
             if omega.ndim > 1:
@@ -93,21 +82,25 @@ class Filter:
             omega, idx = np.unique(omega, return_inverse=True)
             omega = np.vstack((omega, np.zeros_like(omega)))
 
-        h = self._evaluate(omega)
+        h = self._evaluate(omega, **kwargs)
 
         if self.radial:
             h = np.take(h, idx)
 
         return h
 
-    def _evaluate(self, omega):
+    def _evaluate(self, omega, **kwargs):
         raise NotImplementedError("Subclasses should implement this method")
 
-    def fb_mat(self, fbasis):
+    def basis_mat(self, basis, **kwargs):
         """
-        Represent the filter in FB basis matrix
+        Represent the filter in `basis`.
+
+        :param basis: 2D Basis.
+        :return: `basis` representation of this filter.
+            Return type will depend on `basis`.
         """
-        return filter_to_fb_mat(self.evaluate, fbasis)
+        return basis.filter_to_basis_mat(self, **kwargs)
 
     def scale(self, c=1):
         """
@@ -119,7 +112,8 @@ class Filter:
         """
         return ScaledFilter(self, c)
 
-    def evaluate_grid(self, L, dtype=np.float32, *args, **kwargs):
+    @lru_cache(maxsize=config["cache"]["filter_cache_size"].get())  # noqa: B019
+    def evaluate_grid(self, L, *args, dtype=np.float32, **kwargs):
         """
         Generates a two dimensional grid with prescribed dtype,
         yielding the values (omega) which are then evaluated by
@@ -132,7 +126,6 @@ class Filter:
         :return: Filter values at omega's points.
         """
 
-        # Note we can probably unwind the "F"/m_reshape here
         grid2d = grid_2d(L, indexing="yx", dtype=dtype)
         omega = np.pi * np.vstack((grid2d["x"].flatten(), grid2d["y"].flatten()))
         h = self.evaluate(omega, *args, **kwargs)
@@ -161,8 +154,8 @@ class DualFilter(Filter):
         self._filter = filter_in
         super().__init__()
 
-    def evaluate(self, omega):
-        return self._filter.evaluate(-omega)
+    def evaluate(self, omega, **kwargs):
+        return self._filter.evaluate(-omega, **kwargs)
 
 
 class FunctionFilter(Filter):
@@ -184,7 +177,8 @@ class FunctionFilter(Filter):
         # (i.e. at runtime, we will still expect the incoming omega values to have x and y components).
         super().__init__(dim=dim, radial=dim > n_args)
 
-    def _evaluate(self, omega):
+    def _evaluate(self, omega, **kwargs):
+        # Note kwargs are not used here, this might be trouble
         return self.f(*omega)
 
 
@@ -193,15 +187,26 @@ class PowerFilter(Filter):
     A Filter object that is composed of a regular `Filter` object, but evaluates it to a specified power.
     """
 
-    def __init__(self, filter, power=1):
+    def __init__(self, filter, power=1, epsilon=None):
+        """
+        Initialize PowerFilter instance.
+
+        :param filter: A Filter instance.
+        :param power: Exponent to raise filter values.
+        :param epsilon: Threshold on filter values that get raised to a negative power.
+            `filter` values below this threshold will be set to zero during evaluation.
+            Default uses machine epsilon for filter.dtype.
+        """
         self._filter = filter
         self._power = power
+        self._epsilon = epsilon
         super().__init__(dim=filter.dim, radial=filter.radial)
 
-    def _evaluate(self, omega):
-        return self._filter.evaluate(omega) ** self._power
+    def _evaluate(self, omega, **kwargs):
+        return self._filter.evaluate(omega, **kwargs) ** self._power
 
-    def evaluate_grid(self, L, dtype=np.float32, *args, **kwargs):
+    @lru_cache(maxsize=config["cache"]["filter_cache_size"].get())  # noqa: B019
+    def evaluate_grid(self, L, *args, dtype=np.float32, **kwargs):
         """
         Calls the provided filter's evaluate_grid method in case there is an optimization.
 
@@ -209,10 +214,26 @@ class PowerFilter(Filter):
 
         See `Filter.evaluate_grid` for usage.
         """
+        filter_vals = self._filter.evaluate_grid(L, *args, dtype=dtype, **kwargs)
 
-        return (
-            self._filter.evaluate_grid(L, dtype=dtype, *args, **kwargs) ** self._power
-        )
+        # Place safeguard on values below machine epsilon for negative powers.
+        if self._power < 0:
+            eps = self._epsilon
+            if eps is None:
+                eps = np.finfo(filter_vals.dtype).eps
+                eps = (100 * eps) ** (-1 / self._power)
+            condition = abs(filter_vals) < eps
+            num_less_eps = np.count_nonzero(condition)
+            if num_less_eps > 0:
+                logger.warning(
+                    f"{self} setting {num_less_eps} extremal filter value(s) to zero."
+                )
+
+            filter_vals = np.where(condition, 0, filter_vals**self._power)
+
+            return filter_vals
+
+        return filter_vals**self._power
 
 
 class LambdaFilter(Filter):
@@ -225,8 +246,8 @@ class LambdaFilter(Filter):
         self._f = f
         super().__init__(dim=filter.dim, radial=filter.radial)
 
-    def _evaluate(self, omega):
-        return self._f(self._filter.evaluate(omega))
+    def _evaluate(self, omega, **kwargs):
+        return self._f(self._filter.evaluate(omega, **kwargs))
 
 
 class MultiplicativeFilter(Filter):
@@ -238,10 +259,10 @@ class MultiplicativeFilter(Filter):
         super().__init__(dim=args[0].dim, radial=all(c.radial for c in args))
         self._components = args
 
-    def _evaluate(self, omega):
+    def _evaluate(self, omega, **kwargs):
         res = 1
         for c in self._components:
-            res *= c.evaluate(omega)
+            res *= c.evaluate(omega, **kwargs)
         return res
 
 
@@ -255,8 +276,8 @@ class ScaledFilter(Filter):
         self._scale = scale
         super().__init__(dim=filt.dim, radial=filt.radial)
 
-    def _evaluate(self, omega):
-        return self._filter.evaluate(omega / self._scale)
+    def _evaluate(self, omega, **kwargs):
+        return self._filter.evaluate(omega / self._scale, **kwargs)
 
     def __str__(self):
         """
@@ -303,8 +324,7 @@ class ArrayFilter(Filter):
 
         self.xfer_fn_array = xfer_fn_array
 
-    def _evaluate(self, omega):
-
+    def _evaluate(self, omega, **kwargs):
         _input_pts = tuple(np.linspace(1, x, x) for x in self.xfer_fn_array.shape)
 
         # TODO: This part could do with some documentation - not intuitive!
@@ -318,7 +338,8 @@ class ArrayFilter(Filter):
         #  for values slightly outside the interpolation grid bounds.
         interpolator = RegularGridInterpolator(
             _input_pts,
-            self.xfer_fn_array,
+            # https://github.com/scipy/scipy/issues/17718
+            self.xfer_fn_array.astype(np.float64),
             method="linear",
             bounds_error=False,
             fill_value=None,
@@ -335,7 +356,8 @@ class ArrayFilter(Filter):
 
         return result
 
-    def evaluate_grid(self, L, dtype=np.float32, *args, **kwargs):
+    # No need to cache ArrayFilter
+    def evaluate_grid(self, L, *args, dtype=np.float32, **kwargs):
         """
         Optimized evaluate_grid method for ArrayFilter.
 
@@ -350,7 +372,6 @@ class ArrayFilter(Filter):
 
         See Filter.evaluate_grid for usage.
         """
-
         if all(dim == L for dim in self.xfer_fn_array.shape):
             logger.debug(
                 "Size of transfer function matches evaluate_grid size L exactly,"
@@ -359,7 +380,7 @@ class ArrayFilter(Filter):
             res = self.xfer_fn_array
         else:
             # Otherwise call parent code to generate a grid then evaluate.
-            res = super().evaluate_grid(L, dtype=dtype, *args, **kwargs)
+            res = super().evaluate_grid(L, *args, dtype=dtype, **kwargs)
         return res
 
 
@@ -371,7 +392,7 @@ class ScalarFilter(Filter):
     def __repr__(self):
         return f"Scalar Filter (dim={self.dim}, value={self.value})"
 
-    def _evaluate(self, omega):
+    def _evaluate(self, omega, **kwargs):
         return self.value * np.ones_like(omega)
 
 
@@ -386,9 +407,15 @@ class IdentityFilter(ScalarFilter):
 
 
 class CTFFilter(Filter):
+    """
+    Reproduce MATLAB's cryo_CTF_relion CTF (Contrast Transfer Function) Filter
+
+    Note if comparing to legacy MATLAB cryo_CTF_Relion,
+    take care regarding defocus unit conversion to/from nm.
+    """
+
     def __init__(
         self,
-        pixel_size=10,
         voltage=200,
         defocus_u=15000,
         defocus_v=15000,
@@ -400,17 +427,18 @@ class CTFFilter(Filter):
         """
         A CTF (Contrast Transfer Function) Filter
 
-        :param pixel_size:  Pixel size in angstrom
+        Note if comparing to legacy MATLAB cryo_CTF_Relion,
+        take care regarding defocus unit conversion to nm.
+
         :param voltage:     Electron voltage in kV
         :param defocus_u:   Defocus depth along the u-axis in angstrom
         :param defocus_v:   Defocus depth along the v-axis in angstrom
         :param defocus_ang: Angle between the x-axis and the u-axis in radians
-        :param Cs:          Spherical aberration constant
+        :param Cs:          Spherical aberration constant in mm
         :param alpha:       Amplitude contrast phase in radians
         :param B:           Envelope decay in inverse square angstrom (default 0)
         """
         super().__init__(dim=2, radial=defocus_u == defocus_v)
-        self.pixel_size = pixel_size
         self.voltage = voltage
         self.wavelength = voltage_to_wavelength(self.voltage)
         self.defocus_u = defocus_u
@@ -420,52 +448,67 @@ class CTFFilter(Filter):
         self.alpha = alpha
         self.B = B
 
-        self.defocus_mean = 0.5 * (self.defocus_u + self.defocus_v)
-        self.defocus_diff = 0.5 * (self.defocus_u - self.defocus_v)
+        # Convert angstrom to nm and divide by 2
+        self._defocus_mean_nm = 0.05 * (self.defocus_u + self.defocus_v)
+        self._defocus_diff_nm = 0.05 * (self.defocus_u - self.defocus_v)
 
-    def _evaluate(self, omega):
-        om_x, om_y = np.vsplit(omega / (2 * np.pi * self.pixel_size), 2)
+    def _evaluate(self, omega, **kwargs):
+        # Ensure we have a pixel size,
+        pixel_size = kwargs.get("pixel_size", None)
+        if pixel_size is None:
+            raise RuntimeError(
+                f"{self.__class__.__name__}.evaluate must be passed kwarg `pixel_size`."
+            )
+        # and that it is a floating point value.
+        pixel_size = float(pixel_size)
 
-        eps = np.finfo(np.pi).eps
-        ind_nz = (np.abs(om_x) > eps) | (np.abs(om_y) > eps)
-        angles_nz = np.arctan2(om_y[ind_nz], om_x[ind_nz])
-        angles_nz -= self.defocus_ang
+        # Reference MATLAB code, includes reference to paper
+        #    Mindell, J. A.; Grigorieff, N. (2003).
+        # https://github.com/PrincetonUniversity/aspire/blob/760a43b35453e55ff2d9354339e9ffa109a25371/projections/cryo_CTF_Relion.m#L34
+        #
+        # s, theta should match MATLAB's RadiusNorm up to a transpose
+        # To accomplish this given ASPIRE-Python's default `omega` grid,
+        # we unpack and remove the pi scaling,
+        # and further rescale the radii `s` by half below.
+        #
+        # Additionally we upcast so downstream computations remain in doubles.
+        x, y = omega.astype(np.float64, copy=False) / np.pi
 
-        defocus = np.zeros_like(om_x)
-        defocus[ind_nz] = self.defocus_mean + self.defocus_diff * np.cos(2 * angles_nz)
+        # Returns radii such that when multiplied by the
+        # bandwidth of the signal, we get the correct radial frequencies
+        # corresponding to each pixel in our nxn grid.
+        theta, s = cart2pol(x, y)
+        s = s / 2
 
-        c2 = -np.pi * self.wavelength * defocus
-        c4 = 0.5 * np.pi * (self.Cs * 1e7) * self.wavelength**3
+        # Wavelength in nm.
+        lamb = 1.22639 / np.sqrt(self.voltage * 1000 + 0.97845 * self.voltage**2)
 
-        r2 = om_x**2 + om_y**2
-        r4 = r2**2
-        gamma = c2 * r2 + c4 * r4
-        h = np.sqrt(1 - self.alpha**2) * np.sin(gamma) - self.alpha * np.cos(gamma)
+        # Divide by 10 to make pixel size in nm. BW is the
+        # bandwidth of the signal corresponding to the given pixel size.
+        BW = 1 / (pixel_size / 10)
+
+        s = s * BW
+        DFavg = self._defocus_mean_nm  # (DefocusU+DefocusV)/2
+        DFdiff = self._defocus_diff_nm  # (DefocusU-DefocusV)
+        # Note division by 2 is pre-computed in _defocus_diff_nm
+        df = DFavg + DFdiff * np.cos(2 * (theta - self.defocus_ang))
+
+        k2 = np.pi * lamb * df
+        # 10*6 converts Cs from mm to nm.
+        k4 = np.pi / 2 * 10**6 * self.Cs * lamb**3
+        chi = k4 * s**4 - k2 * s**2
+
+        h = np.sqrt(1 - self.alpha**2) * np.sin(chi) - self.alpha * np.cos(chi)
 
         if self.B:
-            h *= np.exp(-self.B * r2)
+            h *= np.exp(-self.B * s**2)
 
-        return h.squeeze()
-
-    def scale(self, c=1):
-        return CTFFilter(
-            pixel_size=self.pixel_size * c,
-            voltage=self.voltage,
-            defocus_u=self.defocus_u,
-            defocus_v=self.defocus_v,
-            defocus_ang=self.defocus_ang,
-            Cs=self.Cs,
-            alpha=self.alpha,
-            B=self.B,
-        )
+        return h
 
 
 class RadialCTFFilter(CTFFilter):
-    def __init__(
-        self, pixel_size=10, voltage=200, defocus=15000, Cs=2.26, alpha=0.07, B=0
-    ):
+    def __init__(self, voltage=200, defocus=15000, Cs=2.26, alpha=0.07, B=0):
         super().__init__(
-            pixel_size=pixel_size,
             voltage=voltage,
             defocus_u=defocus,
             defocus_v=defocus,
@@ -474,3 +517,45 @@ class RadialCTFFilter(CTFFilter):
             alpha=alpha,
             B=B,
         )
+
+
+class BlueFilter(Filter):
+    """
+    Filter where power increases with frequency.
+    """
+
+    def __init__(self, dim=None, var=1):
+        super().__init__(dim=dim, radial=True)
+        self.var = var
+
+    def __repr__(self):
+        return f"BlueFilter(dim={self.dim}, var={self.var})"
+
+    def _evaluate(self, omega, **kwargs):
+        f = np.sqrt(omega[0])
+        m = np.mean(f)
+        f = f / m
+
+        return self.var * f
+
+
+class PinkFilter(Filter):
+    """
+    Filter where power decreases with frequency.
+    """
+
+    def __init__(self, dim=None, var=1):
+        super().__init__(dim=dim, radial=True)
+        self.var = var
+
+    def __repr__(self):
+        return f"PinkFilter(dim={self.dim}, var={self.var})"
+
+    def _evaluate(self, omega, **kwargs):
+        step = np.abs(np.subtract(*omega[0][:2]))
+        # Avoid zero division
+        f = np.sqrt(2 * step / (omega[0] + step))
+        m = np.mean(f)
+        f = f / m
+
+        return self.var * f
