@@ -2,6 +2,8 @@
 Define a Rotation Class for customized rotation operations used by ASPIRE.
 """
 
+import warnings
+
 import numpy as np
 from numpy.linalg import norm
 from scipy.linalg import svd
@@ -11,11 +13,13 @@ from aspire.utils.random import Random
 
 
 class Rotation:
-    def __init__(self, matrices):
+    def __init__(self, matrices, gimble_lock_warnings=True):
         """
          Initialize a Rotation object
 
         :param matrices: Rotation matrices to initialize Rotation object.
+        :param gimble_lock_warnings: Optionally show gimble lock warnings.
+             Defaults to `True`, setting to `False` filters warnings.
         """
         if matrices.ndim == 2:
             matrices = matrices.reshape((1, 3, 3))
@@ -26,6 +30,8 @@ class Rotation:
         self._matrices = matrices
         self._seq_order = "ZYZ"
 
+        self.gimble_lock_warnings = bool(gimble_lock_warnings)
+
     def __str__(self):
         """
         String representation.
@@ -65,7 +71,22 @@ class Rotation:
         :return: Rotation matrices as a n x 3 array
         """
         rotations = sp_rot.from_matrix(self._matrices.astype(self.dtype))
-        return rotations.as_euler(self._seq_order, degrees=False).astype(self.dtype)
+
+        # Filter Gimbal lock warnings.
+        with warnings.catch_warnings():
+            msg = "Gimbal lock detected*"
+
+            disposition = "once"
+            if not self.gimble_lock_warnings:
+                disposition = "ignore"
+
+            warnings.filterwarnings(disposition, message=msg, category=UserWarning)
+
+            euler_angles = rotations.as_euler(self._seq_order, degrees=False).astype(
+                self.dtype
+            )
+
+        return euler_angles
 
     def invert(self):
         """
@@ -212,7 +233,7 @@ class Rotation:
         r2 = self._matrices[j]
         ut = np.dot(r2, r1.T)
         alpha_ij = np.arctan2(ut[2, 0], -ut[2, 1]) + np.pi
-        alpha_ji = np.arctan2(ut[0, 2], -ut[1, 2]) + np.pi
+        alpha_ji = np.arctan2(-ut[0, 2], ut[1, 2]) + np.pi
 
         ell_ij = alpha_ij * ell / (2 * np.pi)
         ell_ji = alpha_ji * ell / (2 * np.pi)
@@ -223,30 +244,41 @@ class Rotation:
         return ell_ij, ell_ji
 
     @staticmethod
-    def from_euler(values, dtype=np.float32):
+    def from_euler(values, dtype=None):
         """
         build rotation object from Euler angles in radians
 
-        :param dtype:  data type for rotational angles and matrices
         :param values: Rotation angles in radians, as a n x 3 array
+        :param dtype: Optional output dtype.  Infers from `values` by default when possible,
+            otherwise defaults to `float32`.
+
         :return: new Rotation object
         """
+        dtype = dtype or getattr(values, "dtype", np.float32)
         rotations = sp_rot.from_euler("ZYZ", values, degrees=False)
-        matrices = rotations.as_matrix().astype(dtype)
+        matrices = rotations.as_matrix().astype(dtype, copy=False)
         return Rotation(matrices)
 
     @staticmethod
-    def about_axis(axis, angles, dtype=np.float32):
+    def about_axis(axis, angles, dtype=None, gimble_lock_warnings=True):
         """
         Build rotation object from axis and angles of rotation.
 
         :param axis: A string denoting the axis of rotation. "x", "y", or "z".
         :param angles: Rotation angles in radians. `angles` can be a single value,
             or an array of shape (N,) or (N,1).
-        :param dtype: Data type for rotation matrices.
+        :param dtype: Optional output dtype.  Infers from `angles` by default when possible,
+            otherwise defaults to `float32`.
+        :param gimble_lock_warnings: Optionally show gimble lock warnings.
+             Defaults to `True`, setting to `False` filters warnings.
 
         :return: Rotation object
         """
+        dtype = dtype or (
+            np.float64
+            if isinstance(angles, float)
+            else getattr(angles, "dtype", np.float32)
+        )
         axes = ["x", "y", "z"]
         if axis.lower() not in axes:
             raise ValueError("`axis` must be 'x', 'y', or 'z'.")
@@ -263,18 +295,53 @@ class Rotation:
 
         rotation = sp_rot.from_euler(axis, angles, degrees=False)
         matrix = rotation.as_matrix().astype(dtype)
-        return Rotation(matrix)
+        rot = Rotation(matrix, gimble_lock_warnings=gimble_lock_warnings)
+
+        return rot
+
+    def as_rotvec(self):
+        """
+        Return this `Rotation` object as a rotation vector.
+
+        A rotation vector is a 3D vector which is co-directional to
+        the axis of rotation and whose norm gives the angle of
+        rotation in radians. The angle of rotation is
+        counter-clockwise about the axis.
+
+        See `from_rotvec`.
+
+        :return: (..., 3) array
+        """
+        return sp_rot.from_matrix(self.matrices).as_rotvec()
 
     @staticmethod
-    def from_matrix(values, dtype=np.float32):
+    def from_rotvec(vec, dtype=None):
+        """
+        Build a Rotation object from rotation vectors. A rotation vector is a
+        3D vector which is co-directional to the axis of rotation and whose norm
+        gives the angle of rotation in radians. The angle of rotation is counter-clockwise
+        about the axis.
+
+        :param vec: array_like, shape (N, 3) or (3,)
+        :param dtype: Optional output dtype.  Infers from `vec` by default.
+        :return: Rotation object
+        """
+        dtype = dtype or vec.dtype
+        rots = sp_rot.from_rotvec(vec)
+        matrices = rots.as_matrix().astype(dtype, copy=False)
+        return Rotation(matrices)
+
+    @staticmethod
+    def from_matrix(values, dtype=None):
         """
         build rotation object from rotational matrices
 
-        :param dtype:  data type for rotational angles and matrices
+        :param dtype: Optional output dtype.  Infers from `values` by default.
         :param values: Rotation matrices, as a n x 3 x 3 array
         :return: new Rotation object
         """
-        return Rotation(values.astype(dtype))
+        dtype = dtype or values.dtype
+        return Rotation(values.astype(dtype, copy=False))
 
     @staticmethod
     def generate_random_rotations(
@@ -304,23 +371,71 @@ class Rotation:
         return Rotation.from_euler(angles, dtype=dtype)
 
     @staticmethod
-    def angle_dist(r1, r2, dtype=np.float32):
+    def angle_dist(r1, r2, dtype=None):
         """
-        Find the angular distance between two rotation matrices. We first compute
-        the rotation between the two rotation matrices, r = r1 @ r2.T. Then using
-        the axis-angle representation of r we find the angle between r1 and r2.
+        Find the angular distance between two sets of rotation matrices. We first compute
+        the rotation between the two sets of rotation matrices, r = r1 @ r2.T. Then using
+        the axis-angle representation of r we find the angle between r1 and r2. Computations
+        will broadcast in the case of a singleton rotation and a set of rotations.
 
-        :param r1: A 3x3 rotation matrix
-        :paran r2: A 3x3 rotation matrix
+        :param r1: An nx3x3 array of rotation matrices, a singleton rotation,
+            or a `Rotation` object.
+        :param r2: An nx3x3 array of rotation matrices, a singleton rotation,
+            or a `Rotation` object.
+        :param dtype: Computation datatype. Default `None` infers from `r1`.
 
-        :return: The angular distance between r1 and r2 in radians.
+        :return: The element-wise angular distance between r1 and r2 in radians.
         """
-        r = r1 @ r2.T
-        tr_r = np.trace(r, dtype=dtype)
-        if np.allclose(tr_r, 3):
-            dist = 0
-        else:
-            theta = (tr_r - 1) / 2
-            theta = max(min(theta, 1), -1)  # Clamp theta in [-1,1]
-            dist = np.arccos(theta, dtype=dtype)
+        if isinstance(r1, Rotation):
+            r1 = r1.matrices
+        if isinstance(r2, Rotation):
+            r2 = r2.matrices
+
+        # Handle singletons.
+        r1 = r1.reshape(-1, 3, 3)
+        r2 = r2.reshape(-1, 3, 3)
+
+        # Check that shapes are compatible.
+        if (r1.shape[0] != 1) and (r2.shape[0] != 1):
+            if r1.shape[0] != r2.shape[0]:
+                raise ValueError(r"r1 and r2 are not broadcastable.*")
+
+        dtype = np.dtype(dtype or r1.dtype)
+
+        r2_T = np.transpose(r2, axes=(0, 2, 1))
+        r = r1 @ r2_T
+        tr_r = np.trace(r, axis1=1, axis2=2, dtype=dtype)
+
+        dist = np.zeros(max(len(r1), len(r2)), dtype=dtype)
+        non_zero_dist_ind = np.where(abs(tr_r - 3.0) > np.finfo(tr_r.dtype).resolution)
+
+        theta = (tr_r[non_zero_dist_ind] - 1) / 2
+        theta = np.maximum(np.minimum(theta, 1), -1)  # Clamp theta in [-1,1]
+        dist[non_zero_dist_ind] = np.arccos(theta, dtype=dtype)
+
+        # Return scalar for single value.
+        if dist.size == 1:
+            dist = dist.flat[0]
         return dist
+
+    @staticmethod
+    def mean_angular_distance(rots_1, rots_2, dtype=None):
+        """
+        Find the mean angular distance between two sets of rotation matrices. Computations
+        will broadcast in the case of a singleton rotations and a set of rotations.
+
+        :param rots_1: An nx3x3 array of rotation matrices, a singleton rotation,
+            or a `Rotation` object.
+        :param rots_2: An nx3x3 array of rotation matrices, a singleton rotation,
+            or a `Rotation` object.
+        :param dtype: Data type for computation. Default infers dtype from `rots_1`.
+
+        :return: The mean angular distance between rotations in radians.
+        """
+
+        dtype = np.dtype(dtype or rots_1.dtype)
+
+        # Shapes will be checked by `angle_dist`.
+        ang_dist = Rotation.angle_dist(rots_1, rots_2, dtype=dtype)
+
+        return np.mean(ang_dist)
